/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "types.h"
#include "proto.h"

/* some systems don't declare these in the header */
extern char *optarg;
extern int optind, opterr;

char buffer[160];

u8 mem[65536];
u16 loc = 0;

struct options opt;

/* sort out the options */

void parse_options(int argc, char **argv) {
	char optchar;

	memset(&opt, 0, sizeof(struct options));

	while((optchar = getopt(argc, argv, "b:def:ho:sv")) >= 0) {
		switch(optchar) {
			case 'b':
				opt.baseaddr = new_atoi(optarg, 0xffff);
				break;
			case 'd':
				opt.debug = 1;
				break;
			case 'e':
				opt.export = 1;
				break;
			case 'f':
				if(!strcasecmp(optarg, "cbm"))
					opt.format = F_CBM;
				else if(!strcasecmp(optarg, "segment"))
					opt.format = F_SEG;
				else if(!strcasecmp(optarg, "mem"))
					opt.format = F_MEM;
				else if(!strcasecmp(optarg, "mem8"))
					opt.format = F_MEM8;
				else if(!strcasecmp(optarg, "mem16"))
					opt.format = F_MEM16;
				else if(!strcasecmp(optarg, "mem32"))
					opt.format = F_MEM32;
				else {
					fputs("orgasm: unknown output format\n", stderr);
					exit(1);
					}
				break;
			case 'o':
				opt.outfilename = optarg;
				break;
			case 'h':
				fputs("orgasm " VERSION " (" __DATE__ ")\n", stderr);
				fputs("Usage: orgasm [options] inputfile\n"
				      "Options are:\n"
				      "     -b addr       base addr for mem8, mem16 & mem32 formats\n"
				      "     -d            enable debugging code\n"
				      "     -e            output exported symbols to a file\n"
				      "     -f format     select output format\n"
				      "                   (cbm/segment/mem/mem8/mem16/mem32)\n"
				      "     -h            display this message\n"
				      "     -o outfile    select output filename\n"
				      "     -v            be verbose\n", stderr);
				exit(0);
			case 'v':
				opt.verbose = 1;
				break;
			default:
				exit(1);
			}
		}
	
	opt.infilename = argv[optind];
	if(!opt.infilename) {
		fputs("orgasm: no input file\n", stderr);
		exit(1);
		}
	
	if(!opt.outfilename)
		opt.outfilename = "a.out";

	return;
	}

void free_all(void) {
	file_free();
	macro_free();
	segment_free();
	free_symtab();
	free_buffer();
	}

int main(int argc, char **argv) {
	char *line;

	memset(mem, 0, 65536);

	parse_options(argc, argv);

	preprocess(opt.infilename);

	build_symtab();

	validate_symtab();

	assemble();

	save(opt.outfilename);

	if(opt.export)
		export_symbols(opt.outfilename);

	free_all();

	return 0;
	}
