/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>
#endif

#include "types.h"
#include "proto.h"

struct segment *segs = NULL;
struct segment *seg_last = NULL;
int num_segs = 0;

/* add a new segment struct */

struct segment *add_segment(u16 org) {
	struct segment *new, *scan;

	for(scan = segs; scan; scan = scan->next)
		if(scan->org == org)
			do_error("a segment with org $%04x already exists\n", org);
	
	if(!(new = (struct segment *) malloc(sizeof(struct segment))))
		do_oom();
		
	new->org = org;
	new->len = 0;

	new->next = NULL;
	
	if(!segs)
		segs = new;
	else
		seg_last->next = new;
	
	seg_last = new;

	num_segs++;

	return new;
	}

void save(char *filename) {
	FILE *file;
	struct segment *scan;
	int org = 0x10000, end = 0;
	u8 val;

	if(!(file = fopen(filename, "w+")))
		do_error("couldn't open '%s' for writing\n", filename);

	if(opt.verbose) {
		if(!num_segs) {
			printf("orgasm: no segments, so not saving anything.\n");
			return;
			}
		else
			printf("orgasm: %d segment(s) assembled.\n", num_segs);
		}
	
	switch(opt.format) {
		case F_CBM:
			for(scan = segs; scan; scan = scan->next) {
				if(scan->org < org)
					org = scan->org;
				if(end < (scan->org + scan->len))
					end = (scan->org + scan->len);
				}
			val = org & 0xff;
			fwrite(&val, sizeof(u8), 1, file);
			val = org >> 8;
			fwrite(&val, sizeof(u8), 1, file);
			fwrite(&mem[org], sizeof(u8), (end - org), file);
			break;
		case F_SEG:
			for(scan = segs; scan; scan = scan->next) {
				val = scan->org & 0xff;
				fwrite(&val, sizeof(u8), 1, file);
				val = scan->org >> 8;
				fwrite(&val, sizeof(u8), 1, file);
				val = scan->len & 0xff;
				fwrite(&val, sizeof(u8), 1, file);
				val = scan->len >> 8;
				fwrite(&val, sizeof(u8), 1, file);
				fwrite(&mem[scan->org], sizeof(u8), scan->len, file);
				}
			break;
		case F_MEM:
			fwrite(mem, sizeof(u8), 0x10000, file);
			break;
		case F_MEM8:
			if(opt.baseaddr > 0xe000)
				goto err_too_high;
			fwrite(&mem[opt.baseaddr], sizeof(u8), 0x2000, file);
			break;
		case F_MEM16:
			if(opt.baseaddr > 0xc000)
				goto err_too_high;
			fwrite(&mem[opt.baseaddr], sizeof(u8), 0x4000, file);
			break;
		case F_MEM32:
			if(opt.baseaddr > 0x8000)
				goto err_too_high;
			fwrite(&mem[opt.baseaddr], sizeof(u8), 0x8000, file);
			break;
		}
	
	fclose(file);
	
	return;

err_too_high:
	do_error("base address too high for this format\n");
	}

void segment_free(void) {
	struct segment *scan, *next;

	for(scan = segs; scan; scan = next) {
		next = scan->next;
		free(scan);
		}
	}
