/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>
#endif

#include "types.h"
#include "proto.h"

struct symbol *symtab = NULL;
struct symbol *sym_last = NULL;

/* create a new symbol in the table */

struct symbol *create_symbol(char *name) {
	struct symbol *new;
	
	if(!(new = (struct symbol *) malloc(sizeof(struct symbol))))
		do_oom();
		
	if(!(new->name = strdup(name)))
		do_oom();
		
	new->defined = 0;
	new->exported = 0;

	if(*name == '@')
		new->vol = 1;
	else
		new->vol = 0;

	new->next = NULL;
	
	if(!symtab)
		symtab = new;
	else
		sym_last->next = new;
	
	sym_last = new;

	return new;
	}

/* add an empty symbol into the table */

struct symbol *add_symbol(char *name) {
	struct symbol *scan;

	for(scan = symtab; scan; scan = scan->next)
		if(!strcmp(scan->name, name))
			return NULL;
	
	scan = create_symbol(name);

	scan->defined = 0;
	scan->filename = file_top->filename;
	scan->lineno = file_top->lineno;

	return scan;
	}

/* add a valued symbol into the table, or update a value if it already exists */

struct symbol *add_symbol_value(char *name, u16 value) {
	struct symbol *scan;

	for(scan = symtab; scan; scan = scan->next)
		if(!strcmp(scan->name, name))
			break;
	
	if(!scan)
		scan = create_symbol(name);
	
	if(!scan->defined || scan->vol) {
		scan->value = value;
		scan->defined = 1;
		}
	else
		do_error("multiple definition of '%s'\n", name);
	
	return scan;
	}

struct symbol *get_symbol(char *name) {
	struct symbol *scan;

	for(scan = symtab; scan; scan = scan->next)
		if(!strcmp(scan->name, name)) {
			return scan;
			}
	
	return NULL;
	}

void validate_symtab(void) {
	struct symbol *scan;
	
	for(scan = symtab; scan; scan = scan->next)
		if(!scan->defined)
			do_error("undefined symbol '%s'\n", scan->name);
	}

void dump_symtab(void) {
	struct symbol *scan;
	
	for(scan = symtab; scan; scan = scan->next)
		printf("$%04x\t%s\n", scan->value, scan->name);
	}

void free_symtab(void) {
	struct symbol *scan, *next;

	next = symtab;
	for(; scan; scan = next) {
		next = scan->next;
		free(scan);
		}
	}

void export_symbols(char *filename) {
	FILE *file;
	struct symbol *scan;

	if(!(file = fopen(filename, "w+"))) {
		do_warning("couldn't open '%s' for writing\n", filename);
		return;
		}

	for(scan = symtab; scan; scan = scan->next)
		if(scan->exported) {
			if(scan->value & 0xff00)
				fprintf(file, "%s\t\t= $%04x\n", scan->name, scan->value);
			else
				fprintf(file, "%s\t\t= $%02x\n", scan->name, scan->value);
			}

	fclose(file);
	}

inline int issymchar(char c) {
	if(isalnum(c) || (c == '_') || (c == '@'))
		return 1;
	return 0;
	}

int issymbol(char *token) {
	int i;
	char c;

	if(!token || !issymchar(token[0]) || isdigit(token[0]))
		return 0;

	for(i = 0; i < strlen(token); i++)
		if(!issymchar(token[i])) {
			break;
			}
	
	c = token[i];
	token[i] = 0;
	
	if(get_opcode(token)) {
		token[i] = c;
		return 0;
		}
	
	token[i] = c;
	return 1;
	}

int islabel(char *token) {
	int i = strlen(token) - 1;

	if((i == 1) || (token[i] != ':'))
		return 0;

	token[i] = '\0';

	if(!issymbol(token)) {
		token[i] = ':';
		return 0;
		}

	token[i] = ':';
	return 1;
	}
