/* 
 * Copyright (c) Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * Abelstate() -- this function takes an IVF data structure and symbol
 *                table, and converts it to an Abel model. The Abel model
 *                is printed to stdout.
 */

#include "cyco.h"

/* You are not going to believe this: Data I/O's Abel language actually
 * imposes a limit on the length of a line. That limit is 150 chars, but
 * we will look for an opportunity to print a newline as soon as we've
 * printed MAX_LINE chars:
 */
#define MAX_LINE 55

extern int no_tidy;             /* command line option */
extern int tidy_before;         /* command line option */

static void abl_print_com(const char* operator, int inv, struct gate *gate);
static void abl_print_reg(struct gate *gate, int async_sr);
static void abl_print_mux(struct gate *gate);
static void abl_print_addc(struct gate *gate, int subc);
static void abl_print_nlgate(struct symbol *net);
static void abl_print_nlgate_reversed(struct symbol *net);
static void abl_print_err(const char *gate_name);
static void abl_print_adder(struct adder *adder);
static int abl_is_broken_vector(struct bitvec *bv);
static int abl_must_declare(struct symbol *s);


/* Print Abel model to stdout
 */
void abelstate(const char *modelname,
               int numgates,
               struct gate **sortedgates,
               struct port *ports,
               struct symbol *symchain)
{
   struct gate *g;
   struct gateport *gp;
   int i;
   struct symbol *s;
   struct port *p;
   struct adder *adder, *adder_list = 0;
   char modname[MAX_NAME+1], *dot;

   /* fixup model name which is VHDL-style ENTITY.ARCHITECTURE: */
   strcpy(modname, modelname);
   dot = strchr(modname, '.');
   if (dot != 0)
      *dot = 0;                 /* drop ARCHITECTURE part */

   if (!no_tidy && tidy_before)
      cleanup_netlist(sortedgates, numgates, ports);

   /* inspect all symbols and find out if they are part of a vector: */
   for (s = symchain; s != 0; s = s->nextinsymchain)
   {
      s->alias = 0;
      s->back = 0;
      extract_vector_range(s);
   }

   /* make a pass over all gates to see what sigs are being used */
   for (i = 0; i < numgates; i += 1)
   {
      g = sortedgates[i];
      if (g->flags & (GF_MODEL_USE | GF_REMOVED))
         continue;
      /* first mark all signal names */
      if ((g->type == NL_GATE) && (g->incnt == 1)
          && (g->outcnt == 1) && (g->in->port->name[0] == '$'))
      {                         /* this is a constant ... */
         if (g->in->port->name[1] == '0')
            g->out->port->flags |= SYMF_CONST_0;
         else if (g->in->port->name[1] == '1')
            g->out->port->flags |= SYMF_CONST_1;
         else
            g->out->port->flags |= SYMF_CONST_X;
      }
      else
      {
         for (gp = g->in; gp != 0; gp = gp->next)
            gp->port->flags |= SYMF_USED;
         for (gp = g->out; gp != 0; gp = gp->next)
            gp->port->flags |= SYMF_USED;
      }
      /* if a latch, then Abel replaces assign(=) to D with assign(:=) to Q */
      if ((g->type==LATCH)
          || (g->type==LATCH_N)
          || (g->type==LATCH_SR)
          || (g->type==LATCH_SR_N))
      {
         if (g->outcnt != 1)
         {
            fprintf(stderr,"** latch with multiple outputs\n");
            exit(1);
         }
         g->out->port->flags |= SYMF_IS_LATCH_Q;
      }
      else if ((adder = extract_adder(g)) != 0)
      {  /* this gate is part of an adder structure */
         int i;
         const char *op = (adder->type == A_add ? "adder" : "subtractor");
         for (i = 0; i < adder->size; i += 1)
         {
            struct gate *g_carry = adder->vec[i].g_carry;
            /* internal carry chain disappears from Abel netlist: */
            if (g_carry != 0)
            {
               g_carry->out->port->flags |= SYMF_MODEL_USE;
               g_carry->flags |= GF_MODEL_USE;
            }
            adder->vec[i].g_sum->flags |= GF_MODEL_USE;
         }
         fprintf(stderr,
                 "-- INFO: extracted %1d-bit %s, msb of sum is %s\n",
                 adder->size, op, adder->name);
         adder->next = adder_list;
         adder_list = adder;
      }
   }

   if (!no_tidy && !tidy_before)
      cleanup_netlist(sortedgates, numgates, ports);
   
   /* ok, let's go */
   printf("\nMODULE %s\n", modname);

   printf("\nDECLARATIONS\n");

   printf("\n\" ---------- input/output:\n\n");
   for (p = ports; p != 0; p = p->next)
   {
      struct symbol *s = p->port;
      s->flags |= (p->type == 1) ? SYMF_OUTPUT : SYMF_INPUT;
      if ((s->bitvec == 0) || (s->bitvec->size == 1))
      {
         if (p->type == 0)         /* input */
            printf("%s PIN;\n", sym_name_real(s));
         else                      /* output */
            printf("%s PIN ISTYPE '%s';\n", sym_name_real(s),
                   (s->flags & SYMF_IS_LATCH_Q) ? "reg" : "com");
      }
      else if (!(s->bitvec->flags & BVF_DECL_DONE))
      {                         /* this is a vector */
         struct bitvec *bv = s->bitvec;
         struct symbol *sym_hi = bv->bits[0];
         struct symbol *sym_lo = bv->bits[bv->size-1];
         char hi[MAX_NAME+1], lo[MAX_NAME+1];
         strcpy(hi, sym_name_real(sym_hi));
         strcpy(lo, sym_name_real(sym_lo));
         bv->flags |= BVF_DECL_DONE;
         if (p->type == 0)         /* input */
            printf("%s..%s PIN;\n", hi, lo);
         else                      /* output */
            printf("%s..%s PIN ISTYPE '%s';\n", hi, lo,
                   (s->flags & SYMF_IS_LATCH_Q) ? "reg" : "com");
      }
   }

   printf("\n\" ---------- internal wires/regs:\n\n");

   for (s = symchain; s != 0; s = s->nextinsymchain)
      if (abl_must_declare(s))
         if ((s->bitvec == 0)
             || (s->bitvec->size == 1)
             || abl_is_broken_vector(s->bitvec))
            printf("%s NODE ISTYPE '%s';\n", sym_name(s),
                   (s->flags & SYMF_IS_LATCH_Q) ? "reg" : "com");
         else if (!(s->bitvec->flags & BVF_DECL_DONE))
         {                         /* this is a vector */
            struct bitvec *bv = s->bitvec;
            struct symbol *sym_hi = bv->bits[0];
            struct symbol *sym_lo = bv->bits[bv->size-1];
            char hi[MAX_NAME+1], lo[MAX_NAME+1];
            strcpy(hi, sym_name(sym_hi));
            strcpy(lo, sym_name(sym_lo));
            bv->flags |= BVF_DECL_DONE;
            printf("%s..%s NODE ISTYPE '%s';\n", hi, lo,
                   (s->flags & SYMF_IS_LATCH_Q) ? "reg" : "com");
         }

   printf("\nEQUATIONS\n\n");

   for (adder = adder_list; adder != 0; adder = adder->next)
      abl_print_adder(adder);

   for (i = 0; i < numgates; i += 1)
   {
      g = sortedgates[i];
      if (g->flags & (GF_MODEL_USE | GF_REMOVED))
         continue;
      switch (g->type)
      {
      case NL_GATE:      abl_print_com("", 0, g); break;
      case NL_OPTIMIZED: abl_print_err("nl_optimized"); break;
      case NOT:          abl_print_com("!", 0, g); break;
      case AND:          abl_print_com("&", 0, g); break;
      case OR:           abl_print_com("#", 0, g); break;
      case XOR:          abl_print_com("$", 0, g); break;
      case NAND:         abl_print_com("&", 1, g); break;
      case NOR:          abl_print_com("#", 1, g); break;
      case XNOR:         abl_print_com("$", 1, g); break;
      case LATCH:        abl_print_reg(g, 0); break;
      case LATCH_SR:     abl_print_reg(g, 1); break;
      case LATCH_N:      abl_print_err("latch_n"); break;
      case LATCH_SR_N:   abl_print_err("latch_sr_n"); break;
      case NEGC:         abl_print_err("negc"); break;
      case ADDC:         abl_print_addc(g, 0); break;
      case SUBC:         abl_print_addc(g, 1); break;
      case BUF:          abl_print_com("", 0, g); break;
      case MUX:          abl_print_mux(g); break;
      case BUFIF0:       abl_print_err("bufif0"); break;
      case PMOS:         abl_print_err("pmos"); break;
      case BUFIF1:       abl_print_err("bufif1"); break;
      case NMOS:         abl_print_err("nmos"); break;
      case NOTIF0:       abl_print_err("notif0"); break;
      case NOTIF1:       abl_print_err("notif1"); break;
      case WIRE:         abl_print_com("", 0, g); break;
      case TRI:          abl_print_err("tri"); break;
      case WAND:         abl_print_err("wand"); break;
      case TRIAND:       abl_print_err("triand"); break;
      case WOR:          abl_print_err("wor"); break;
      case TRIOR:        abl_print_err("trior"); break;
      case TRI0:         abl_print_err("tri0"); break;
      case TRI1:         abl_print_err("tri1"); break;
      case SUPPLY0:      abl_print_err("supply0"); break;
      case SUPPLY1:      abl_print_err("supply1"); break;
      default:           abl_print_err("<unknown>"); break;
      }
   }   

   /* OUT ports that are constant or aliased need explicit assignment */
   for (p = ports; p != 0; p = p->next)
      if (p->port->flags & SYMF_OUTPUT)
      {
         if ((p->port->flags & (SYMF_CONST_0 | SYMF_CONST_1))
             || ((p->port->alias != 0) && (p->port->back != p->port)))
            abl_print_nlgate(p->port);
      }
      else if ((p->port->flags & SYMF_INPUT)
               && ((p->port->alias != 0) && (p->port->back != p->port)))
         /* the circuit refers to inport by its alias ... We need to
          * make an explicit connection between inport and alias: */
         abl_print_nlgate_reversed(p->port);

   printf("\nEND %s\n\n", modname);
}


/* Abel model of a combinational gate.
 */
static void abl_print_com(const char* operator,
                          int inv,
                          struct gate *gate)
{
   struct gateport *gp = gate->in;
   int len;
   if ((gate->outcnt != 1) || (gate->incnt <= 0))
   {
      fprintf(stderr, "** gate has unexpected number of ports\n");
      exit(1);
   }
   if (is_const(gate->out->port))
      return;
   len = printf("%s = ", gp_name(gate->out));
   if (inv)
      len += printf("!(");
   if (gp->next == 0)
      printf("%s%s", operator, gp_name(gp));
   else
   {
      for (; gp->next != 0; gp = gp->next)
         if ((len += printf("%s %s ", gp_name(gp), operator)) > MAX_LINE)
            len = printf("\n     ");
      printf("%s", gp_name(gp));
   }
   if (inv)
      printf(")");
   printf(";\n");
}


/* Abel model of a flipflop.
 */
static void abl_print_reg(struct gate *gate,
                          int async_sr)
{
   struct gateport *gp_d = gate->in;
   struct gateport *gp_s, *gp_r, *gp_ck;
   char bf1[MAX_NAME+1], bf2[MAX_NAME+1];
   if ((gate->outcnt != 1)
       || (!async_sr && (gate->incnt != 2))
       || (async_sr && (gate->incnt != 4)))
   {
      fprintf(stderr, "** latch has unexpected number of ports\n");
      exit(1);
   }
   strcpy(bf1, gp_name(gate->out));
   strcpy(bf2, gp_name(gp_d));
   printf("%s := %s;\n", bf1, bf2);
   if (async_sr)
   {
      gp_s = gp_d->next;
      gp_r = gp_s->next;
      gp_ck = gp_r->next;
   }
   else
      gp_ck = gp_d->next;
   strcpy(bf1, gp_name(gate->out));
   strcpy(bf2, gp_name(gp_ck));
   printf("%s.CLK = %s;\n", bf1, bf2);
   if (async_sr)
   {
      strcpy(bf1, gp_name(gp_s)); /* async set */
      if (strcmp(bf1, "0") != 0)
         printf("%s.ASET = %s;\n", gp_name(gate->out), bf1);
      strcpy(bf1, gp_name(gp_r)); /* async reset */
      if (strcmp(bf1, "0") != 0)
         printf("%s.ACLR = %s;\n", gp_name(gate->out), bf1);
   }
}


/* Abel model of a mux
 */
static void abl_print_mux(struct gate *gate)
{
   struct gateport *gp_l, *gp_c, *gp_t, *gp_f;
   char bf1[MAX_NAME+1], bf2[MAX_NAME+1], bf3[MAX_NAME+1];
   if ((gate->outcnt != 1) || (gate->incnt != 3))
   {
      fprintf(stderr, "** mux has unexpected number of ports\n");
      exit(1);
   }
   gp_l = gate->out;
   gp_c = gate->in;
   gp_t = gp_c->next;
   gp_f = gp_t->next;
   strcpy(bf1, gp_name(gp_l));
   strcpy(bf2, gp_name(gp_t));
   strcpy(bf3, gp_name(gp_f));
   if (strcmp(bf2, "0") == 0)
      /* special case that strength-reduction cannot handle because we have
       * no gate in the IVF library with function (!A & B).
       */
      printf("%s = !%s & %s;\n", bf1, gp_name(gp_c), bf3);
   else if (strcmp(bf3, "1") == 0)
      /* other special case for which no strength-reduced IVF gate exists */
      printf("%s = !%s + %s;\n", bf1, gp_name(gp_c), bf2);
   else
      printf("WHEN %s THEN %s = %s ELSE %s = %s;\n",
             gp_name(gp_c), bf1, bf2, bf1, bf3);
}


/* print Abel model for the carry propagate cell ADDC or SUBC
 */
static void abl_print_addc(struct gate *gate, int subc)
{
   struct gateport *gp_a, *gp_b, *gp_c = 0;
   char s_a[MAX_NAME+1], s_b[MAX_NAME+1], s_c[MAX_NAME+1];
   if ((gate->outcnt != 1) || ((gate->incnt != 3) && (gate->incnt != 2)))
   {
      fprintf(stderr, "** addc/subc has unexpected number of ports\n");
      exit(1);
   }
   gp_a = gate->in;
   gp_b = gp_a->next;
   strcpy(s_a, gp_name(gp_a));
   sprintf(s_b, "%s%s", (subc ? "!" : ""), gp_name(gp_b));
   if (gate->incnt == 3)
   {
      gp_c = gp_b->next;
      strcpy(s_c, gp_name(gp_c));
      printf("%s = (%s & %s) # (%s & %s) # (%s & %s);\n",
             gp_name(gate->out), s_a, s_b, s_a, s_c, s_b, s_c);
   }
   else if (!subc)
      printf("%s = %s & %s;\n", gp_name(gate->out), s_a, s_b);
   else
      printf("%s = %s # %s;\n", gp_name(gate->out), s_a, s_b);
}


static void abl_print_nlgate(struct symbol *net)
{
   struct symbol sym = *net;
   sym.flags &= ~(SYMF_CONST_0|SYMF_CONST_1);
   sym.alias = 0;
   sym.back = 0;
   printf("%s = %s;\n", sym_name(&sym), sym_name(net));
}

static void abl_print_nlgate_reversed(struct symbol *net)
{
   struct symbol sym = *net;
   sym.flags &= ~(SYMF_CONST_0|SYMF_CONST_1);
   sym.alias = 0;
   sym.back = 0;
   printf("%s = %s;\n", sym_name(net), sym_name(&sym));
}

static void abl_print_err(const char *gate_name)
{
   fprintf(stderr, "** cannot handle gate type %s\n", gate_name);
   exit(1);
}


/* print an add or sub operation, but be careful not to make the lines too
 * long because the Data I/O Abel parser chokes on lines longer than 150
 * chars.
 */
static void abl_print_adder(struct adder *adder)
{
   int i, len;
   const carry_group_size = 1;
   enum adder_type ty = adder->type;
   const char *op = (ty == A_add ? "+" : "-");
   const char *op2 = (ty == A_add ? "Adder" : "Subtractor");
   printf("\" %s: size = %1d bits, name = %s, carry group size = %1d\n",
          op2, adder->size, adder->name, carry_group_size);
   printf("@CARRY %1d;\n", carry_group_size);
   len = printf("  [");
   for (i = 0; i < adder->size - 1; i += 1)
      if ((len += printf("%s,", sym_name(adder->vec[i].y))) > MAX_LINE)
         len = printf("\n  ");
   printf("%s]\n", sym_name(adder->vec[i].y));
   len = printf("= [");
   for (i = 0; i < adder->size - 1; i += 1)
      if ((len += printf("%s,", sym_name(adder->vec[i].a))) > MAX_LINE)
         len = printf("\n  ");
   printf("%s]\n", sym_name(adder->vec[i].a));
   len = printf("%s [", op);
   for (i = 0; i < adder->size - 1; i += 1)
      if ((len += printf("%s,", sym_name(adder->vec[i].b))) > MAX_LINE)
         len = printf("\n  ");
   printf("%s];\n\n", sym_name(adder->vec[i].b));
}


/* if not all elements of the vector BV should be declared, then
 * return TRUE, else return FALSE.
 */
static int abl_is_broken_vector(struct bitvec *bv)
{
   int i;
   if (bv->flags & BVF_ANALIZED)
      return bv->flags & BVF_BROKEN; /* cached value */
   bv->flags &= !BVF_BROKEN;    /* clear `broken' flag */
   for (i = 0; i < bv->size; i += 1)
      if (!abl_must_declare(bv->bits[i]))
      {
         bv->flags |= BVF_BROKEN;
         break;
      }
   bv->flags |= BVF_ANALIZED;
   return bv->flags & BVF_BROKEN;;
}


static int abl_must_declare(struct symbol *s)
{
   return ((s->flags & SYMF_USED)
           && (s->flags & (SYMF_REFERENCED | SYMF_MODEL_OUTPUT))
           && !(s->flags & SYMF_MODEL_USE)
           && !is_ioport(s)
           && !is_const(s)
           && !(((s->id == 0) && (s->name[0] == '$'))
                || ((s->id != 0) && (s->id[0] == '$')))
      );
}
