/* 
 * Copyright (c) Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/* extract.c contains various functions to tidy -up the IVF netlist
 * and to extract models from the IVF netlist.
 */

#include "cyco.h"

/* used for extract_adder(), the MAX_XOR constant defines how many XOR gates
 * can be connected in the close neighborhood of an ADDC or SUBC cell before
 * we start failing to detect an adder that is actually there. (There is
 * nothing wrong with failing to detect an adder, it just is less efficient.)
 */
#define MAX_XOR 100


extern int no_model_extract;    /* command line option */
extern int no_const_prop;       /* command line option */
extern int remove_latches;      /* command line option */
extern int no_strength_reduce;  /* command line option */

static struct bitvec *new_bitvec(const char *name,
                                 int range_hi,
                                 int range_lo);
static void make_adder_section(struct gate *g_carry, struct gate *g_sum,
                               struct adder_section **vec, int *j);
static int is_full_adder(struct gate *fa_carry, struct gate **fa_sum,
                         enum adder_type adder_type);
static int is_half_adder(struct gate *fa_carry, struct gate **fa_sum,
                         enum adder_type adder_type);
static int count_fanout(struct gateport *gp);
static int is_stuck(struct gate *g, int *symf);
static void propagate_const(struct symbol *sym);
static void remove_gate(struct gate *g, int *n);
static void reduce_strength(struct gate *g, int *n_reduced);
static void make_alias(struct gate *g, struct symbol *sym);
static struct alias *merge_aliases(struct symbol *s1, struct symbol *s2);
static int reduce_fanin(struct gate *g, int symf);



/* print the name of a net
 */
const char *gp_name(struct gateport *gp)
{
   return sym_name(gp->port);
}


/* print a symbol
 */
const char *sym_name(struct symbol *s)
{
   const char *name;
   struct symbol *sym = s->alias ? s->back : s;
   char *nm = sym->id ? sym->id : sym->name;
   if (sym->flags & SYMF_CONST_0)
      name = "0";
   else if (sym->flags & SYMF_CONST_1)
      name = "1";
   else if (sym->flags & SYMF_CONST_X)
   {
      fprintf(stderr,"** illegal constant 'X' used\n");
      name = "X";
   }
   else if (nm[0] == '$')
   {
      if ((nm[1] != '0') && (nm[1] != '1'))
         fprintf(stderr,"** illegal constant '$%c' used\n", nm[1]);
      name = &nm[1];
   }
   else
      name = sym_name_real(sym);
   return name;
}


const char *sym_name_real(struct symbol *s)
{
   const char *name;
   struct symbol *sym = s->alias ? s->back : s;
   char *nm = sym->id ? sym->id : sym->name;
   char *p;
   static char buf[MAX_NAME+1];
   int len, i;
   p = strchr(nm, '/');
   if (p)
      nm = p + 1;
   strcpy(buf,nm);
   p = strrchr(buf, '/');
   if (p == 0)
      ;
   else if (sym->bitvec == 0)
      *p = 0;                /* remove trailing "/0" */
   else                      /* vector element */
      sprintf(p, "_%1d", sym->index);
   len = strlen(buf);
   for (i = 0; i < len; i += 1)
      if ((buf[i] == '!') && (buf[i+1] == 'T'))
      {
         buf[i] = 't';          /* names cannot start with "!T" */
         buf[i+1] = '_';
      }
      else if ((buf[i] == '!') && (buf[i+1] == 'W'))
      {
         buf[i] = '_';          /* names cannot start with "!T" */
         buf[i+1] = 'w';
      }
      else if ((buf[i] == '/') || (buf[i] == '[') || (buf[i] == ']'))
         buf[i] = '_';
   name = (const char *)buf;
   return name;
}


/* return TRUE if SYM is a constant
 */ 
int is_const(struct symbol *sym)
{
   return
      (sym->flags & (SYMF_CONST_0 | SYMF_CONST_1 | SYMF_CONST_X)) != 0;
}


/* return TRUE if SYM is an I/O port of the module
 */
int is_ioport(struct symbol *sym)
{
   return (sym->flags & (SYMF_INPUT | SYMF_OUTPUT)) != 0;
}

/* if SYM is an element in a vector then create and return a bitvec structure
 * for it. All elements in the bitvec will have their symbol->bitvec field
 * point to the newly created bitvec struct.
 * Return 0 if bitvec already exists in SYM->bitvec or if SYM is not an
 * element of a vector.
 */
struct bitvec *extract_vector_range(struct symbol *sym)
{
   const char *name = sym->id ? sym->id : sym->name;
   char t[MAX_NAME+1], *p;
   int this_index, range_hi, range_lo, i;
   struct symbol *s;
   struct bitvec *bitvec = 0;

   if (sym->bitvec != 0)
      return 0;                 /* bitvec already created on other occasion */
   if (strchr(name, '!') != 0)
      return 0;                 /* temp sigs are never vectors */
   strcpy(t, name);
   p = strrchr(t, '/');
   if (p == 0)
      return 0;                 /* never happens... */
   p = p + 1;
   this_index = atoi(p);
   range_hi = range_lo = this_index;
   /* find upper bound */
   for (i = this_index + 1; ; i += 1)
   {
      sprintf(p, "%1d", i);
      s = symfind(t, 0);
      if (s == 0)
         break;
      range_hi = i;
   }
   /* find lower bound */
   for (i = this_index - 1; i >= 0 ; i -= 1)
   {
      sprintf(p, "%1d", i);
      s = symfind(t, 0);
      if (s == 0)
         break;
      range_lo = i;
   }
   *(p-1) = 0;
   if ((this_index != 0) || (range_hi != range_lo))
      bitvec = new_bitvec(t, range_hi, range_lo);
   return bitvec;
}

/* allocate and initialize a new bitvec
 */
static struct bitvec *new_bitvec(const char *name,
                                 int range_hi,
                                 int range_lo)
{
   int namelen = strlen(name);
   char *nm = calloc(1, namelen+1);
   struct bitvec *bitvec = calloc(1, sizeof(struct bitvec));
   struct symbol **bits = calloc(range_hi - range_lo + 1,
                                 sizeof(struct symbol *));
   int i;
   char str[MAX_NAME+1];
   bitvec->name = strcpy(nm, name);
   bitvec->size = range_hi - range_lo + 1;
   bitvec->range_hi = range_hi;
   bitvec->range_lo = range_lo;
   bitvec->bits = bits;
   for (i = range_hi; i >= range_lo; i -= 1)
   {
      struct symbol *sym;
      sprintf(str, "%s/%1d", name, i);
      sym = symfind(str, 0);
      bits[range_hi - i] = sym;
      if (sym != 0)
      {
         sym->bitvec = bitvec;  /* back reference */
         sym->index = i;
      }
   }
   return bitvec;
}

/* starting from SEED, search its neighborhood to see if it is part of a
 * adder or subtractor structure. If so, create an adder struct for it and
 * mark all gates and nets that participate in the add/sub module with the
 * GF_MODEL_USE and SYMF_MODEL_USE flags.
 */
struct adder *extract_adder(struct gate* seed)
{
   int gate_type = seed->type;
   struct gate *fa_carry, *fa_sum = 0;
   int up_cnt, down_cnt, j = 0, k = 0;
   int i, sum_type, ctype = seed->type;
   struct adder_section *vec = 0, *vec2;
   struct adder *adder;
   const char *name;
   enum adder_type adder_type;
   if (no_model_extract)
      return 0;                 /* no model extraction requested */
   switch (gate_type)
   {
   case ADDC: adder_type = A_add; sum_type = XOR;  break;
   case SUBC: adder_type = A_sub; sum_type = XNOR; break;
   default: return 0;
   }
   if (!is_full_adder(seed, &fa_sum, adder_type))
      return 0;
   make_adder_section(seed, fa_sum, &vec, &j);
   /* search for the upper bound of the adder */
   for (fa_carry = seed, up_cnt = 0; ; up_cnt += 1)
   {
      struct symbol *outport = fa_carry->out->port;
      struct sensitivity *fanout;
      struct gate *g1 = 0, *g2 = 0, *g3;
      if (outport == 0) {free(vec); return 0;}
      fanout = outport->togate;
      if (fanout == 0) {free(vec); return 0;}
      g1 = fanout->gate;
      if (g1->flags & GF_REMOVED)
      {free(vec); return 0;}
      if (fanout->next == 0)    /* carry fans out to only 1 gate */
         if ((g1->type == sum_type) && (g1->incnt == 3) && (g1->outcnt == 1))
         {
            make_adder_section(0, g1, &vec, &j);
            break; /* found the most significant bit of the adder */
         }
         else
            {free(vec); return 0;}
      g2 = fanout->next->gate;
      if (g2->flags & GF_REMOVED)
      {free(vec); return 0;}
      if (fanout->next->next != 0) {free(vec); return 0;}    /* fanout != 2 */
      if (g1->type != ctype) {g3=g1; g1=g2; g2=g3;} /* exchange */
      if ((g1->type != ctype) || (g2->type != sum_type)) {free(vec); return 0;}
      if (!is_full_adder(g1, &g3, adder_type)) {free(vec); return 0;}
      if (g2 != g3) {free(vec); return 0;}
      make_adder_section(g1, g2, &vec, &j);
      fa_carry = g1;
   }
   /* search for the lower bound of the adder */
   for (fa_carry = seed, down_cnt = 0; ; down_cnt += 1)
   {
      struct gate *g1 = 0, *g2 = 0;
      struct sensitivity *fanin;
      /* 3rd input of ADDC/SUBC is used as the carry-in: */
      struct symbol *inport = fa_carry->in->next->next->port;
      if (inport == 0) {free(vec); return 0;}
      fanin = inport->fromgate;
      if (fanin == 0) {free(vec); return 0;}
      if (fanin->gate->flags & GF_REMOVED)
      {free(vec); return 0;}
      if (fanin->next != 0) {free(vec); return 0;} /* more than one driver */
      g1 = fanin->gate;
      if (is_half_adder(g1, &g2, adder_type))
      {
         make_adder_section(g1, g2, &vec, &j);
         break;                 /* found low-end of the adder */
      }
      if (!is_full_adder(g1, &g2, adder_type)) {free(vec); return 0;}
      if (count_fanout(g1->out) != 2) {free(vec); return 0;}
      make_adder_section(g1, g2, &vec, &j);
      fa_carry = g1;
   }
   adder = calloc(1, sizeof(struct adder));
   adder->type = adder_type;
   adder->size = up_cnt + down_cnt + 3;
   vec2 = calloc(adder->size, sizeof(struct adder_section));
   adder->vec = vec2;
   for (k = 0, i = up_cnt + 1; i >= 1; i -= 1)
      vec2[k++] = vec[i];
   vec2[k++] = vec[0];
   for (i = up_cnt + 2; i <= up_cnt + down_cnt + 2; i += 1)
      vec2[k++] = vec[i];
   free(vec);
   name = gp_name(vec2[0].g_sum->out);
   adder->name = calloc(1, strlen(name) + 1);
   strcpy(adder->name, name);
   for (i = 0; i < adder->size; i += 1)
      vec2[i].y->flags |= SYMF_MODEL_OUTPUT;
   return adder;
}

static void make_adder_section(struct gate *g_carry,
                               struct gate *g_sum,
                               struct adder_section **vec,
                               int *j)
{
   static int size;
   if (*vec == 0)
   {
      *j = 0;
      size = 300;
      *vec = calloc(size, sizeof(struct adder_section));
   }
   if (*j >= size)
   {
      size *= 4;
      *vec = realloc(*vec, size * sizeof(struct adder_section));
   }
   if (*vec == 0)
   {
      fprintf(stderr, "\n** not enough memory...\n\n");
      exit(1);
   }
   (*vec)[*j].g_carry = g_carry;
   (*vec)[*j].g_sum = g_sum;
   (*vec)[*j].y = g_sum->out->port;
   (*vec)[*j].a = g_sum->in->port;
   (*vec)[*j].b = g_sum->in->next->port;
   *j += 1;
}


/* Check to see that all 3 inputs of the FA_CARRY gate are also connected
 * to a 3-input XOR gate. If so, we consider the pair {FA_CARRY,FA_SUM} to
 * form a full-adder.
 */
static int is_full_adder(struct gate *fa_carry,
                         struct gate **fa_sum,
                         enum adder_type adder_type)
{
   struct gateport *gp;
   struct sensitivity *fanout;
   struct gate *xor[3][MAX_XOR+1];
   int i, j, k, sum_type;
   if ((fa_carry->incnt != 3) || (fa_carry->outcnt != 1)
       || ((adder_type == A_add) && (fa_carry->type != ADDC))
       || ((adder_type == A_sub) && (fa_carry->type != SUBC))
       || (fa_carry->flags & GF_REMOVED))
      return 0;
   sum_type = (adder_type == A_add) ? XOR : XNOR;
   for (i = 0; i <= 2; i += 1)
      for (j = 0; j <= MAX_XOR; j += 1)
         xor[i][j] = 0;
   for (i = 0, gp = fa_carry->in; gp != 0; i += 1, gp = gp->next)
      for (fanout = gp->port->togate, j = 0;
           (fanout != 0) && (j < MAX_XOR); fanout = fanout->next)
      {
         struct gate *g = fanout->gate;
         if (g->flags & GF_REMOVED)
            continue;
         if ((g->type == sum_type) && (g->incnt == 3) && (g->outcnt == 1))
            xor[i][j++] = g;    /* found a candidate */
      }
   for (i = 0; xor[0][i] != 0; i += 1)
   {
      for (j = 0; (xor[1][j] != xor[0][i]) && (xor[1][j] != 0); j += 1)
         ;
      if (xor[1][j] == 0)
         continue;
      for (k = 0; (xor[2][k] != xor[0][i]) && (xor[2][k] != 0); k += 1)
         ;
      if (xor[2][k] == 0)
         continue;
      *fa_sum = xor[2][k];
      return 1;
   }
   return 0;
}


/* Check to see that all 2 inputs of the FA_CARRY gate are also connected
 * to a 2-input XOR gate. If so, we consider the pair {FA_CARRY,FA_SUM} to
 * form a half-adder.
 */
static int is_half_adder(struct gate *fa_carry,
                         struct gate **fa_sum,
                         enum adder_type adder_type)
{
   struct gateport *gp;
   struct sensitivity *fanout;
   struct gate *xor[2][MAX_XOR+1];
   int i, j;
   if ((fa_carry->incnt != 2) || (fa_carry->outcnt != 1)
       || ((adder_type == A_add) && (fa_carry->type != AND))
       || ((adder_type == A_sub) && (fa_carry->type != SUBC))
       || (fa_carry->flags & GF_REMOVED))
      return 0;
   for (i = 0; i <= 1; i += 1)
      for (j = 0; j <= MAX_XOR; j += 1)
         xor[i][j] = 0;
   for (i = 0, gp = fa_carry->in; gp != 0; i += 1, gp = gp->next)
      for (fanout = gp->port->togate, j = 0;
           (fanout != 0) && (j < MAX_XOR); fanout = fanout->next)
      {
         struct gate *g = fanout->gate;
         if (g->flags & GF_REMOVED)
            continue;
         if ((g->type == XOR) && (g->incnt == 2) && (g->outcnt == 1))
            xor[i][j++] = g;    /* found a candidate */
      }
   for (i = 0; xor[0][i] != 0; i += 1)
   {
      for (j = 0; (xor[1][j] != xor[0][i]) && (xor[1][j] != 0); j += 1)
         ;
      if (xor[1][j] == 0)
         continue;
      *fa_sum = xor[1][j];
      return 1;
   }
   return 0;
}

/* count how many gates are reading from GP
 */
static int count_fanout(struct gateport *gp)
{
   int n = 0;
   struct symbol *sym = gp->port;
   struct sensitivity *fanout;
   if (sym == 0) return 0;
   for (fanout = sym->togate; fanout != 0; fanout = fanout->next)
      if (!(fanout->gate->flags & GF_REMOVED))
         n += 1;
   return n;
}

/* Scan the netlist and mark all gates that are not dangling. Note that after 
 * marking a gate as dangling, the gates on its fanin-list are checked to see 
 * of they now have also become dangling, etc.   
 */
void cleanup_netlist(struct gate **sortedgates,
                     int numgates,
                     struct port *ports)
{
   struct gate *g;
   struct gateport *gp;
   struct port *p;
   int i, do_sweep, n = 0, m = 0, symf;
   int new_numgates, model_numgates;
   struct symbol *sym;
   if ((sym = symfind("$0", 0)) != 0)
      sym->flags |= SYMF_CONST_0; /* mark as constant 0 */
   if ((sym = symfind("$1", 0)) != 0)
      sym->flags |= SYMF_CONST_1; /* mark as constant 1 */
   /* initialize all referenced nets with the SYMF_REFERENCED flag */
   for (p = ports; p != 0; p = p->next)
      if (p->type == 1)
      {  /* OUT port */
         p->port->flags |= SYMF_OUTPUT;
         p->port->flags |= SYMF_REFERENCED; /* OUT ports are always ref'ed */
      }
      else
         p->port->flags |= SYMF_INPUT;
   for (i = 0; i < numgates; i += 1)
   {
      g = sortedgates[i];
      if (!no_const_prop)
         if (is_const(g->out->port))
            propagate_const(g->out->port);
         else if (is_stuck(g, &symf))
         {  /* mark OUT port as constant */
            g->out->port->flags |= symf;
            propagate_const(g->out->port);
         }
      for (gp = g->in; gp != 0; gp = gp->next)
      {
         gp->port->flags |= SYMF_REFERENCED;
         gp->port->alias = 0;
         gp->port->back = 0;
      }
   }
   /* now remove those gates that have unreferenced outputs.  This is done by
    * performing sweeps though the list of gates.  As long as a sweep found
    * at least one gate that had to be removed, we continue sweeping.
    */
   for (n = 0, do_sweep = 1; do_sweep; )
   {
      for (do_sweep = 0, i = 0; i < numgates; i += 1)
      {
         g = sortedgates[i];
         if (g->flags & (GF_REMOVED | GF_MODEL_USE))
            continue;
         if (no_const_prop || !is_const(g->out->port))
            if (g->out->port->flags & SYMF_REFERENCED)
               continue;
         if (!remove_latches && is_latch(g))
         {
            g->out->port->flags |= SYMF_REFERENCED;
            continue;
         }
         do_sweep = 1;          /* removed at least one gate in this sweep */
         remove_gate(g, &n);
      }
   }
   /* Now that all constants have been propagated as far as possible, many
    * opportunities for strength-reduction have been created. For example,
    * the function XOR(y, $1, a) can be replaced with NOT(y, a), and the
    * function AND(y, $1, a) can be replaced with the alias y = a.
    * Strength reduction is a single-pass operation because we don't change
    * the property of the gate outputs (they remain variable signals) and
    * we don't discard non-constant input signals (strength-reduction only
    * discards constant input signals). Therefore, there is nothing to
    * propagate, and a single pass is sufficient to do all the work.
    */
   if (!no_strength_reduce)
      for (m = 0, i = 0; i < numgates; i += 1)
      {
         g = sortedgates[i];
         if (g->flags & (GF_REMOVED | GF_MODEL_USE))
            continue;
         reduce_strength(g, &m);
      }
   /* Done.
    */
   fprintf(stderr,"-- INFO: removed %1d unused gates out of %1d (%1.1f%%)\n",
           n, numgates, 100.0*(double)n/numgates);
   if (!no_strength_reduce)
      fprintf(stderr,"-- INFO: strength-reduced %1d gates out of %1d (%1.1f%%)\n",
              m, numgates, 100.0*(double)m/numgates);
   for (new_numgates = 0, model_numgates = 0, i = 0; i < numgates; i += 1)
   {
      g = sortedgates[i];
      if (g->flags & GF_REMOVED)
         continue;
      new_numgates += 1;
      if (g->flags & GF_MODEL_USE)
         model_numgates += 1;
   }
   fprintf(stderr,"-- INFO: optimized gate count is %1d (%1.1f%% overall reduction)\n",
           new_numgates, 100.0*(1.0 - (double)new_numgates/numgates));
   if(model_numgates)
   fprintf(stderr,"-- INFO: new model gate count is %1d (%1.1f%% of optimized gate count)\n",
           model_numgates, 100.0*(double)model_numgates/new_numgates);
}


/* return TRUE if the output of gate G is stuck to '1' or '0', regardless
 * of its inputs. This does not cover all imaginable cases, but it comes
 * close to catching most common cases. And it is easy to add rules for
 * more obscure cases if you happen to run into them and you think it is
 * worth the trouble to add 'm.
 */
static int is_stuck(struct gate *g, int *symf)
{
   struct gateport *gp;
   int t = g->type, v = 1;;
   if ((t == AND) || (t == NAND))
   {
      for (gp = g->in; gp != 0; gp = gp->next)
         if (gp->port->flags & SYMF_CONST_0)
         {  /* at least one input of AND function is constant 0 */
            *symf = (t == AND) ? SYMF_CONST_0 : SYMF_CONST_1;
            return 1;
         }
         else if (!(gp->port->flags & SYMF_CONST_1))
            v = 0;
      if (v)
      {  /* AND gate with all inputs tied to 1 */
         *symf = (t == AND) ? SYMF_CONST_1 : SYMF_CONST_0;
         return 1;
      }
   }
   else if ((t == OR) || (t == NOR))
   {
      for (gp = g->in; gp != 0; gp = gp->next)
         if (gp->port->flags & SYMF_CONST_1)
         {  /* at least one input of OR function is constant 1 */
            *symf = (t == OR) ? SYMF_CONST_1 : SYMF_CONST_0;
            return 1;
         }
         else if (!(gp->port->flags & SYMF_CONST_0))
            v = 0;
      if (v)
      {  /* OR gate with all inputs tied to 0 */
         *symf = (t == OR) ? SYMF_CONST_0 : SYMF_CONST_1;
         return 1;
      }
   }
   else if (((t == NL_GATE) || (t == NOT)) && (g->incnt == 1))
   {
      int f = g->in->port->flags;
      if (f & SYMF_CONST_0)
         *symf = (t == NL_GATE) ? SYMF_CONST_0 : SYMF_CONST_1;
      else if (f & SYMF_CONST_1)
         *symf = (t == NL_GATE) ? SYMF_CONST_1 : SYMF_CONST_0;
      else
         return 0;              /* input not constant */
      return 1;
   }
   else if (t == MUX)
   {
      int fs = g->in->port->flags & (SYMF_CONST_0 | SYMF_CONST_1);
      int fa = g->in->next->port->flags & (SYMF_CONST_0 | SYMF_CONST_1);
      int fb = g->in->next->next->port->flags & (SYMF_CONST_0 | SYMF_CONST_1);
      if ((fa != 0) && (fa == fb))
      {  /* both inputs of the MUX connect to the same constant */
         *symf = fa;
         return 1;
      }
      else if ((fs == SYMF_CONST_1) && (fa != 0))
      {  /* select input is stuck at 1, and a input is constant */
         *symf = fa;
         return 1;
      }
      else if ((fs == SYMF_CONST_0) && (fb != 0))
      {  /* select input is stuck at 0, and b input is constant */
         *symf = fb;
         return 1;
      }
   }
   else if (t == ADDC)
   {
      int a0, b0, c0, n_zero = 0;
      int a1, b1, c1, n_one = 0;
      struct gateport *pa, *pb, *pc;
      int incnt = g->incnt;
      if (incnt < 2) return 0;
      pa = g->in;
      pb = g->in->next;
      pc = g->in->next->next;
      a0 = (pa->port->flags & SYMF_CONST_0);
      b0 = (pb->port->flags & SYMF_CONST_0);
      c0 = (pc == 0) ? 1 : (pc->port->flags & SYMF_CONST_0);
      if (a0) n_zero += 1;
      if (b0) n_zero += 1;
      if (c0) n_zero += 1;
      if (n_zero >= 2)
      {                         /* two or more zero's force y = 0 */
         *symf = SYMF_CONST_0;
         return 1;
      }
      a1 = (pa->port->flags & SYMF_CONST_1);
      b1 = (pb->port->flags & SYMF_CONST_1);
      c1 = (pc == 0) ? 0 : (pc->port->flags & SYMF_CONST_1);
      if (a1) n_one += 1;
      if (b1) n_one += 1;
      if (c1) n_one += 1;
      if (n_one >= 2)
      {                         /* two or more one's force y = 1 */
         *symf = SYMF_CONST_1;
         return 1;
      }
   }
   else if (t == SUBC)
   {
      int a0, b0, c0, n_zero = 0;
      int a1, b1, c1, n_one = 0;
      struct gateport *pa, *pb, *pc;
      int incnt = g->incnt;
      if (incnt == 2)
	{
	      pa = g->in;
	      pb = g->in->next;
	      a0 = (pa->port->flags & SYMF_CONST_0);
	      b0 = (pb->port->flags & SYMF_CONST_1);
	      if (a0) n_zero += 1;
	      if (b0) n_zero += 1;
	      if (n_zero == 2)
	      {                         /* two zero's force y = 0 */
	         *symf = SYMF_CONST_0;
	         return 1;
	      }
	      a1 = (pa->port->flags & SYMF_CONST_1);
	      b1 = (pb->port->flags & SYMF_CONST_0);
	      if (a1) n_one += 1;
	      if (b1) n_one += 1;
	      if (n_one >= 1)
	      {                         /* one or more one's force y = 1 */
	         *symf = SYMF_CONST_1;
	         return 1;
	      }
	}
	else if(incnt == 3)
	{
	      pa = g->in;
	      pb = g->in->next;
	      pc = g->in->next->next;
	      a0 = (pa->port->flags & SYMF_CONST_0);
	      b0 = (pb->port->flags & SYMF_CONST_1); 	/* ab* + ac + b*c */
	      c0 = (pc->port->flags & SYMF_CONST_0);
	      if (a0) n_zero += 1;
	      if (b0) n_zero += 1;
	      if (c0) n_zero += 1;
	      if (n_zero >= 2)
	      {                         /* two or more zero's force y = 0 */
	         *symf = SYMF_CONST_0;
	         return 1;
	      }
	      a1 = (pa->port->flags & SYMF_CONST_1);
	      b1 = (pb->port->flags & SYMF_CONST_0);	/* ab* + ac + b*c */
	      c1 = (pc->port->flags & SYMF_CONST_1);
	      if (a1) n_one += 1;
	      if (b1) n_one += 1;
	      if (c1) n_one += 1;
	      if (n_one >= 2)
	      {                         /* two or more one's force y = 1 */
	         *symf = SYMF_CONST_1;
	         return 1;
	      }
	}

   }
   return 0;
}


/* propagate the constant net SYM to the output of all gates that
 * are sensitive to SYM.
 */
static void propagate_const(struct symbol *sym)
{
   struct sensitivity *sens;
   int symf;
   for (sens = sym->togate; sens != 0; sens = sens->next)
      if (is_stuck(sens->gate, &symf))
      {
         struct symbol *outsym = sens->gate->out->port;
         outsym->flags |= symf;
         propagate_const(outsym);
      }

}

/* remove gate G from the netlist and see if this removal triggers
 * the removel of other gates.
 */
static void remove_gate(struct gate *g, int *n)
{
   struct gateport *gp;
   int symf;
   if (g->flags & GF_REMOVED)
      return;                   /* already removed */
   symf = g->out->port->flags;
   g->flags |= GF_REMOVED;      /* mark the gate as `removed' */
   *n += 1;
   /* Decrement the reference count of the
    * nets that connect to this gate's inputs
    */
   for (gp = g->in; gp != 0; gp = gp->next)
   {
      struct sensitivity *sens;
      struct symbol *net = gp->port;
      net->incount -= 1;    /* decrement the reference count */
      if (net->incount < 0)
      {
         fprintf(stderr, "Internal: net->incount < 0\n");
         exit(1);
      }
      if (net->incount > 0)
         continue;
      /* this net has become unreferenced: */
      net->flags &= ~SYMF_REFERENCED;
      for (sens = net->fromgate; sens != 0; sens = sens->next)
         if (sens->gate->flags & GF_REMOVED)
         {
            if (!is_const(sens->gate->out->port))
            {
               fprintf(stderr, "Internal: gate already removed ???\n");
               exit(1);
            }
         }
         else if (sens->gate->flags & GF_MODEL_USE)
            /* cannot remove gate because it is adsorbed by model */
            net->flags |= SYMF_MODEL_OUTPUT;
         else if (!remove_latches && is_latch(sens->gate))
            sens->gate->out->port->flags |= SYMF_REFERENCED;
         else
            remove_gate(sens->gate, n);
   }
}

static void reduce_strength(struct gate *g, int *n_reduced)
{
   int t = g->type;
   int incnt = g->incnt, orig_incnt = g->incnt;
   switch (t)
   {
   case NL_GATE:
      make_alias(g, g->in->port);
      break;
   case OR:
   case AND:
      if ((incnt = reduce_fanin(g, (t == OR) ? SYMF_CONST_0 : SYMF_CONST_1)) == 1)
         make_alias(g, g->in->port);
      break;
   case NOR:
   case NAND:
      if ((incnt = reduce_fanin(g, (t == OR) ? SYMF_CONST_0 : SYMF_CONST_1)) == 1)
         g->type = NOT;         /* change into an inverter */
      break;
   case XOR:
   case XNOR:
      if ((incnt = reduce_fanin(g, SYMF_CONST_0)) == 1) /* remove $0 inputs */
         ;
      else
      {
         int incnt1 = incnt;
         incnt = reduce_fanin(g, SYMF_CONST_1); /* remove $1 inputs */
         if (((incnt1 - incnt) % 2) == 1)
            /* removed an odd number of ones: invert the output */
            g->type = (t == XOR) ? XNOR : XOR;
      }
      if (incnt == 1)
         if (g->type == XOR)
            make_alias(g, g->in->port);
         else
            g->type = NOT;
      break;
   case MUX:
   {
      struct symbol *s, *a, *b;
      if (g->incnt != 3) break;
      s = g->in->port;
      a = g->in->next->port;
      b = g->in->next->next->port;
      /* Mux function: y = (s & a) + (!s & b).
       * The case (a == b == const) is already covered by const propagation.
       */
      if ((a->flags & SYMF_CONST_1) && (b->flags & SYMF_CONST_0)
          && !(s->flags & (SYMF_CONST_0 | SYMF_CONST_1)))
         make_alias(g, s);      /* identity y = s */
      else if (((a->flags & SYMF_CONST_0) && (b->flags & SYMF_CONST_1)))
      {                         /* negation y = !s */
         g->in->next = 0;       /* remove a and b from the port list */
         g->incnt = 1;
         g->type = NOT;
      }
      else if (a->flags & SYMF_CONST_1) /* y = s + (!s & b) = s + b */
      {
         g->in->next = g->in->next->next; /* remove a from port list */
         g->incnt = 2;
         g->type = OR;
      }
      /* the case a = 0 reduces to (y = !s & b) but we have no gate for that.
       * Likewise, b = 1 reduces to (y = !s + a) but there is no IVF gate for
       * that. In both cases, the backend function may detect such MUXes as
       * special cases and express them in there reduced form (see abel.c for
       * an example of this)
       */
   }
   break;
   case ADDC:                   /* y = (a & b) + (a & c) + (b & c) */
   case SUBC:                   /* y = (a & !b) + (a & c) + (!b & c) */
   {
      int a0, b0, c0, n_zero = 0;
      int a1, b1, c1, n_one = 0;
      struct gateport *pa, *pb, *pc;
      int incnt = g->incnt;
      if (incnt < 2) break;
      pa = g->in;
      pb = g->in->next;
      pc = g->in->next->next;
      a0 = (pa->port->flags & SYMF_CONST_0);
      b0 = (pb->port->flags) & (t == ADDC ? SYMF_CONST_0 : SYMF_CONST_1);
      c0 = (pc == 0) ? 1 : (pc->port->flags & SYMF_CONST_0);
      if (a0) n_zero += 1;
      if (b0) n_zero += 1;
      if (c0) n_zero += 1;
      a1 = (pa->port->flags & SYMF_CONST_1);
      b1 = (pb->port->flags) & (t == ADDC ? SYMF_CONST_1 : SYMF_CONST_0);
      c1 = (pc == 0) ? 0 : (pc->port->flags & SYMF_CONST_1);
      if (a1) n_one += 1;
      if (b1) n_one += 1;
      if (c1) n_one += 1;
      /* the case where we have >= 2 zero's or >= 2 one's is already
       * handled by the constant propagation functionality
       */
      if ((n_zero == 1) && (n_one == 1))
      {  /* one input $0, one input $1, one input variable */
         if (!is_const(pa->port))
            make_alias(g, pa->port); /* y = a */
         else if (!is_const(pb->port))
            if (t == ADDC)
               make_alias(g, pb->port); /* y = b */
            else
            {                   /* y = !b */
               g->type = NOT;
               g->incnt = 1;
               g->in = pb;
               pb->next = 0;
            }
         else if ((pc != 0) && !is_const(pc->port))
            make_alias(g, pc->port); /* y = c */
      }
      else if ((n_zero == 1) && (n_one == 0))
      {  /* one input 0, two inputs variable */
         /* Note that AND(x,y) => x & y, while SUBC(x,y) => x & !y
          */
         g->type = ((t == SUBC) && (a0 || c0)) ? SUBC : AND;
         g->incnt = 2;
         if (a0)
            if (t == SUBC)
            {                   /* y = !b & c */
               g->in = pc;      /* remove a, swap b and c */
               pc->next = pb;
               pb->next = 0;
            }
            else
               g->in = pb;      /* remove input a */
         else if (b0)
            pa->next = pc;      /* remove input b */
         else if (c0)
            pb->next = 0;       /* remove input c */
      }
      else if ((n_zero == 0) && (n_one == 1) && (t == ADDC))
      {
         g->type = OR;
         g->incnt = 2;
         if (a1)
            g->in = pb;         /* remove input a */
         else if (b1)
            pa->next = pc;      /* remove input b */
         else if (c1)
            pb->next = 0;       /* remove input c */
      }
      else if ((n_zero == 0) && (n_one == 1) && (t == SUBC) && b1)
      {                         /* y = a + c */
         g->type = OR;
         g->incnt = 2;
         pa->next = pc;
      }
   }
   break;
   } /* end switch (t) */
   /*
    * keep the statistics ...
    */
   if ((incnt != orig_incnt) || (g->type != t) || (g->flags & GF_REMOVED))
      *n_reduced += 1;
}

/* Remove gate G and create an alias from SYM to G's output port.
 */
static void make_alias(struct gate *g, struct symbol *sym)
{
struct port *op=g->out->port->isport;

if((op)&&(op->type==1)&&(sym->isport)&&(sym->isport->type==0)) return; 
/* aliasing an outport to an inport causes trouble in csim */

   if (is_const(g->out->port) || is_const(sym))
   {
      fprintf(stderr, "Internal: alias to a const ????\n");
      exit(1);
   }
   g->flags |= GF_REMOVED;
   merge_aliases(g->out->port, sym);
}


/* Make symbols S1 and S2 part of the same alias group.
 */
static struct alias *merge_aliases(struct symbol *s1, struct symbol *s2)
{
   struct alias *a1 = s1->alias, *a2 = s2->alias;
   struct alias *a, *a_port = 0, *a_base = 0;
   struct symbol *base;
   if (a1 == 0)
   {
      a1 = calloc(1, sizeof(struct alias));
      a1->sym = s1;
   }
   if (a2 == 0)
   {
      a2 = calloc(1, sizeof(struct alias));
      a2->sym = s2;
   }
   for (a = a1; a->next != 0; a = a->next)
      ;
   a->next = a2;                /* chain a2 to the end of a1 */
   s1->alias = s2->alias = a1;
   /*
    * now select a new base symbol for the aliases
    */
   for (a = a1; a != 0; a = a->next)
   {
      const char *nm = a->sym->id ? a->sym->id : a->sym->name;
      if (a->sym->flags & (SYMF_INPUT | SYMF_OUTPUT))
         a_port = a;
      else if ((a_base == 0) && (strchr(nm, '!') == 0))
         a_base = a;            /* first non-temp name is new base symbol */
   }
   if (a_port != 0)             /* port names are most wanted */
      base = a_port->sym;
   else if (a_base != 0)        /* non-temp names are next */
      base = a_base->sym;
   else
      base = a1->sym;           /* no preferred base symbol: just take a1 */
   for (a = a1; a != 0; a = a->next)
   {
      a->sym->alias = a1;
      a->sym->back = base;      /* make new base symbol known to all aliases */
   }
   return a1;
}


static int reduce_fanin(struct gate *g, int symf)
{
   struct gateport *gp, *gp_prev = 0;
   int num_to_remove = 0, n = 0;
   for (gp = g->in; gp != 0; gp = gp->next)
      if (gp->port->flags & symf)
         num_to_remove += 1;
   if ((g->incnt - num_to_remove) <= 1)
      num_to_remove = g->incnt - 1; /* always leave at least one input */
   n = 0;
   for (gp = g->in; n < num_to_remove; gp = gp->next)
      if (gp->port->flags & symf)
      {  /* unlink input pin from gate */
         if (gp_prev == 0)
            g->in = gp->next;
         else
            gp_prev->next = gp->next;
         n += 1;
      }
      else
         gp_prev = gp;
   g->incnt -= n;
   return g->incnt;
}
