/* 
 * Copyright (c) Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include "cyco.h"

extern int compressed_ids;      /* this model uses compressed ids */
extern int no_tidy;             /* command line option */
extern int tidy_before;         /* command line option */
extern int alias_disable;       /* command line option */

static int ivf_gcnt = 1;

static void ivf_print_com(const char* operator, struct gate *gate);
static void ivf_print_nlgate(struct symbol *net);
static void ivf_print_nlgate_reversed(struct symbol *net);
static void ivf_print_err(const char *gate_name);
static const char *ivf_gp_name(struct gateport *gp);
static const char *ivf_sym_name(struct symbol *s);


/* Print IVF model to stdout
 */
void ivfstate(const char *modelname,
              int numgates,
              struct gate **sortedgates,
              struct port *ports,
              struct symbol *symchain)
{
   struct gate *g;
   struct gateport *gp;
   int i;
   struct symbol *s;
   struct port *p, *ptemp, *porthead=NULL;
   struct adder *adder, *adder_list = 0;

   if (!no_tidy && tidy_before)
      cleanup_netlist(sortedgates, numgates, ports);

   /* inspect all symbols and find out if they are part of a vector: */
   for (s = symchain; s != 0; s = s->nextinsymchain)
   {
      s->alias = 0;
      s->back = 0;
      extract_vector_range(s);
   }

   /* make a pass over all gates to see what sigs are being used */
   for (i = 0; i < numgates; i += 1)
   {
      g = sortedgates[i];
      if (g->flags & (GF_MODEL_USE | GF_REMOVED))
         continue;
      /* first mark all signal names */
      if ((g->type == NL_GATE) && (g->incnt == 1)
          && (g->outcnt == 1) && (g->in->port->name[0] == '$'))
      {                         /* this is a constant ... */
         if (g->in->port->name[1] == '0')
            g->out->port->flags |= SYMF_CONST_0;
         else if (g->in->port->name[1] == '1')
            g->out->port->flags |= SYMF_CONST_1;
         else
            g->out->port->flags |= SYMF_CONST_X;
      }
      else
      {
         for (gp = g->in; gp != 0; gp = gp->next)
            gp->port->flags |= SYMF_USED;
         for (gp = g->out; gp != 0; gp = gp->next)
            gp->port->flags |= SYMF_USED;
      }
      if ((adder = extract_adder(g)) != 0)
      {  /* this gate is part of an adder structure */
         int i;
         const char *op = (adder->type == A_add ? "adder" : "subtractor");
         for (i = 0; i < adder->size; i += 1)
         {
            struct gate *g_carry = adder->vec[i].g_carry;
            /* internal carry chain disappears from netlist: */
            if (g_carry != 0)
            {
               g_carry->out->port->flags |= SYMF_MODEL_USE;
               g_carry->flags |= GF_MODEL_USE;
            }
            adder->vec[i].g_sum->flags |= GF_MODEL_USE;
         }
         fprintf(stderr,
                 "-- INFO: extracted %1d-bit %s, msb of sum is %s\n",
                 adder->size, op, adder->name);
         adder->next = adder_list;
         adder_list = adder;
      }
   }

   if (!no_tidy && !tidy_before)
      cleanup_netlist(sortedgates, numgates, ports);
   
   /* ok, let's go */
   printf("MX %s{\n", modelname);

   p=ports;
   while(p)        /* invert ports to normal order */
        {  
        ptemp=(struct port *)malloc(sizeof(struct port));
        ptemp->port=p->port;
        ptemp->type=p->type;
        ptemp->next=porthead;
        porthead=ptemp;
        p=p->next;
        }
 
   if (!compressed_ids)
   {
      printf("MP{\n");
      for (p = porthead; p != 0; p = p->next)
      {
         struct symbol *s = p->port;
         s->flags |= (p->type == 1) ? SYMF_OUTPUT : SYMF_INPUT;
         if (p->type == 0)         /* input */
            printf("%s IN;", s->name);
         else                      /* output */
            printf("%s OUT;", s->name);
      }
      printf("}\n");               /* end of MP block */
   }
   else
   {
      printf("MIP{\n");
      for (p = porthead; p != 0; p = p->next)
      {
         struct symbol *s = p->port;
         s->flags |= (p->type == 1) ? SYMF_OUTPUT : SYMF_INPUT;
         if (p->type == 0)         /* input */
            printf("%s %s IN;", s->name, s->id);
         else                      /* output */
            printf("%s %s OUT;", s->name, s->id);
      }
      printf("}\n");               /* end of MP block */
   }

   printf("}\n");               /* end of MX block */

   printf("\nMF{\n");

   for (s = symchain; s != 0; s = s->nextinsymchain)
      if ((s->flags & SYMF_USED)
          && (s->flags & (SYMF_REFERENCED | SYMF_MODEL_OUTPUT))
          && !is_const(s))
         if (compressed_ids)
         {
            if ((s->id == 0) && (s->name[0] == '$'))
               continue;           /* happens in optimized netlists */
            if (!s->alias || alias_disable)
               printf("FIS{%s %s}\n", s->name, s->id);
            else if (s->back == s)
            {
               struct alias *alias;
               printf("FISA{%s %s", s->back->name, s->back->id);
               for (alias = s->alias; alias != 0; alias = alias->next)
                  if (alias->sym != alias->sym->back)
                     printf(" %s %s", alias->sym->name, alias->sym->id);
               printf("}\n");
            }
         }
         else
         {
            if ((s->id == 0) && (s->name[0] == '$'))
               continue;           /* happens in optimized netlists */
            if (!s->alias || alias_disable)
               printf("FS{%s}\n", ivf_sym_name(s));
            else if (s->back == s)
            {
               struct alias *alias;
               printf("FSA{%s", s->back->name);
               for (alias = s->alias; alias != 0; alias = alias->next)
                  if (alias->sym != alias->sym->back)
                     printf(" %s", alias->sym->name);
               printf("}\n");
            }
         }

   printf("\n");                /* end of wire declarations */

   for (i = 0; i < numgates; i += 1)
   {
      g = sortedgates[i];
      if (g->flags & (GF_REMOVED))
         continue;
      switch (g->type)
      {
      case NL_GATE:      ivf_print_com("NL_GATE", g); break;
      case NL_OPTIMIZED: ivf_print_err("nl_optimized"); break;
      case NOT:          ivf_print_com("NOT", g); break;
      case AND:          ivf_print_com("AND", g); break;
      case OR:           ivf_print_com("OR", g); break;
      case XOR:          ivf_print_com("XOR", g); break;
      case NAND:         ivf_print_com("NAND", g); break;
      case NOR:          ivf_print_com("NOR", g); break;
      case XNOR:         ivf_print_com("XNOR", g); break;
      case LATCH:        ivf_print_com("LATCH", g); break;
      case LATCH_SR:     ivf_print_com("LATCH_SR", g); break;
      case LATCH_N:      ivf_print_com("LATCH_N", g); break;
      case LATCH_SR_N:   ivf_print_com("LATCH_SR_N", g); break;
      case NEGC:         ivf_print_com("NEGC", g); break;
      case ADDC:         ivf_print_com("ADDC", g); break;
      case SUBC:         ivf_print_com("SUBC", g); break;
      case BUF:          ivf_print_com("BUF", g); break;
      case MUX:          ivf_print_com("MUX", g); break;
      case BUFIF0:       ivf_print_com("BUFIF0", g); break;
      case PMOS:         ivf_print_com("PMOS", g); break;
      case BUFIF1:       ivf_print_com("BUFIF1", g); break;
      case NMOS:         ivf_print_com("NMOS", g); break;
      case NOTIF0:       ivf_print_com("NOTIF0", g); break;
      case NOTIF1:       ivf_print_com("NOTIF1", g); break;
      case WIRE:         ivf_print_com("WIRE", 0); break;
      case TRI:          ivf_print_com("TRI", g); break;
      case WAND:         ivf_print_com("WAND", g); break;
      case TRIAND:       ivf_print_com("TRIAND", g); break;
      case WOR:          ivf_print_com("WOR", g); break;
      case TRIOR:        ivf_print_com("TRIOR", g); break;
      case TRI0:         ivf_print_com("TRI0", g); break;
      case TRI1:         ivf_print_com("TRI1", g); break;
      case SUPPLY0:      ivf_print_com("SUPPLY0", g); break;
      case SUPPLY1:      ivf_print_com("SUPPLY1", g); break;
      default:           ivf_print_err("<UNKNOWN>"); break;
      }
   }   

   /* OUT ports that are constant or aliased need explicit assignment */
   for (p = porthead; p != 0; p = p->next)
      if (p->port->flags & SYMF_OUTPUT)
      {
         if ((p->port->flags & (SYMF_CONST_0 | SYMF_CONST_1))
             || ((p->port->alias != 0) && (p->port->back != p->port)))
            ivf_print_nlgate(p->port);
      }
      else if ((p->port->flags & SYMF_INPUT)
               && ((p->port->alias != 0) && (p->port->back != p->port)))
         /* the circuit refers to inport by its alias ... We need to
          * make an explicit connection between inport and alias: */
         ivf_print_nlgate_reversed(p->port);

   printf("}\n");               /* end of MF block */
}


/* print gate */
static void ivf_print_com(const char* operator,
                          struct gate *gate)
{
   struct gateport *gp;

if(gate)
   {
   gp=gate->in;
   if (is_const(gate->out->port))
      return;
   printf("FG %s.%1d{%s 0 I %1d ",
          operator, ivf_gcnt++, operator, gate->incnt);
   for (; gp != 0; gp = gp->next)
      printf("%s ", ivf_gp_name(gp));
   printf("%1d ", gate->outcnt);
   for (gp = gate->out; gp->next != 0; gp = gp->next)
      printf("%s ", ivf_gp_name(gp));
   printf("%s}\n", ivf_gp_name(gp));
   }
   else
   {
   /* gate is missing here.. */
   }
}

static void ivf_print_nlgate(struct symbol *net)
{
   printf("FG NL_GATE.%1d{NL_GATE 0 I 1 %s 1 %s}\n",
          ivf_gcnt++, ivf_sym_name(net), net->name);
}

static void ivf_print_nlgate_reversed(struct symbol *net)
{
   printf("FG NL_GATE.%1d{NL_GATE 0 I 1 %s 1 %s}\n",
          ivf_gcnt++, net->name, ivf_sym_name(net));
}

static void ivf_print_err(const char *gate_name)
{
   fprintf(stderr, "** cannot handle gate type %s\n", gate_name);
   exit(1);
}


static const char *ivf_gp_name(struct gateport *gp)
{
   return ivf_sym_name(gp->port);
}

static const char *ivf_sym_name(struct symbol *sym)
{
   const char *name = sym->name;
   if (sym->flags & SYMF_CONST_0)
      name = "$0";
   else if (sym->flags & SYMF_CONST_1)
      name = "$1";
   else if (sym->alias)
      name = sym->back->name;
   return name;
}
