/*
 * $Header: /home/paul/cvsroot/ver/behave.c,v 1.38 1999/12/19 04:10:31 paul Exp $
 *
 * Copyright (c) Paul Stravers 1999.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */


#include "symbol.h"
#include "behave.h"
#include "y.tab.h"

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

/* muxtree is a private data type and therefore does not belong in "behave.h":
 */
struct muxtree
{
   struct stmt *stmt;           /* statement from where this node originates */
   struct symslot *outsig;      /* signal name of this node in the tree */
   struct bitvec *sel;          /* the select input of the mux */
   struct muxtree *t, *f;       /* true and false branches of the mux */
   struct muxtree *choices;     /* choice list of a case switch */
   struct muxtree *null_choice; /* null_choice of incomplete CASE stmt */
   struct muxtree *next;        /* next choice in a choice list */
   int flags;                   /* see #define MF_* below */
};

#define MF_IS_PARALLEL_CASE    0x01   /* not a priority mux */
#define MF_HAS_DEFAULT_CLAUSE  0x02   /* default case present */
#define MF_IS_COMPLETE         0x04   /* complete assignment */
#define MF_IS_NULL_CHOICE      0x08   /* choice does not assign lvalue */
#define MF_IS_LINKED_TO_SELF   0x10   /* null branches linked to self */
#define MF_REG_VAL_DONE        0x20   /* replace_reg_with_val() done */
#define MF_HAS_NULL_BRANCH     0x40   /* case switch with a null branch */

/* struct async is used to figure out asynchronous set/reset on a flipflop:
 */
struct async
{
   struct symlist *sig_list;
   struct muxtree *set_reset_list;
   struct async   *next;
};

/* prototype declarations of some local functions: */
static void error(const char *msg);
static void merge_assignments(struct regval **to_list,
                              struct regval *from_list);
static struct regval *assigns_to_sym(struct symslot *sym,
                                     struct stmt *statement);
static struct regval *assigns_to(struct regval *lvalue,
                                 struct stmt *statement);
static struct choice *pick_default_choice(struct choice **list);
static struct symslot *generate_choice_guard(struct stmt *this,
                                             struct choice *choice,
                                             struct bitvec *not_switch_expr);
static void generate_default_guard(struct bitvec *guard_vec, int j_dflt);
static struct symslot *generate_inv_gate(struct symslot *gate_input);
static struct muxtree *new_muxtree(struct stmt *stmt);
static int find_duplicate_choice(struct choice *list,
                                 struct choice_selector *choice_sel);
static int bitvec_equivalent(struct bitvec *a,
                             struct bitvec *b);
static int is_latch(struct gate *g);
static int is_async_set_reset(struct bitvec *outvec,
                              struct async *async,
                              int lineno);
static struct stmt *find_previous_stmt(struct stmt *stmt);
static void find_reg(struct symslot **sigptr,
                     struct stmt *stmt, int *loop);
static struct symslot *find_value(struct symslot *sig,
                                  struct stmt *stmt, int *loop);
static void replace_reg_with_val(struct muxtree *mt, int *loop);
static int is_vector_elmt(struct symslot *vec, struct symslot *elmt);
static void generate_gate_and_leave_in_bitvec(int type,
                                              struct bitvec *gate_inputs,
                                              int j_dest);
static struct choice *find_default_choice(struct choice *list);


/*
 * Create a struct stmt:
 */
struct stmt *new_stmt(enum stmt_type type, int lineno)
{
   /* calloc() initializes the structure to all zeros: */
   struct stmt *s = calloc(1, sizeof(struct stmt));
   s->type = type;
   s->lineno = lineno;
   return s;
}

/*
 * Assign RVAL to the register LVAL in the range LLORANGE to LHIRANGE.
 */
struct stmt *stmt_assign_expr(struct stmt *this,
                              int llorange, int lhirange,
                              struct reg *lval,
                              struct bitvec *rval)
{
   int j;
   int no_mesg = FALSE;
   struct assign_reg *assign_reg = 0;
   if (this->type != Stmt_Assign)
      error("Internal: Stmt_Assign expected");
   for (j = llorange; j <= lhirange; j += 1)
   {
      char s[200];
      struct symslot *sym;
      struct regval *lv;
      sprintf(s, "%s/%1d", lval->port->name, j);
      sym = symfind(s);
      if (sym == 0)
         error("Internal: cannot find symbol with bitnum");
      if (((sym->flags & SF_NON_BLOCKING)
           && !(this->flags & Is_Non_Blocking))
          || ((sym->flags & SF_BLOCKING)
              && (this->flags & Is_Non_Blocking)))
         if (!no_mesg)
         {
            mesg("** cannot mix blocking and non-blocking assignments to reg %s, in line %1d of file '%s'\n",
                 lval->port->name, this->lineno, yyfilename);
            no_mesg = TRUE;
         }
      if (this->flags & Is_Non_Blocking)
         sym->flags |= SF_NON_BLOCKING;
      else
         sym->flags |= SF_BLOCKING;
      lv = calloc(1, sizeof(struct reg));
      lv->symbol = sym;
      lv->next = this->lvalues;
      this->lvalues = lv;
   }
   assign_reg = calloc(1, sizeof(struct assign_reg));
   assign_reg->next = this->t.assign.reg_list;
   this->t.assign.reg_list = assign_reg;
   assign_reg->reg = lval->port;
   assign_reg->expr = rval;
   assign_reg->hi_range_offset = lval->port->hirange - lhirange;
   if (lval->port->dir == -1)
   {  /* mirror the rval vector */
      int i, len = rval->length;
      struct symslot *t;
      for (i = 0; i < len/2; i += 1)
      {
         t = rval->bits[i];
         rval->bits[i] = rval->bits[len - i - 1];
         rval->bits[len - i - 1] = t;
      }
   }
   return this;
}

/*
 * append SUBSTMT to the end of the GROUP. Note that SUBSTMT
 * can also be a list of sub-statements.
 */
struct stmt *stmt_group_append(struct stmt *this,
                               struct stmt *substmt)
{
   struct stmt *sub, *last_sub = 0;
   if (this->type != Stmt_Group)
      error("Internal: Stmt_Group expected");
   for (sub = substmt; sub != 0; sub = sub->next)
   {
      if (sub->parent != 0)
         error("Internal: empty parent expected");
      sub->parent = this;
      last_sub = sub;
   }
   if (this->t.group.first == 0)
   {
      this->t.group.first = substmt;
      this->t.group.last = last_sub;
   }
   else
   {
      this->t.group.last->next = substmt;
      substmt->prev = this->t.group.last;
      this->t.group.last = last_sub;
   }
   /* merge the sub-statements lvalues into the group's lvalues */
   for (sub = substmt; sub != 0; sub = sub->next)
      merge_assignments(&this->lvalues, sub->lvalues);
   return this;
}

/*
 * add CONDITION and TRUESTMT to the branch
 */
struct stmt *stmt_branch_true(struct stmt *this,
                              struct bitvec *condition,
                              struct stmt *truestmt)
{
   if (this->type != Stmt_Branch)
      error("Internal: Stmt_Branch expected");
   if (this->t.branch.if_true != 0)
      error("Internal: empty if_true branch expected");
   if (this->t.branch.cond != 0)
      error("Internal: empty branch condition expected");
   if (truestmt->parent != 0)
      error("Internal: empty parent expected");
   truestmt->parent = this;
   this->t.branch.cond = condition;
   this->t.branch.if_true = truestmt;
   merge_assignments(&this->lvalues, truestmt->lvalues);
   return this;
}

/*
 * add ELSESTMT to the branch
 */
struct stmt *stmt_branch_false(struct stmt *this,
                               struct stmt *elsestmt)
{
   if (this->type != Stmt_Branch)
      error("Internal: Stmt_Branch expected");
   if (this->t.branch.if_false != 0)
      error("Internal: empty if_false branch expected");
   this->t.branch.if_false = elsestmt;
   if (elsestmt != 0)
   {
      if (elsestmt->parent != 0)
         error("Internal: empty parent expected");
      elsestmt->parent = this;
      merge_assignments(&this->lvalues, elsestmt->lvalues);
   }
   return this;
}

/*
 * Scan the choices in THIS case statement, and compute for each choice
 * a choice guard. Put all the choice guards in this->t.choice.guards:
 */
static void stmt_case_compute_guards(struct stmt *this)
{
   struct choice *choice;
   int num_guards = 0, j = 0, j_dflt = -1;
   int expr_length = this->t.choice.expr->length;
   struct bitvec *not_switch_expr = new_bitvec(expr_length); 
   struct bitvec *guard_vec;
   struct choice *deflt = this->t.choice.deflt;
   /* count the number of choices we have in this case statement: */
   for (choice = this->t.choice.list; choice != 0; choice = choice->next)
      num_guards += 1;
   if (num_guards == 0)
      error("Internal: num_guards should be at least 1");
   guard_vec = new_bitvec(num_guards);
   this->t.choice.guard = guard_vec;
   for (choice = this->t.choice.list; choice != 0; choice = choice->next)
      if (choice == deflt)
         j_dflt = j++;
      else
         /* create logic that guards this choice, then cache it: */
         guard_vec->bits[j++] =
            generate_choice_guard(this, choice, not_switch_expr);
   free(not_switch_expr);
   if (deflt != 0)
      if (j_dflt < 0)
         error("internal: j_dflt < 0");
      else
         generate_default_guard(guard_vec, j_dflt);
}

/*
 * Initialize a Stmt_Case structure.
 */
struct stmt *stmt_case_init(struct stmt *this,
                            struct bitvec *switch_expr,
                            struct choice *choice_list)
{
   struct choice *choice;
   if (this->type != Stmt_Case)
      error("Internal: Stmt_Case expected");
   /* keep track of the lvalues that we are assigning to */
   for (choice = choice_list; choice != 0; choice = choice->next)
   {
      merge_assignments(&this->lvalues, choice->statement->lvalues);
      choice->statement->parent = this;
   }
   /* Find the "default" choice and remove it from choice_list: */
   this->t.choice.deflt = find_default_choice(choice_list);
   this->t.choice.expr = switch_expr;
   this->t.choice.list = choice_list;
   stmt_case_compute_guards(this);
   return this;
}

/*
 * Create a new struct choice
 */
struct choice *new_choice(struct choice_selector *sel,
                          struct stmt *statement,
                          int lineno)
{
   struct choice *this = calloc(1, sizeof(struct choice));
   this->sel = sel;
   this->statement = statement;
   this->lineno = lineno;
   return this;
}

/*
 * Create a new struct choice_selector
 */
struct choice_selector *new_choice_selector(struct bitvec *const_expr)
{
   struct choice_selector *this = calloc(1, sizeof(struct choice_selector));
   this->const_expr = const_expr;
   return this;
}

/*
 * set the FLAG for this Stmt_Case.
 */
struct stmt *stmt_case_setflag(struct stmt *this,
                               stmt_flags flag)
{
   if (this->type != Stmt_Case)
      error("Internal: Stmt_Case expected");
   this->flags |= flag;
   return this;
}

/*
 * return TRUE if FLAG is set in this Stmt_Case.
 */
int stmt_case_testflag(struct stmt *this,
                       stmt_flags flag)
{
   if (this->type != Stmt_Case)
      error("Internal: Stmt_Case expected");
   return (this->flags & flag) != 0;
}

/*
 * print a message on stderr, then exit.
 */
static void error(const char *msg)
{
   fprintf(stderr, "\n\n*******************************************\n");
   fprintf(stderr, msg);
   fprintf(stderr, "\n*******************************************\n\n");
   exit(1);
}

/*
 * Merge a copy of the FROM_LIST into the TO_LIST without creating
 * duplicates. The FROM_LIST will not be modified.
 */
static void merge_assignments(struct regval **to_list,
                              struct regval *from_list)
{
   /* Actually, TO_LIST and FROM_LIST are symbol tables.  Currently, we
    * implement these symbol tables as linear lists, so we're doomed to use a
    * N-square merging algorithm (slow...).  Need to fix this sometime!
    */
   struct regval *last_in_to_list = 0, *to, *from;

   /* first we need to find the end of the TO_LIST */
   for (to = *to_list; to != 0; to = to->next)
      last_in_to_list = to;

   for (from = from_list; from != 0; from = from->next)
   {
      /* check for duplicate */
      for (to = *to_list; to != 0; to = to->next)
         if (strcmp(from->symbol->name, to->symbol->name) == 0)
            break;
      if (to == 0)
      {  /* no duplicate found, append a shallow copy at the end of TO_LIST */
         struct regval *newreg = calloc(1, sizeof(struct regval));
         newreg->symbol = from->symbol;
         newreg->muxtree = from->muxtree;
         if (last_in_to_list == 0)
            *to_list = newreg;
         else
            last_in_to_list->next = newreg;
         last_in_to_list = newreg;
      }
   }
}

/*
 * find all NULL branches in the ROOT tree, and replace
 * these NULL branches with SUBTREE.
 */
static void link_to_null_branches(struct muxtree *root,
                                  struct muxtree *subtree)
{
if(!root) return; /* may be wrong: fix for cpu6502.v compiler hang */

   if (root->flags & MF_IS_COMPLETE)
      ; /* there are no null branches here */
   else if (root->sel == 0)
      ; /* we've reached a leave: this is not a mux, it is a value */
   else if (root->choices == 0)
   {                            /* this is an incomplete branch */
      if (root->t == 0)
         root->t = subtree;
      else
	 {
	 if(root->t==root) { root->t = NULL; }	/* may be wrong: fix for cpu6502.v compiler hang */
         if(root->t!=subtree) link_to_null_branches(root->t, subtree); /* recurse */
         }
      if (root->f == 0)
         root->f = subtree;
      else
	 {
	 if(root->f==root) { root->f = NULL; } /* may be wrong: fix for cpu6502.v compiler hang */
         if(root->f!=subtree) link_to_null_branches(root->f, subtree); /* recurse */
         }
   }
   else
   {  /* this is an incomplete case switch */
      struct muxtree *choice = root->choices;
      for (; choice != 0; choice = choice->next)
         link_to_null_branches(choice, subtree);
      if (root->flags & MF_HAS_NULL_BRANCH)
         if (root->null_choice == 0)
            root->null_choice = subtree;
         else
            link_to_null_branches(root->null_choice, subtree);
   }
}

/*
 * Create a MUX gate and connect its output to a newly created signal.  Note:
 * this function interfaces with the gate and port creation functions that
 * are part of the original VER code. Since I do not fully understand that
 * code, I suspect this function can be much improved. Please do so!
 */
static struct symslot *create_mux_gate(struct muxtree *mux)
{
   struct symslot *sig, *select_sig;
   struct gateport *outport;
   saveports();
   isgate = 1;
   if (mux->sel->length > 1)
   {                            /* must reduce this to 1 bit using an OR */
      int j;
      gptr = 0;
      select_sig = createtemp();
      outport = gptr;
      gtype = OR;
      addgate();
      gptr = outport;
      for (j = 0; j < mux->sel->length; j += 1)
         addport(mux->sel->bits[j]);
      currentmod->gates->port = gptr;
   }
   else
      select_sig = mux->sel->bits[0];
   gptr = 0;
   sig = createtemp();          /* create unique name for new signal */
   /* the call to createtemp() also created an output port with addport() */
   outport = gptr;
   gtype = MUX;
   addgate();                   /* create the MUX gate */
   gptr = outport;
   addport(select_sig);         /* Add a port for the select input, ... */
   addport(mux->t->outsig);     /* and a port for the sel=1 input ... */
   addport(mux->f->outsig);     /* ..and a port for the sel=0 input . */
   currentmod->gates->port = gptr; /* this is a hack, how to avoid this ? */
   retrports();
   return sig;
}

/*
 * Count how many selectors are ipart of this CHOICE:
 */
static int count_selectors(struct choice *choice)
{
   int n = 0;
   struct choice_selector *sel;
   for (sel = choice->sel; sel != 0; sel = sel->next)
      n += 1;
   return n;
}

/*
 * Scan STMT and translate the procedural statements it contains into a set
 * of concurrent assignments.
 */
static struct muxtree *build_mux_tree(struct stmt *this,
                                      struct regval *lvalue,
                                      struct async *async)
{
   struct muxtree *node = 0;
   struct regval *regv = 0;
   if (this == 0)
      return node;
   switch (this->type)
   {
   case Stmt_Assign:
   {                            /* create a leave in the mux tree */
      int lbitnum = lvalue->symbol->bitnum; /* bitnum of the lvalue */
      int lhirange, offset, rval_index;
      struct assign_reg *lhs = this->t.assign.reg_list;
      for (; lhs != 0; lhs = lhs->next)
         if (is_vector_elmt(lhs->reg, lvalue->symbol))
            break;
      if (lhs == 0)
         error("Internal: build_mux_tree: no assignment found");
      lhirange = lhs->reg->hirange; /* lo-range of the lvalue */
      offset = lhs->hi_range_offset;
      rval_index = lhirange - lbitnum - offset; /* which bit to select */
      node = new_muxtree(this);
      /* select the bit from the rvalue bitvec that corresponds to lvalue.
       * Note that bitvec[0] corresponds to the most significant bit!
       */
      node->outsig = lhs->expr->bits[rval_index];
      node->flags |= MF_IS_COMPLETE; /* assign statement is always complete */
      return node;
   }
   case Stmt_Group:
   {
      /* Later assignments overrule earlier assignments. Therefore we
       * scan the list of sub-statements backwards, stopping as soon as
       * we found a complete assignment.
       */
      struct stmt *seq_stmt = this->t.group.last;
      struct muxtree *ret_node = 0;
      for (; seq_stmt != 0; seq_stmt = seq_stmt->prev)
         if ((regv = assigns_to(lvalue, seq_stmt)) != 0)
         {
            ret_node = build_mux_tree(seq_stmt, lvalue, async);
            regv->muxtree = ret_node;
            break;
         }
      if (ret_node == 0)
         return ret_node;       /* no lvalue assignment found in this group */
      node = ret_node;
      /* We try to make the assignment complete by substituting the NULL
       * branches with assignments that were made earlier in the group.
       */
      for (seq_stmt = seq_stmt->prev; seq_stmt != 0; seq_stmt = seq_stmt->prev)
         if ((regv = assigns_to(lvalue, seq_stmt)) != 0)
         {
            struct muxtree *mt = build_mux_tree(seq_stmt, lvalue, async);
            regv->muxtree = mt;
            if (node->flags & MF_IS_COMPLETE)
               node = mt;
            else
            {
               link_to_null_branches(node, mt); /* link MT under NODE tree */
               if (mt->flags & MF_IS_COMPLETE)
                  /* We just replaced all dangling branches of NODE with
                   * pointers to the MT subtree which is complete. Therefore
                   * now the entire tree is complete.
                   */
                  node->flags |= MF_IS_COMPLETE;
            }
         }
      return ret_node;          /* return last assignment in block */
   }
   case Stmt_Branch:
   {
      struct regval *t_branch = assigns_to(lvalue, this->t.branch.if_true);
      struct regval *f_branch = assigns_to(lvalue, this->t.branch.if_false);
      struct bitvec *cond = this->t.branch.cond;
      if (!t_branch && !f_branch)
         error("Internal: at least one assignment expected in branch");
      if (is_async_set_reset(cond, async, this->lineno))
      {  /* flipflop with asynchronous set/reset does not translate to MUX */
         struct muxtree *m;
         if (!f_branch)
         {
            mesg("** flipflop has async set/reset but no normal input,"
                 " line %1d of file '%s'\n", this->lineno, yyfilename);
            return 0;
         }
         else if (t_branch)
         {
            /* the `true' branch contains the asynchronous assignments... */
            m = build_mux_tree(this->t.branch.if_true, lvalue, async);
            m->next = async->set_reset_list;
            async->set_reset_list = m;
            m->sel = cond;         /* we are abusing the SEL field here ... */
         }
         /* ...and the `false' branch contains all normal assignments */
         node = build_mux_tree(this->t.branch.if_false, lvalue, async);
         this->flags |= Is_Async_Latch;
      }
      else
      {  /* normal if-else that translates to a MUX */
         node = new_muxtree(this);
         node->sel = cond;
         if (t_branch)
         {
            node->t = build_mux_tree(this->t.branch.if_true, lvalue, async);
            t_branch->muxtree = node->t;
         }
         if (f_branch)
         {
            node->f = build_mux_tree(this->t.branch.if_false, lvalue, async);
            f_branch->muxtree = node->f;
         }
         if ((node->t != 0)
             && (node->f != 0)
             && (node->t->flags & MF_IS_COMPLETE)
             && (node->f->flags & MF_IS_COMPLETE))
            node->flags |= MF_IS_COMPLETE;
      }
      return node;}
   
   case Stmt_Case:
   {
      struct choice *choice = this->t.choice.list;
      struct muxtree *prev_choice = 0;
      struct regval *default_regv = 0;
      int j = 0, m = 0;
      int num_not_assigned = 0, num_selectors = 0;
      int all_permutations_used, all_choices_used, all_assigns_complete;
      int default_assignment;
      int num_assigned = 0;
      int length = this->t.choice.guard->length;
      struct bitvec *guard_vec;
      struct choice *default_choice = this->t.choice.deflt;
      for (choice = this->t.choice.list; choice != 0; choice = choice->next)
         if (assigns_to(lvalue, choice->statement))
         {
            num_assigned += 1;
            num_selectors += count_selectors(choice);
         }
         else
            num_not_assigned += 1;
      default_assignment =
         ((default_choice != 0)
          && (default_regv = assigns_to(lvalue, default_choice->statement)));
      guard_vec = new_bitvec(num_assigned);
      node = new_muxtree(this);
      node->sel = guard_vec;
      all_assigns_complete = TRUE;
      for (j = 0, m = 0, choice = this->t.choice.list; choice != 0;
           m += 1, choice = choice->next)
         if ((regv = assigns_to(lvalue, choice->statement)) != 0)
         {
            struct muxtree *mt
               = build_mux_tree(choice->statement, lvalue, async);
            regv->muxtree = mt;
            all_assigns_complete
               = (all_assigns_complete && (mt->flags & MF_IS_COMPLETE));
            guard_vec->bits[j++] = this->t.choice.guard->bits[m];
            if (prev_choice == 0)
               node->choices = mt;
            else
               prev_choice->next = mt;
            prev_choice = mt;
         }
      /* here we do some heuristics and guessing to see if the assignment
       * could possibly be complete. For a real analysis we should be doing
       * some serious boolean function analysis, but that is left as an
       * exercise to the reader.
       */
      all_permutations_used =
         /* this assumes that all guards are different: */
         ((1 << this->t.choice.expr->length) == num_selectors);
      all_choices_used =
         /* this case statement assigns under all conditions: */
         ((num_not_assigned == 0) && default_assignment);
      if ((all_permutations_used && all_assigns_complete)
          ||
          (all_choices_used && all_assigns_complete))
         node->flags |= MF_IS_COMPLETE;
      if ((!all_choices_used)
          || (!all_permutations_used && !default_assignment))
         node->flags |= MF_HAS_NULL_BRANCH;
      if (stmt_case_testflag(this, Case_Parallel))
         node->flags |= MF_IS_PARALLEL_CASE;
      else
         node->flags &= ~MF_IS_PARALLEL_CASE;
      return node;
   }
   case Stmt_Null:
   default:
      /* simply ignore */
      return 0;
   }
}

/* return TRUE if ELMT is a element of the vector VEC.
 */
static int is_vector_elmt(struct symslot *vec, struct symslot *elmt)
{
   char *p = strrchr(elmt->name, '/');
   int len = p ? (p - elmt->name) : -1;
   if (len >= 0)
      return strncmp(vec->name, elmt->name, p - elmt->name) == 0;
   else
      return FALSE;
}

/*
 * Traverse the tree of muxes/assignments, while translating them
 * to gates in the process.
 */
static void generate_gates(struct muxtree *mtree)
{
   if (mtree->outsig == 0)
   {                            /* this is a mux */
      if (mtree->sel == 0)
         error("Internal: mtree->sel should have been assigned");
      if (mtree->choices == 0)
      {
         /* this is an if-else construction (2-way branch) */
         generate_gates(mtree->f); /* follow the false/else branch */
         generate_gates(mtree->t); /* follow the true/then branch */
         mtree->outsig = create_mux_gate(mtree);
      }
      else if (mtree->flags & MF_IS_PARALLEL_CASE)
      {
         /* this is a non-prioritized case-endcase construction
          * (multi-way branch). Implement as an AND-OR structure.
          */
         int or_width = mtree->sel->length;
         int j = 0;
         struct muxtree *mt;
         if (mtree->null_choice != 0)
            or_width += 1;
         if (or_width == 1)
         {  /* there is only one choice, no need for AND-OR structure: */
            mt = mtree->choices;
            generate_gates(mt);
            mtree->outsig = mt->outsig; /* just pass it on */
         }
         else
         {
            struct bitvec *or_inputs = new_bitvec(or_width);
            for (mt = mtree->choices; mt != 0; mt = mt->next, j += 1)
            {
               generate_gates(mt);
               or_inputs->bits[j] = generate_gate2(AND, mt->outsig,
                                                   mtree->sel->bits[j]);
            }
            if (mtree->null_choice != 0)
            {
               struct symslot *null_guard
                  = generate_gate(NOR, mtree->sel, mtree->sel->length);
               generate_gates(mtree->null_choice);
               or_inputs->bits[j++]
                  =generate_gate2(AND, mtree->null_choice->outsig, null_guard);
            }
            if (j != or_width)
               error("Internal: n != or_width");
            mtree->outsig = generate_gate(OR, or_inputs, or_width);
            free(or_inputs);
         }
      }
      else
      {
         /* this is a prioritized case-endcase construction
          * (multi-way branch). Implement as a MUX tree.
          */
         int j;
         int length = mtree->sel->length;
         struct muxtree **choices, *mt;
         struct symslot *outsig;
         if (length < 1)
            error("Internal: length should be >= 1");
         choices = calloc(length, sizeof(struct mtree *));
         for (j = 0, mt = mtree->choices; mt != 0; j += 1, mt = mt->next)
            choices[j] = mt;    /* create array of choices */
         generate_gates(choices[length - 1]);
         if (mtree->null_choice == 0)
            outsig = (choices[length - 1])->outsig;
         else
         {
            generate_gates(mtree->null_choice);
            outsig = generate_gate3(MUX,
                                    mtree->sel->bits[length - 1],
                                    (choices[length - 1])->outsig,
                                    mtree->null_choice->outsig);
         }
         for (j = length - 2; j >= 0; j -= 1)
         {
            generate_gates(choices[j]); /* recurse */
            outsig = generate_gate3(MUX,
                                    mtree->sel->bits[j],
                                    (choices[j])->outsig,
                                    outsig);
         }
         mtree->outsig = outsig;
         free(choices);
      }
   }
}


static int is_latch(struct gate *g)
{
   return (g->type == LATCH)
      || (g->type == LATCH_N)
      || (g->type == LATCH_SR)
      || (g->type == LATCH_SR_N);
}


/*
 * Append all signals needed to compute OUTSIG to the SIGS list
 * and return the new list. A symbol occurs at most once in SIGS.
 */
struct symlist *append_read_sigs(struct symslot *outsig,
                                 struct symlist *sigs,
                                 int *loop)
{
   char *name;
   struct gate *g;
   struct gateport *gp;
   if (outsig == 0)
      return sigs;
   name = outsig->name;
   if (name[0] != '!')
   {  /* OUTSIG is not an anonymous (temporary) signal */
      char *end = strrchr(name, '/');
      char pure_name[200];
      struct symslot *pure_sym;
      strncpy(pure_name, name, end - name); /* copy until '/' char */
      pure_name[end - name] = 0;   /* remove index part from the name */
      pure_sym = symfind(pure_name);
      if (pure_sym != 0)
      {
         struct symlist *t;
         for (t = sigs; t != 0; t = t->next)
            if (strcmp(t->symbol->name,pure_name) == 0)
               break;                 /* pure_name is already in SIGS */
         if (t == 0)
         {  /* add an entry for PURE_SYM to the SIGS list */
            t = calloc(1, sizeof(struct symlist));
            t->symbol = pure_sym;
            t->next = sigs;
            sigs = t;
         }
      }
   }
   g = outsig->vhdlgate;        /* get the gate that drives OUTSIG */
   if (g == 0)
      return sigs;              /* maybe this is a module input */
   if (is_latch(g))
      return sigs;              /* not a combinational gate */
   if (g->flags & GF_IN_USE)
   {  /* we are recursing far too deep, something must be wrong! */
      mesg("** Combinational loop detected in '%s' line: %d\n"
           "   Here is the trace:\n"
           "# %1d) %s\n", yyfilename, yylineno, ++*loop, g->outport->port->name);
      return sigs;
   }
   g->flags |= GF_IN_USE;       /* mark for loop detection */
   /* now iterate over all input ports of this gate: */
   for (gp = g->port; (gp != g->outport) && !*loop; gp = gp->next)
   {
      struct symslot *sym = gp->port;
      sigs = append_read_sigs(sym, sigs, loop); /* recurse */
   }
   g->flags &= ~GF_IN_USE;      /* unmark for loop detection */
   if (*loop)
      mesg("# %1d) %s\n", ++*loop, g->outport->port->name);
   return sigs;
}

/*
 * Synthesize a latch with output Q, input D and clock and
 * asynchronous set/reset from from ASYNC:
 */
static int addlatch(struct symslot *q,
                    struct symslot *d,
                    struct async *async,
                    int lineno)
{
   struct symslot *async_set = 0, *async_reset = 0;
   struct muxtree *m;
   struct symlist *sig = async->sig_list, *ck = 0;
   int num_async = 0;
   for (; sig != 0; sig = sig->next)
      if ((sig->flags == SYML_NEGEDGE) || (sig->flags == SYML_POSEDGE))
         if (ck == 0)
            ck = sig;
         else
         {
            mesg("** More than one clock present for statement in line %1d of file '%s'\n",
                 lineno, yyfilename);
            break;
         }
      else if (sig->flags == SYML_ASYNC)
         num_async += 1;
   if (num_async > 2)
      mesg("** Cannot have more than two async set/reset signals per reg, line %1d of file '%s'\n",
           lineno, yyfilename);
   for (m = async->set_reset_list; m != 0; m = m->next)
   {
      if (m->outsig == 0)
         mesg("** Async set/reset must be a simple assignment, in line %1d of file '%s'\n",
              m->stmt->lineno, yyfilename);
      else
      {
         int typ = get_driver_type(m->outsig);
         if (typ == CONST0)
         {
            if (async_reset != 0)
               mesg("** more than one async reset, in line %1d of file '%s'\n",
                    m->stmt->lineno, yyfilename);
            async_reset = m->sel->bits[0]; /* SEL abuse, see build_mux_tree */
         }
         else if (typ == CONST1)
         {
            if (async_set != 0)
               mesg("** more than one async set, in line %1d of file '%s'\n",
                    m->stmt->lineno, yyfilename);
            async_set = m->sel->bits[0]; /* SEL abuse, see build_mux_tree */
         }
         else
            mesg("** Assignment for async set/reset must be constant, line %1d of file '%s'\n",
                 m->stmt->lineno, yyfilename);
      }
   }
   if (async_set || async_reset)
   {
      if (!async_set)
         async_set = generate_gate0(CONST0);
      if (!async_reset)
         async_reset = generate_gate0(CONST0);
   }
   saveports();
   isgate = 1;
   gptr = NULL;
   addport(q);
   addport(d);
   if (async_set || async_reset)
   {
      addport(async_set);
      addport(async_reset);
      if (ck->symbol) addport(ck->symbol);	/* clock pin */
      gtype = (ck->flags & SYML_NEGEDGE) ? LATCH_SR_N : LATCH_SR;
   }
   else
   {
      if (ck->symbol) addport(ck->symbol);	/* clock pin */
      gtype = (ck->flags & SYML_NEGEDGE) ? LATCH_N : LATCH;
   }
   addgate();
   retrports();
   return num_async;
}

/*
 * In Verilog everything that is being assigned from within an ALWAYS
 * block must have been declared "reg". That does not mean that such
 * a signal is actually a latch! Whenever we find that a "reg" is
 * actually just a wire, we call this_reg_is_not_a_latch() to notify
 * the module that it can decrement its "reg" count by 1.
 */
static void this_reg_is_not_a_latch(struct symslot *sym)
{
   struct reg *r = currentmod->reg;
   for (; r != 0; r = r->next)
      if (r->port == sym)
      {
         r->not_a_latch = 1;
         currentmod->numregs -= 1;
         break;
      }
}

static int already_assigned_to_reg(struct regval* lvalue)
{
   struct reg *reg = currentmod->reg;
   for (; reg != 0; reg = reg->next)
      if (lvalue->symbol == reg->port)
      {
         if (reg->already_assigned)
            return TRUE;
         reg->already_assigned = TRUE;
         return FALSE;
      }
   /* lvalue is not a reg ... */
   return FALSE;
}

/*
 * Scan STMT and for each lvalue that it assigns to translate the procedural
 * statements into a set of concurrent gate-level assignments.
 */
void elaborate_always(struct stmt *this,
                      struct symlist *evt_list,
                      int evt_lineno)
{
   static int give_hint = TRUE, loop = 0;
   struct regval *lvalue;
   struct symlist *read_sigs = 0;
   struct async *async_list = 0, *prev_async = 0, *async = 0;
   int clocked_process = (evt_list != 0) && evt_list->flags;
   for (lvalue = this->lvalues; lvalue != 0; lvalue = lvalue->next)
   {
      async = calloc(1, sizeof(struct async));
      if (clocked_process)
         async->sig_list = evt_list;
      if (already_assigned_to_reg(lvalue))
         mesg("** multiple drivers for reg %s in file '%s'\n", lvalue->symbol->name, yyfilename);
      else
         lvalue->muxtree = build_mux_tree(this, lvalue, async);
      if (prev_async == 0)
         async_list = async;
      else
         prev_async->next = async;
      prev_async = async;
   }
   /* at this point, we have built a muxtree for every assignment in the
    * always block. Now we must descent each muxtree to see if it reads
    * register values that have been written previously in this same always
    * block, e.g. "a=5; b=a" yields b=5. Without the following for-loop, this
    * would have behaved as "a<=5; b=a" which is something different because
    * b is connected to the wire a instead of to the value 5.
    */
   for (lvalue = this->lvalues; lvalue != 0; lvalue = lvalue->next)
   {
      int loop = 0;
      replace_reg_with_val(lvalue->muxtree, &loop);
      if (loop)
         mesg("# %1d) Assignment to %s in file '%s'\n", ++loop, lvalue->symbol->name, yyfilename);
   }
   for (lvalue = this->lvalues, async = async_list;
        lvalue != 0; lvalue = lvalue->next, async = async->next)
   {
      struct muxtree *t = lvalue->muxtree;
      if (t == 0)
         continue;              /* can happen if there were errors */
      if (clocked_process && !(t->flags & MF_IS_COMPLETE))
         /* if a clocked process does not assign to lvalue under
          * some conditions, then lvalue remains unchanged:
          */
         if (t->flags & MF_IS_LINKED_TO_SELF)
            t->flags |= MF_IS_COMPLETE; /* null branches already linked  */
         else
         {
            struct muxtree *reg_out = new_muxtree(this);
            reg_out->outsig = lvalue->symbol;
            reg_out->flags |= MF_IS_COMPLETE;
            /* replace all dangling branches with register output: */
            link_to_null_branches(t, reg_out);
            t->flags |= (MF_IS_COMPLETE | MF_IS_LINKED_TO_SELF);
         }
      if (!(t->flags & MF_IS_COMPLETE))
      {
         /* (This can only happen with combinational assignments). The LVALUE
          * is not assigned under all possible input conditions.  This means
          * that the ALWAYS block describes a latch, but it is not a clocked
          * process.
          */
         mesg("** incomplete assignment to %s in line: %d of file '%s'\n",
              lvalue->symbol->name, this->lineno, yyfilename);
         if (give_hint)
         {
            give_hint = FALSE;
            printf("   (HINT: it seems that not all possible control paths\n"
                   "   through the ALWAYS block assign to %s in file '%s'. An easy way\n"
                   "   to fix this is to assign a default value to %s\n"
                   "   at the top of the ALWAYS block)\n",
                   lvalue->symbol->name, yyfilename, lvalue->symbol->name);
         }
      }
      else
      {
         /* we have a complete assignment. Now generate the gates. */
         generate_gates(t);
         /* and finally, assign the output of the gate network to lvalue: */
         if (clocked_process)
         {
            int n = addlatch(lvalue->symbol, t->outsig,
                             async, this->lineno);
            if (n > 0)          /* n is number of async set/reset inputs */
            {
               struct stmt *s = this, *s2 = 0;
               while (s && (s->type == Stmt_Group))
                  s = s->t.group.first;
               if ((s == 0)
                   || (s->type != Stmt_Branch)
                   || !(s->flags & Is_Async_Latch)
                   || ((n > 1) &&
                       (((s2 = s->t.branch.if_false) == 0)
                        || (s2->type != Stmt_Branch)
                        || !(s2->flags & Is_Async_Latch))))
                  mesg("** Async set/reset reg has non-conforming syntax, line %1d of file '%s'\n",
                       this->lineno, yyfilename);
            }
         }
         else
         {
            addwire(lvalue->symbol, t->outsig);
            /* find the signals that we are reading when computing OUTSIG: */
            read_sigs = append_read_sigs(t->outsig, read_sigs, &loop);
         }
      }
      if (!clocked_process)
         /* we must mark this reg special, otherwise the netlist
          * backend in flatten.c will do the wrong thing ... */
         this_reg_is_not_a_latch(lvalue->symbol);
   }
   if (!clocked_process)
      verify_event_list(evt_list, read_sigs, evt_lineno);
}


static void replace_reg_with_val(struct muxtree *mt, int *loop)
{
   if (mt == 0)
      return;
   if (mt->flags & MF_REG_VAL_DONE)
      return;
   mt->flags |= MF_REG_VAL_DONE;
   if (mt->outsig != 0)
   {  /* this is an assignment */
      if (mt->stmt)
         /* e.g. when linking-to-self, the self-mux has a NULL stmt */
         find_reg(&mt->outsig, mt->stmt, loop);
   }
   else if ((mt->choices == 0) && (mt->sel != 0))
   {  /* this is an if-else construction (2-way branch) */
      find_reg(&mt->sel->bits[0], mt->stmt, loop);
      replace_reg_with_val(mt->f, loop); /* follow the false/else branch */
      replace_reg_with_val(mt->t, loop); /* follow the true/then branch */
   }
   else if ((mt->choices != 0) && (mt->sel != 0))
   {  /* this is a case switch */
      struct muxtree *t;
      int i;
      for (i = 0, t = mt->choices; t != 0; i += 1, t = t->next)
      {
         find_reg(&mt->sel->bits[i], mt->stmt, loop);
         replace_reg_with_val(t, loop);
      }
   }
}

static void find_reg(struct symslot **sigptr,
                     struct stmt *stmt,
                     int *loop)
{
   struct gate *g;
   struct gateport *gp;
   struct symslot *sig = *sigptr;
   if (sig == 0)
      return;
   if (sig->flags & SF_REGISTER)
   {
      struct stmt *prev_stmt = find_previous_stmt(stmt);
      struct symslot *val = find_value(sig, prev_stmt, loop);
      if (*loop)
         mesg("# %1d) Assignment to %s, returned by statement in line %d of file '%s'\n",
              ++*loop, sig->name, prev_stmt->lineno, yyfilename);
      if (val != 0)
         /* replace register name with its actual value */
         *sigptr = val;
      return;
   }
   g = sig->vhdlgate;           /* get the gate that drives SIG */
   if (g == 0)
      return;                   /* maybe this is a module input */
   if (is_latch(g))
      return;                   /* not a combinational gate */
   if (g->flags & GF_IN_USE)
   {  /* we are recursing far too deep, something must be wrong! */
      mesg("** Combinational loop detected in '%s' line: %d\n"
           "   Here is the trace:\n"
           "# %1d) %s\n", yyfilename, yylineno, ++*loop, g->outport->port->name);
      return;
   }
   g->flags |= GF_IN_USE;       /* mark for loop detection */
   /* now iterate over all input ports of this gate: */
   for (gp = g->port; (gp != g->outport) && !*loop; gp = gp->next)
      find_reg(&gp->port, stmt, loop); /* recurse */
   g->flags &= ~GF_IN_USE;      /* unmark for loop detection */
   if (*loop)
      mesg("# %1d) %s\n", ++*loop, g->outport->port->name);
}


/*
 * find the statement that is executed sequentially before STMT
 */
static struct stmt *find_previous_stmt(struct stmt *stmt)
{
   struct stmt *s = stmt;
   while ((s != 0) && (s->parent != 0))
      if ((s->parent->type == Stmt_Group) && (s->prev != 0))
         return s->prev;
      else
         s = s->parent;
   return 0;
}


/* Starting at STMT, search for a blocking assignment to the register SIG and
 * return its value (i.e. the right-hand side of that assignment).  Make sure
 * all references to register names that may occur in the value expression
 * have been replaced by their proper value (i.e. this may cause indirect
 * recursion through replace_reg_with_val()).
 */
static struct symslot *find_value(struct symslot *sig,
                                  struct stmt *stmt,
                                  int *loop)
{
   struct regval *rv;
   struct symslot *ret_val = 0;
   if (stmt == 0)
      ;
   else if ((rv = assigns_to_sym(sig, stmt))
            && (rv->symbol->flags & SF_BLOCKING))
   {
      struct muxtree *mt = rv->muxtree;
      if (!(mt->flags & MF_IS_COMPLETE))
      {
         struct muxtree *t = new_muxtree(0);
         t->outsig = sig;
         t->flags |= MF_IS_COMPLETE;
         link_to_null_branches(mt, t);
         mt->flags |= MF_IS_LINKED_TO_SELF;
      }
      replace_reg_with_val(mt, loop);
      generate_gates(mt);
      ret_val = mt->outsig;
   }
   else
      ret_val = find_value(sig, find_previous_stmt(stmt), loop);
   return ret_val;
}


/*
 * if STATEMENT assigns to SYM then return the regval that describes
 * this assignment in detail. Return 0 if STATEMENT does not assign
 * to SYM.
 */
static struct regval *assigns_to_sym(struct symslot *sym,
                                     struct stmt *statement)
{
   struct regval *r;
   if (statement == 0)
      return 0;
   for (r = statement->lvalues; r != 0; r = r->next)
      if (strcmp(sym->name, r->symbol->name) == 0)
         break;
   return r;
}


static struct regval *assigns_to(struct regval *lvalue,
                                 struct stmt *statement)
{
   return assigns_to_sym(lvalue->symbol, statement);
}


/*
 * This function synthesizes gates for the == and != operators.
 */
struct bitvec *compare_vectors(struct bitvec *b1,
                               struct bitvec *b2,
                               int inv)
{
   int j;
   int length = b1->length;
   struct bitvec *xor_out = new_bitvec(length);
   struct bitvec *cmp_out;
   /* first generate a layer of XOR gates */
   xor_out->length = length;
   if (b2->length != length)
      error("Internal: vectors must have equal length");
   for (j = 0; j < length; j += 1)
   {
      int b1_const0 = (get_driver_type(b1->bits[j]) == CONST0);
      int b2_const0 = (get_driver_type(b2->bits[j]) == CONST0);
      int b1_const1 = (get_driver_type(b1->bits[j]) == CONST1);
      int b2_const1 = (get_driver_type(b2->bits[j]) == CONST1);
      int b1_constx = ((get_driver_type(b1->bits[j]) == CONSTX)
                       || (get_driver_type(b1->bits[j]) == CONSTZ));
      int b2_constx = ((get_driver_type(b2->bits[j]) == CONSTX)
                       || (get_driver_type(b2->bits[j]) == CONSTZ));
      if (b1_constx || b2_constx)
         mesg("** Cannot compare against 'x' or 'z' in '%s' line: %d\n",yyfilename, yylineno);
      /* see if we can get away without a XOR gate: */
      if (length > 1)
         if (b1_const0)
            xor_out->bits[j] = b2->bits[j];
         else if (b1_const1)
            xor_out->bits[j] = generate_inv_gate(b2->bits[j]);
         else if (b2_const0)
            xor_out->bits[j] = b1->bits[j];
         else if (b2_const1)
            xor_out->bits[j] = generate_inv_gate(b1->bits[j]);
         else
            xor_out->bits[j] = generate_gate2(XOR, b1->bits[j], b2->bits[j]);
      else /* length == 1 */
         if ((b1_const1 && !inv) || (b1_const0 && inv))
            xor_out->bits[j] = b2->bits[j];
         else if ((b1_const0 && !inv) || (b1_const1 && inv))
            xor_out->bits[j] = generate_inv_gate(b2->bits[j]);
         else if ((b2_const1 && !inv) || (b2_const0 && inv))
            xor_out->bits[j] = b1->bits[j];
         else if ((b2_const0 && !inv) || (b2_const1 && inv))
            xor_out->bits[j] = generate_inv_gate(b1->bits[j]);
         else
            xor_out->bits[j] = generate_gate2(inv ? XOR : XNOR,
                                              b1->bits[j], b2->bits[j]);
   }
   /* now reduce the xor_output array to the single boolean cmp_out: */
   if (length > 1)
   {
      struct symslot *cmp = generate_gate(inv ? OR : NOR, xor_out, length);
      free(xor_out);
      cmp_out = new_bitvec(1);
      cmp_out->bits[0] = cmp;
   }
   else
      cmp_out = xor_out;
   return cmp_out;
}

/*
 * Verify that the event list of an ALWAYS statement contains all
 * identifiers that are read by the statements in the ALWAYS block.
 * It is an error if an identifier is missing from the event list,
 * but it is only a warning if an identifier is in the event list
 * while no statement in the ALWAYS block is reading it.
 */
void verify_event_list(struct symlist *event_list,
                       struct symlist *read_list,
                       int lineno)
{
   struct symlist *rd, *ev;
   for (rd = read_list; rd != 0; rd = rd->next)
   {
      for (ev = event_list; ev != 0; ev = ev->next)
         if (strcmp(rd->symbol->name, ev->symbol->name) == 0)
            break;              /* found RD in event_list */
      if (ev == 0)
         mesg("** signal %s is missing from the event list in line: %d of file '%s'\n",
              rd->symbol->name, lineno, yyfilename);
   }
   for (ev = event_list; ev != 0; ev = ev->next)
   {
      for (rd = read_list; rd != 0; rd = rd->next)
         if (strcmp(rd->symbol->name, ev->symbol->name) == 0)
            break;              /* found EV in event_list */
      if (rd == 0)
         warn("-- WARNING: signal %s in event list not used in line: %d of file '%s'\n",
              ev->symbol->name, lineno, yyfilename);
   }
}

/*
 * test whether we are currently expecting a pragma, then set the new
 * expectation to the value of NEW_STATE.
 */
int pragma_expected(int new_state)
{
   static int currently_expecting_a_pragma = FALSE;
   int current_state = currently_expecting_a_pragma;
   currently_expecting_a_pragma = new_state;
   return current_state;
}

/*
 * Find the 'default' choice in the LIST, then return it.
 */
static struct choice *find_default_choice(struct choice *list)
{
   struct choice *p, *prev = 0, *dflt = 0;
   if (list == 0)
      return 0;
   for (p = list; p != 0; prev = p, p = p->next)
      if (p->sel == 0)          /* default choice has no selector */
      {
         if (dflt != 0)
         {
            mesg("** spurious default clause ignored, in line %d of file '%s'\n",
                 p->lineno, yyfilename);
            prev->next = p->next; /* prev cannot be 0 */
            free(p);
         }
         dflt = p;
      }
   if (dflt == 0)
      return 0;                 /* no default choice present */
   if (dflt->next != 0)
      warn("-- WARNING: default clause is not the last in case, line: %d of file '%s'\n",
           dflt->lineno, yyfilename);
   return dflt;
}
   
/*
 * The CHOICE contains a constant expression that must be compared to
 * the non-constant SWITCH_EXPR. The result is a boolean signal (the
 * "guard") that is then stored in choice->cached_guard.
 */
static struct symslot *generate_choice_guard(struct stmt *this,
                                             struct choice *choice,
                                             struct bitvec *not_switch_expr)
{
   struct bitvec *switch_expr = this->t.choice.expr;
   struct choice_selector *choice_sel = choice->sel;
   int length = switch_expr->length;
   int k = 0, m = 0, max_or_inputs = 0;
   struct bitvec *and_inputs, *or_inputs;
   struct symslot *ret_val = 0;
   for (choice_sel = choice->sel;
        choice_sel != 0; choice_sel = choice_sel->next)
      max_or_inputs += 1;
   /* We are going to create an AND-OR structure to compute the choice guard.
    * A choice list like "foo, bar:" would produce two AND gates followed
    * by a 2-input OR gate.
    */
   or_inputs = new_bitvec(max_or_inputs);
   for (choice_sel = choice->sel;
        choice_sel != 0; choice_sel = choice_sel->next)
   {
      struct bitvec *const_expr = choice_sel->const_expr;
      int j, lineno;
      if (!(this->flags & Case_Parallel)
          && !(this->flags & Case_With_Dontcare)
          /* only check duplicate choices for prioritized case switches
             that do not allow dont-cares:
	   */
          && (lineno = find_duplicate_choice(this->t.choice.list, choice_sel)))
         warn("-- WARNING: CASE switch contains duplicate choice in line: %d of file '%s'\n"
              "   (This choice also appears near line: %d)\n",
              choice->lineno, yyfilename, lineno);
      if (const_expr->length != length)
      {
         if (!choice->msg_size)
         {
            mesg("** CASE expression does not match bit size of selector in line: %d of file '%s'\n",
                 choice->lineno, yyfilename);
            choice->msg_size = TRUE;
         }
         /* set length to the smallest of the two: */
         length = (length < const_expr->length) ? length : const_expr->length;
      }
      /* create the AND part of the AND-OR expression that forms the guard: */
      k = 0;
      and_inputs = new_bitvec(length);
      for (j = 0; j < length; j += 1)
      {
         switch (get_driver_type(const_expr->bits[j]))
         {
         case CONST0:
            if (not_switch_expr->bits[j] == 0)
               not_switch_expr->bits[j] /* not yet cached, generate this bit */
                  = generate_inv_gate(switch_expr->bits[j]);
            and_inputs->bits[k++] = not_switch_expr->bits[j];
            break;
         case CONST1:
            and_inputs->bits[k++] = switch_expr->bits[j];
            break;
         case CONSTX:
         case CONSTZ:
            if (!stmt_case_testflag(this, Case_With_Dontcare))
               if (!stmt_case_testflag(this, Mesg_Dontcare))
               {
                  mesg("** Dont-care values only allowed with casex/casez, line %d of file '%s'\n",
                       choice->lineno, yyfilename);
                  stmt_case_setflag(this, Mesg_Dontcare);
               }
            break;
         default:
            if (!choice->msg_const)
            {
               mesg("** Expression must be constant in line %d of file '%s'\n",
                    choice->lineno, yyfilename);
               choice->msg_const = TRUE;
            }
         }
      }
      if (k == 1)
         or_inputs->bits[m++] = and_inputs->bits[0]; /* no AND gate needed */
      else if (k > 1)
         or_inputs->bits[m++] = generate_gate(AND, and_inputs, k);
      free(and_inputs);
   }
   if (m == 1)
      ret_val = or_inputs->bits[0];
   else if (m > 1)
      ret_val = generate_gate(OR, or_inputs, m);
   free(or_inputs);
   return ret_val;
}

/*
 * Search for a choice selector that is equivalent to CHOICE_SEL.
 * Return its line number if found, zero otherwise.
 */
static int find_duplicate_choice(struct choice *list,
                                 struct choice_selector *choice_sel)
{
   struct choice *choice;
   struct choice_selector *sel;

   for (choice = list ; choice != 0; choice = choice->next)
      for (sel = choice->sel; sel != 0; sel = sel->next)
         if (sel == choice_sel)
            return 0;           /* not found */
         else
         {
            struct bitvec *a = choice_sel->const_expr;
            struct bitvec *b = sel->const_expr;
            if (bitvec_equivalent(a, b))
               return choice->lineno;
         }
   return 0;
}


/*
 * Return TRUE if bitvec A and B are numerical equivalents.
 * The vectors A and B must be constants !!!!
 */
static int bitvec_equivalent(struct bitvec *a,
                             struct bitvec *b)
{
   int j;
   if (a->length != b->length)
      return FALSE;             /* apple and orange */
   for (j = 0; j < a->length; j += 1)
   {
      int typa = get_driver_type(a->bits[j]);
      int typb = get_driver_type(b->bits[j]);
      if ((typa == typb) && (typa != -1))
         continue;
      if ((typa == CONSTX) || (typa == CONSTZ)
          || (typb == CONSTX) || (typb == CONSTZ))
         continue;
      return FALSE;
   }
   return TRUE;
}

/*
 * Generate a signal that is TRUE when all bits in the GUARD_VEC
 * are FALSE. In other words, generate a NOR gate.
 */
static void generate_default_guard(struct bitvec *guard_vec, int j_dflt)
{
   int length = guard_vec->length;
   if (length == 1)
      /* we have a case statement with only a default assignment: */
      guard_vec->bits[j_dflt] = generate_gate0(CONST1);
   else if (length == 2)
      /* one choice and one default */
      guard_vec->bits[j_dflt] = 
         generate_inv_gate(guard_vec->bits[1 - j_dflt]);
   else
      generate_gate_and_leave_in_bitvec(NOR, guard_vec, j_dflt);
}

/*
 * Create a new bit vector of LENGTH:
 */
struct bitvec *new_bitvec(int length)
{
   struct bitvec *bitvec =
      calloc(1, sizeof(struct bitvec) + 
		((length)?(length-1)*sizeof(struct symslot *):0)
		);
   bitvec->length = length;
   return bitvec;
}

/*
 * Split bitvec *APTR into two bitvecs, such that bitvec *BPTR receives
 * the B_SIZE most significant bits of *APTR.
 */
void split_bitvec(struct bitvec **aptr, struct bitvec **bptr, int b_size)
{
   struct bitvec *a = *aptr;
   struct bitvec *new_a;
   struct bitvec *b = new_bitvec(b_size);
   int i, j;
   if (b_size > a->length)
      error("Internal: split_bitvec: b_size too large");
   for (i = 0; i < b_size; i += 1)
      b->bits[i] = a->bits[i];
   new_a = new_bitvec(a->length - b_size);
   for (j = 0; i < a->length; i += 1)
      new_a->bits[j++] = a->bits[i];
   *aptr = new_a;
   *bptr = b;
   free(a);
}

/*
 * Resize bitvec *APTR to NEW_SIZE. Pad with zeros to the left, or truncate
 * the bitvec to the left, whatever is necessary.
 */
void resize_bitvec(struct bitvec **aptr, int new_size)
{
   struct bitvec *a = *aptr;
   struct bitvec *new_a;
   int i, j;
   if (a->length == new_size)
      return;
   new_a = new_bitvec(new_size);
   if (a->length < new_size)
   {
      for (i = 0; i < new_size - a->length; i += 1)
         new_a->bits[i] = generate_gate0(CONST0); /* pad zeros */
      for (j = 0; i < new_size; i += 1)
         new_a->bits[i] = a->bits[j++];
   }
   else
   {
      j = a->length - new_size;
      for (i = 0; i < new_size; i += 1)
         new_a->bits[i] = a->bits[j++];
   }
   free(a);
   *aptr = new_a;
}

struct reg_lval *new_reg_lval(struct reg *reg, struct bit_range *bit_range)
{
   struct reg_lval *lval = calloc(1, sizeof(struct reg_lval));
   lval->reg = reg;
   lval->bit_range = bit_range;
   return lval;
};

/*
 * Generate a gate of TYPE, using the first NUM_INPUTS of the bitvec
 * GATE_INPUTS as the inputs to the gate. Create a new output signal,
 * connect it to the output of the gate, and return it.
 */
struct symslot *generate_gate(int type,
                              struct bitvec *gate_inputs,
                              int num_inputs)
{
   struct symslot *gate_output;
   int j;
   saveports();
   isgate = 1;
   gptr = NULL;
   gate_output = createtemp();
   for (j = 0; j < num_inputs; j += 1)
      addport(gate_inputs->bits[j]);
   gtype = type;
   addgate();
   retrports();
   return gate_output;
}

/* Create gate of TYPE taking inputs from GATE_INPUTS, except from
 * signal at index J_DST, because that is where the output goes.
 * Get it?
 */
static void generate_gate_and_leave_in_bitvec(int type,
                                              struct bitvec *gate_inputs,
                                              int j_dest)
{
   struct symslot *gate_output;
   int j;
   saveports();
   isgate = 1;
   gptr = NULL;
   gate_output = createtemp();
   for (j = 0; j < gate_inputs->length; j += 1)
      if (j != j_dest)
         addport(gate_inputs->bits[j]);
   gtype = type;
   addgate();
   retrports();
   gate_inputs->bits[j_dest] = gate_output;
}

/*
 * Generate an inverter gate. Return the output signal.
 */
static struct symslot *generate_inv_gate(struct symslot *gate_input)
{
   struct symslot *gate_output;
   saveports();
   isgate = 1;
   gptr = NULL;
   gate_output = createtemp();
   addport(gate_input);
   gtype = NOT;
   addgate();
   retrports();
   return gate_output;
}

/*
 * Generate 0-input gate (for const directives)
 */
struct symslot *generate_gate0(int type)
{
   static struct symslot *cache_const0 = 0;
   static struct symslot *cache_const1 = 0;
   static struct symslot *cache_constX = 0;
   if ((type == CONST0) && cache_const0)
      return cache_const0;
   else if ((type == CONST1) && cache_const1)
      return cache_const1;
   else if ((type == CONSTX) && cache_constX)
      return cache_constX;
   else
   {
      struct symslot *gate_output;
      saveports();
      isgate = 1;
      gptr = NULL;
      gate_output = createtemp();
      gtype = type;
      addgate();
      retrports();
      if (type == CONST0) cache_const0 = gate_output;
      else if (type == CONST1) cache_const1 = gate_output;
      else if (type == CONSTX) cache_constX = gate_output;
      return gate_output;
   }
}

/*
 * Generate 1-input gate
 */
struct symslot *generate_gate1(int type,
                                      struct symslot *gate_input1)
{
   struct symslot *gate_output;
   saveports();
   isgate = 1;
   gptr = NULL;
   gate_output = createtemp();
   addport(gate_input1);
   gtype = type;
   addgate();
   retrports();
   return gate_output;
}

/*
 * Generate 2-input gate
 */
struct symslot *generate_gate2(int type,
                                      struct symslot *gate_input1,
                                      struct symslot *gate_input2)
{
   struct symslot *gate_output;
   saveports();
   isgate = 1;
   gptr = NULL;
   gate_output = createtemp();
   addport(gate_input1);
   addport(gate_input2);
   gtype = type;
   addgate();
   retrports();
   return gate_output;
}

/*
 * Generate 3-input gate
 */
struct symslot *generate_gate3(int type,
                               struct symslot *gate_input1,
                               struct symslot *gate_input2,
                               struct symslot *gate_input3)
{
   struct symslot *gate_output;
   saveports();
   isgate = 1;
   gptr = NULL;
   gate_output = createtemp();
   addport(gate_input1);
   addport(gate_input2);
   addport(gate_input3);
   gtype = type;
   addgate();
   retrports();
   return gate_output;
}

struct bitvec *generate_adder(struct bitvec *lhs, struct bitvec *rhs,
                              int operator,
                              int sum2, int carry2, int sum3, int carry3)
{
   struct bitvec *sum;
   struct symslot *carry;
   int i, length;
   
   if (lhs->length != rhs->length)
   {
      warn("-- WARNING: Operands for %c operator have different sizes, in '%s' line: %d\n",
           operator, yyfilename, yylineno);
      align_bitvec(&lhs, &rhs);   /* make 'm the same length */
   }

   length=lhs->length;
   sum=new_bitvec(length);
   sum->bits[length-1]=generate_gate2(sum2, lhs->bits[length-1],
                                      rhs->bits[length-1]);
   if(length>1)
   {
      carry=generate_gate2(carry2, lhs->bits[length-1], rhs->bits[length-1]);
      for(i=length-2;i>=0;i--)
      {
         sum->bits[i]=generate_gate3(sum3, lhs->bits[i], rhs->bits[i], carry);
         if (i > 0)
            carry=generate_gate3(carry3, lhs->bits[i], rhs->bits[i], carry);
      }
   }
   free(lhs); free(rhs);
   return sum;
}


/*
 * make the lengths of bitvecs A and B equal to the longest of the two by
 * padding the shortest bitvec with zeros to the left.
 */
void align_bitvec(struct bitvec **aptr, struct bitvec **bptr)
{
   struct bitvec *a = *aptr;
   struct bitvec *b = *bptr;
   int len = MAX2(a->length, b->length);
   if (a->length < len)
   {  /* pad to the left with zeros */
      struct bitvec *new_a = new_bitvec(len);
      int i, j;
      for (i = 0; i < len - a->length; i+= 1)
         new_a->bits[i] = generate_gate0(CONST0);
      for (j = 0; i < len; i += 1)
         new_a->bits[i] = a->bits[j++];
      free(a);
      *aptr = new_a;
   }
   else if (b->length < len)
   {  /* pad to the left with zeros */
      struct bitvec *new_b = new_bitvec(len);
      int i, j;
      for (i = 0; i < len - b->length; i+= 1)
         new_b->bits[i] = generate_gate0(CONST0);
      for (j = 0; i < len; i += 1)
         new_b->bits[i] = b->bits[j++];
      free(b);
      *bptr = new_b;
   }
}

/* convert a NUMBER to a bitvec of size LENGTH
 */
struct bitvec *number_to_bitvec(int number, int length)
{
   struct bitvec *bv = new_bitvec(length);
   int i;
   long num = (long)number;
   for (i = length - 1; i >= 0; i -= 1)
   {
      if ((num & 1) == 1)
         bv->bits[i] = generate_gate0(CONST1);
      else
         bv->bits[i] = generate_gate0(CONST0);
      num >>= 1;
   }
   return bv;
}


/*
 * Create and initialize a new struct muxtree:
 */
static struct muxtree *new_muxtree(struct stmt *stmt)
{
   struct muxtree *mt = calloc(1, sizeof(struct muxtree));
   mt->stmt = stmt;
   return mt;
}

/*
 * Find out what type of gate is driving SIG. Return -1 if there is no
 * gate driving SIG (e.g. when SIG is an input port):
 */
int get_driver_type(struct symslot* sig)
{
   struct gate *gate = sig->vhdlgate;
   if (gate == 0)
      return -1;
   return gate->type;
}


/* Trace the signals that compute OUTSIG. If a signal occurs also in
 * ASYNC_SIG_LIST then increment *NA, else increment *NS.
 */
static void count_async_and_sync(struct symslot *outsig,
                                 int *na, int *ns,
                                 struct symlist *async_sig_list,
                                 int *loop)
{
   char *name;
   struct gate *g;
   struct gateport *gp;
   if (outsig == 0)
      return;

   name = outsig->name;
   if (name[0] != '!')
   {  /* OUTSIG is not an anonymous (temporary) signal */
      struct symlist *asig = async_sig_list;
      for (; asig != 0; asig = asig->next)
         if (asig->symbol == outsig)
            break;
      if (asig != 0)
      {                         /* asig was found in async_sig_list */
         asig->flags = SYML_ASYNC;
         *na += 1;
      }
      else                      /* asig not found in async_sig_list */
         *ns += 1;
   }
   g = outsig->vhdlgate;        /* get the gate that drives OUTSIG */
   if (g == 0)
      return;                   /* maybe this is a module input */
   if (is_latch(g))
      return;                   /* not a combinational gate */
   if (g->flags & GF_IN_USE)
   {  /* we are recursing far too deep, something must be wrong! */
      mesg("** Combinational loop detected in '%s' line: %d\n"
           "   Here is the trace:\n"
           "# %1d) %s\n", yyfilename, yylineno, ++*loop, g->outport->port->name);
      return;
   }
   g->flags |= GF_IN_USE;       /* mark for loop detection */
   /* now iterate over all input ports of this gate: */
   for (gp = g->port; (gp != g->outport) && !*loop; gp = gp->next)
   {
      struct symslot *sym = gp->port;
      count_async_and_sync(sym, na, ns, async_sig_list, loop); /* recurse */
   }
   g->flags &= ~GF_IN_USE;      /* unmark for loop detection */
   if (*loop)
      mesg("# %1d) %s\n", ++*loop, g->outport->port->name);
}


static int is_async_set_reset(struct bitvec *outvec,
                              struct async *async,
                              int lineno)
{
   int len = outvec->length;
   int i, na = 0, ns = 0, loop = 0;
   struct symlist *async_sig_list;
   if (async == 0)
      return FALSE;             /* list of async sigs is empty */
   async_sig_list = async->sig_list;
   for (i = 0; i < len; i += 1)
      count_async_and_sync(outvec->bits[i], &na, &ns, async_sig_list, &loop);
   if ((na > 0) && (ns > 0))
      mesg("** cannot mix normal signals with async set/reset, line %1d of file '%s'\n",
           lineno, yyfilename);
   if ((na > 0) && (len != 1))
      mesg("** asynchronous condition cannot be a vector, line %1d of file '%s'\n",
           lineno, yyfilename);
   return na > 0;
}
