
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        cmd_do.c
**
**   Functions:   adoit (static), cdoit (static), cpath (static),
**                walk_net (static)
**		  cmd_doactivity, cmd_dochanges, cmd_doclock,
**		  cmd_doexit, cmd_domsg, cmd_dopath, cmd_dostep, cmd_cont
**
**   Description:
**		  This file contains 'cmd_doXXXXXX' functions and
**                all static routines used by those functions.
**                All static routines appear first in the file.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial
**   08/04/88    ARM      Code cleanup; added comments.
**
**
*/

#include "net.h"


/* helper routine for activity command */

static long     abegin, aend, absize;
static long     acounts[NBUCKETS];



/*
**   Function:     adoit (static)
**
**   File:         cmd_do.c
**
**   Called by:    cmd_doactivity indirectly through call to 'walk_net'
**
**   Description:
**
**   Arguments:    n - pointer to the node of interest
**
**   Return Value: None.
*/
static void
adoit (n)
    register nptr   n;
{
    /* skip over the alias nodes  */
    while (n->nflags & ALIAS)
	n = n->nlink;

    if (n->ctime >= abegin && n->ctime <= aend)
	acounts[(n->ctime - abegin) / absize] += 1;
}


/*
**   Function:     cdoit (static)
**
**   File:         cmd_do.c
**
**   Called by:    cmd_dochanges indirectly through call to 'walk_net'
**
**   Description:  print this node's name if we had a transition scheduled
**                 from 'abegin' to 'aend'
**
**   Arguments:    n - pointer to the node of interest
**
**   Return Value: None.
*/
static void
cdoit (n)
    register nptr   n;
{
    /* skip down the list of aliases to find the real node of interest  */
    while (n->nflags & ALIAS)
	n = n->nlink;

    /*
     * test if this node transitioned during our period of interest if so,
     * print it out
     */
    if (n->ctime >= abegin && n->ctime <= aend)
	lprintf (outfile, "  %s\n", pnode (n));
}


/*
**   Function:     cpath (static)
**
**   File:         cmd_do.c
**
**   Called by:    cmd_dopath
**                 itself, recursively
**
**   Description:  this routine outputs a list of nodes which caused
**                 its transition by traversing
**
**   Arguments:    n - pointer to current node of interest
**                 level - the level of ancestry of the node
**                         (0 == FIRST_LEVEL_CALL)
**
**   Return Value: always 1
*/
static void
cpath (n, level)
    register nptr   n;
    int             level;
{
    static long     ptime;	/* used during trace back */

    if (level == FIRST_LEVEL_CALL)
	lprintf (outfile,
		 "critical path for last transition of %s:\n", pnode (n));

    /* skip through the alias' and find the node of interest  */
    while (n->nflags & ALIAS)
	n = n->nlink;

    /* no last transition! */
    if (n->cause == NULL)
	lprintf (outfile, "there is no previous transition!\n");

    /*
     * here if we come across node which has changed more recently than the
     * time we've reached in the backtrace.  We can't continue the backtrace
     * in any reasonable fashion, so we stop here.
     */
    else if (level != 0 && n->ctime > ptime)
    {
	lprintf (outfile,
		 "transition of %s, which has since changed again\n",
		 pnode (n));
    }

    /*
     * here if there seems to be a cause for this node's transition. If the
     * node appears to have 'caused' its own transition (n->cause == n), that
     * means it was input.  Otherwise continue backtrace...
     */
    else
    {
	if (n->cause != n)
	{
	    if (n->cause->nflags & VISITED)
		lprintf (outfile, "  ... loop in traceback\n");
	    else
	    {
		n->nflags |= VISITED;
		ptime = n->ctime;
		cpath (n->cause, level + 1);
		n->nflags &= ~VISITED;
		lprintf (outfile, "%s -> %c @ %2.1fns\n",
			 pnode (n), n->nvalue, 0.1 * (float) n->ctime);
	    }
	}
	else
	    lprintf (outfile, "%s -> %c @ %2.1fns, node was an input\n",
		     pnode (n), n->nvalue, 0.1 * (float) n->ctime);
    }
}


/*
**   Function:     walk_net (static)
**
**   File:         cmd_do.c
**
**   Called by:    cmd_doactivity (cmd_do.c), cmd_dochanges (cmd_do.c)
**
**   Description:  visit each node in the hash table and call
**                 the function 'fun' with the node as the
**                 argument.
**
**   Return Value: None.
*/
static void
walk_net (fun)
    void            (*fun) ();

{
    register int    index;
    register nptr   n;

    for (index = 0; index < HASHSIZE; index++)
    {
	for (n = hash[index]; n != NULL; n = n->hnext)
	{
	    (*fun) (n);
	}
    }
}

/*
**   Function:     cmd_doactivity
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'activity' command
**
**   Description:  routine to print a histogram of circuit activity
**                 in the specified time interval.  Two forms of invocation:
**
**                   1)  starting time explicitly specified, ending
**                       time used is current time
**                       "activity  starting_time"
**
**                   2)  starting and ending times explicitly specified
**                       "activity  starting_time  ending_time"
**
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_doactivity ()
{
#define  LEN_STARS  50		/* number of stars in 'stars' below  */
    static char     stars[LEN_STARS + 1] =
    "**************************************************";

    char           *star_ptr;
    register int    i;
    long            total;

    if (targc == 2)
    {
	abegin = atol (targv[1]);
	aend = cur_delta;
    }
    else if (targc == 3)
    {
	abegin = atol (targv[1]);
	aend = atol (targv[2]);
    }
    else
    {
	verror ("wrong number of args to 'activity' command\n");
	verror ("%s\n", USAGE_ACTIVITY);
	return (CMD_SAME_LEVEL);
    }

    /* perform some error checking on the arguments  */
    if (abegin > aend)
    {
	verror ("ending time must be > starting time\n");
	verror ("%s\n", USAGE_ACTIVITY);
	return (CMD_SAME_LEVEL);
    }

    /* collect histogram info by walking the network */

    /* zero out the histogram  */
    for (i = 0; i < NBUCKETS; i++)
	acounts[i] = 0;

    if ((absize = (aend - abegin + 1) / NBUCKETS) <= 0)
	absize = 1;

    walk_net (adoit);

    for (total = 0, i = 0; i < NBUCKETS; i++)
	total += acounts[i];

    /* print out what we found */

    lprintf (outfile,
	     "%s\n  %2.1fns to %2.1fns (bucket size = %2.1fns)\n",
	     "Histogram of circuit activity",
	     0.1 * (float) abegin, 0.1 * (float) aend, 0.1 * (float) absize);

    for (i = 0; i < NBUCKETS; i++)
    {
	star_ptr = &stars[LEN_STARS - (50 * acounts[i]) / total - 1];
	lprintf (outfile, "  [%ld,%ld]\t(%d)\t%s\n",
		 abegin + (i * absize), abegin + ((i + 1) * absize),
		 acounts[i], star_ptr);
    }

    return (CMD_SAME_LEVEL);
}



/*
**   Function:     cmd_dochanges
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'changes' command
**
**   Description:  routine to print the nodes which transitioned
**                 in the specified time interval.  Two forms of invocation:
**
**                   1)  starting time explicitly specified, ending
**                       time used is current time
**                       "changes  starting_time"
**
**                   2)  starting and ending times explicitly specified
**                       "changes  starting_time  ending_time"
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_dochanges ()
{

    if (targc == 2)
    {
	abegin = atol (targv[1]);
	aend = cur_delta;
    }
    else if (targc == 3)
    {
	abegin = atol (targv[1]);
	aend = atol (targv[2]);
    }
    else
    {
	verror ("wrong number of args to 'changes' command\n");
	verror ("%s\n", USAGE_CHANGES);
	return (CMD_SAME_LEVEL);
    }

    /* perform some error checking on the arguments  */
    if (abegin > aend)
    {
	verror ("ending time must be > starting time\n");
	verror ("%s\n", USAGE_CHANGES);
	return (CMD_SAME_LEVEL);
    }

    lprintf (outfile,
	     "Nodes with last transition in interval %2.1fns to %2.1fns\n",
	     0.1 * (float) abegin, 0.1 * (float) aend);

    walk_net (cdoit);

    return (CMD_SAME_LEVEL);
}



/*
**   Function:     cmd_doclock
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'cl' command
**
**   Description:  routine to run the network through the specified
**                 number of clock cycles;  if no argument is specified,
**                 one clock cycle is executed.
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_doclock ()
{
    register int    n;

    if (targc > 2)
    {
	verror ("wrong number of args to 'cl' command\n");
	verror ("%s\n", USAGE_CLOCK);
	return (CMD_SAME_LEVEL);
    }
    else
    {
	if (targc == 2)
	{
	    n = atoi (targv[1]);

	    if (n <= 0)
	    {
		verror ("illegal argument specified to 'clock' command\n");
		verror ("%s\n", USAGE_CLOCK);
		return (CMD_SAME_LEVEL);
	    }
	}
	else
	    n = 1;
    }

    /* run through the specified clock cycles  */
    while (n-- > 0)
	cycle ();

    return (CMD_SAME_LEVEL);
}



/*
**   Function:     cmd_doexit
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'exit' command
**
**   Description:  this function caues the simulator to exit with
**                 an exit status of 0, targv[1] or '-1' if there
**                 is an error
**
**   Arguments:    No explicit arguments, implicit argument passed
**                 in 'targv'.
**
**   Return Value: this routine never returns.
*/
int
cmd_doexit ()
{
    if (targc > 2)
	exit (-1);
    else if (targc == 2)
	exit (atoi (targv[1]));
    else
	exit (0);
}


/*
**   Function:     cmd_domsg
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'print' command
**
**   Description:  this routine outputs its arguments to the console /
**                 log file.
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_domsg ()
{
    register int    n;

    for (n = 1; n < targc; n++)
	lprintf (outfile, "%s ", targv[n]);

    lprintf (outfile, "\n");

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_dopath
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'path' command
**
**   Description:  this routine determines and prints the critical
**                 path for the specified node by calling 'cpath'
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_dopath ()
{
    if (targc >= 2)
    {
	apply (cpath, FIRST_LEVEL_CALL);
    }
    else
    {
	verror ("wrong number of arguments to path command\n");
	verror ("%s\n", USAGE_PATH);
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_dostep
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'st' command
**
**   Description:  this routine relaxes the network and optionally
**                 sets the current stepsize
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_dostep ()
{
    int             newsize;

    if (targc > 2)
    {
	verror ("incorrect number of arguments to 'st' command\n");
	verror ("%s\n", USAGE_ST);
	return (CMD_SAME_LEVEL);
    }
    else if (targc == 2)
    {
	newsize = atoi (targv[1]);

	if (newsize <= 0)
	{
	    verror ("bad step size specified to 'st' command");
	    verror ("%s\n", USAGE_ST);
	    return (CMD_SAME_LEVEL);
	}

	stepsize = newsize;
    }

    /* step the network ahead  */
    step (cur_delta + stepsize);

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_cont
**
**   File:         cmd_do.c
**
**   Called by:    invocation of 'cont' command
**
**   Description:  this routine relaxes the network
**
**   Arguments:    None.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_cont ()
{
    int             newsize;

    if (targc > 1)
    {
	verror ("no arguments expected to 'cont' command\n");
	return (CMD_SAME_LEVEL);
    }

    /* step the network ahead as little as possible */
    step (cur_delta + 1);

    return (CMD_SAME_LEVEL);
}



