
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        cmd_set.c
**
**   Functions:   dump_list (static), xtrace (static), cmd_setclock, cmd_setdbg,
**                cmd_setlog, cmd_setstep, cmd_settrace,
**                cmd_set_hi, cmd_set_lo, cmd_set_xinput
**
**   Description:
**		  This file contains 'cmd_setXXXXXX' functions and
**                all static routines used by those functions.
**                All static routines appear first in the file.
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include "net.h"


/*
**   Function:     dump_list (static)
**
**   File:         cmd_set.c
**
**   Called by:    cmd_set_hi (cmd_set.c), cmd_set_lo (cmd_set.c)
**
**   Description:  output the specified node list being tied high/low
**
**   Arguments:    n - the node to delete
**                 list - the list to search for the node
**
**   Return Value: None.
*/
static void
dump_list (list_ptr)
    register iptr   list_ptr;
{
    if (list_ptr == NULL)
    {
	/* no nodes in the list => nothing to print  */
	lprintf (outfile, "   None.\n");
    }
    else
    {
	while (list_ptr != NULL)
	{
	    /* dump out the node name  */
	    lprintf (outfile, "   %s\n", list_ptr->inode->nname);

	    list_ptr = list_ptr->next;
	}
    }
}


/*
**   Function:     xtrace (static)
**
**   File:         cmd_set.c
**
**   Called by:    cmd_settrace (cmd_set.c) via call to fapply
**
**   Description:  this routine sets or clears the trace bit for
**                 the node, 'n' passed in.
**
**   Arguments:    n - the node to set/clear trace bit for
**                 flag - 0 => set trace bit
**                        not 0 => clear trace bit
**
**   Return Value: None.
*/
static void
xtrace (n, flag)
    nptr            n;
    int             flag;
{
    while (n->nflags & ALIAS)
	n = n->nlink;

    if (flag)
	n->nflags &= ~TRACED;
    else
	n->nflags |= TRACED;
}




/*
**   Function:     cmd_setdbg
**
**   File:         cmd_set.c
**
**   Called by:    invocation of 'debug' command
**
**   Description:  this routine sets the current debug level to
**                 the second argument, if no argument is specified,
**                 debug is turned off.
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
/* set debugging level */
int
cmd_setdbg ()
{
    int             new_debug;

    if (targc == 1)
    {
	debug = 0;
	lprintf (outfile, "debug printouts disabled\n");
    }
    else if (targc == 2)
    {
	new_debug = atoi (targv[1]);

	if (new_debug < 0)
	{
	    verror ("illegal debug level specified\n");
	    verror ("%s\n", USAGE_DEBUG);
	}
	else
	{
	    debug = new_debug;
	    lprintf (outfile, "debug level set to %d\n", debug);
	}
    }
    else
    {
	verror ("wrong number of args to 'debug' command\n");
	verror ("%s\n", USAGE_DEBUG);
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_setlog
**
**   File:         cmd_set.c
**
**   Called by:    invocation of 'logfile' command
**
**   Description:  close the current logfile and attempt to open
**                 a new logfile, if one has been specified
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_setlog ()
{
    if (logfile != NULL)
    {
	fclose (logfile);
	logfile = NULL;
	lprintf (outfile, "current logfile closed\n");
    }

    if (targc == 2)
    {
	/* attempt to open the new logfile  */
	if ((logfile = fopen (targv[1], "w")) == NULL)
	{
	    verror ("cannot open logfile %s for writing\n", targv[1]);
	    verror ("%s\n", USAGE_LOGFILE);
	}
    }
    else if (targc > 2)
    {
	verror ("%s\n", "wrong number of args to 'logfile' command");
	verror ("%s\n", USAGE_LOGFILE);
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_setstep
**
**   File:         cmd_set.c
**
**   Called by:    invocation of 'stepsize' command
**
**   Description:  set current stepsize to argument specified;
**                 if no argument, print out current stepsize
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
/* set stepsize */
int
cmd_setstep ()
{
    int             newsize;

    if (targc == 1)
    {
        lprintf (outfile, "stepsize = %d = %d.%d ns\n", 
                 stepsize, stepsize/10, stepsize%10);
    }
    else if (targc == 2)
    {
	newsize = atoi (targv[1]);

	if (newsize <= 0)
	{
	    verror ("bad step size specified to 'stepsize' command\n");
	    verror ("%s\n", USAGE_STEPSIZE);
	}
	else
	{
	    stepsize = newsize;
	    lprintf (outfile, "new stepsize = %d = %d.%d ns\n", 
                     stepsize, stepsize/10, stepsize%10);
	}
    }
    else if (targc > 2)
    {
	verror ("incorrect number of arguments to 'stepsize' command\n");
	verror ("%s\n", USAGE_STEPSIZE);
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_settrace
**
**   File:         cmd_set.c
**
**   Called by:    invocation of 'trace' command
**
**   Description:  add or remove the specified nodes from being traced
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_settrace ()
{
    if (targc >= 2)
    {
	fapply (xtrace);
    }
    else
    {
	verror ("wrong number of arguments to 'trace' command\n");
	verror ("%s\n", USAGE_TRACE);
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_set_hi
**
**   File:         cmd_set.c
**
**   Called by:    invocation of 'hi' command
**
**   Description:  routine to tie the specified input nodes high
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_set_hi ()
{
    if (targc == 1)
    {
	lprintf (outfile, "Input nodes tied high:\n");
	dump_list (hinputs);
    }
    else
    {
	apply (setin, 'h');
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_set_lo
**
**   File:         cmd_set.c
**
**   Called by:    invocation of 'lo' command
**
**   Description:  routine to tie the specified input nodes low
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_set_lo ()
{
    if (targc == 1)
    {
	lprintf (outfile, "Input nodes tied low:\n");
	dump_list (linputs);
    }
    else
    {
	apply (setin, 'l');
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_set_xinput
**
**   File:         cmd_set.c
**
**   Called by:    invocation of 'xinput' and 'input' commands
**
**   Description:  remove any strapping from the specified input nodes
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_set_xinput ()
{
    if (targc == 1 || !strcmp (targv[0], "input"))
    {
	lprintf (outfile, "Input nodes tied high:\n");
	dump_list (hinputs);
	lprintf (outfile, "Input nodes tied low:\n");
	dump_list (linputs);
    }
    else
    {
	/* delete the specified nodes from all input lists  */
	apply (setin, 'x');
    }
    return (CMD_SAME_LEVEL);
}
