
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        cmd_vec.c
**
**   Functions:   cmd_clear_vectors, cmd_runvec, cmd_setbits, cmd_setvec,
**                cmd_sb_vec
**
**   Description: command routines assoctiated with vectors
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include "net.h"

/*  pointer to bit vector used to assign wildcarded bitvec's  */
static  bptr   global_bit_vec_ptr;

/*
**   Function:     add_bit_vec (static)
**
**   File:         cmd_vec.c
**
**   Description:  routine to add bit vector to bitvec structure
**
**   Arguments:    n - pointer to node to add
**                 arg - not used (used by match_net)
**
**   Return Value: None.
*/
static void
add_bit_vec (n, arg)
    nptr n;
    int arg;
{
    bptr b;

    b = global_bit_vec_ptr;

    if (b->bnbits >= MAXBITS)
    {
        verror ("Too many bits to watch (%d)\n", MAXBITS);
    }
    else
    {
        b->bbits[b->bnbits++] = n;
    }
}    


/*
**   Function:     cmd_set_bits
**
**   File:         cmd_vec.c
**
**   Called by:    invocation of 'bitvec' command
**
**   Description:  routine to add/remove bit vectors from display list
**
**   Arguments:    None.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_set_bits ()
{
    register nptr   n;
    register bptr   b;
    register int    i;
    register bptr   prev;
    int             j, node_err = FALSE;
    int             found_it;
    nptr            tmp_nptr;
    int		    k;

    /* if no arguments, simply output all bit vectors  */
    if (targc == 1)
    {
	/* output all bit vectors  */

	/* dump the header  */
	lprintf (outfile, "Bit vectors defined:\n");

	/* dump the list (or a message indicating no bit vectors)  */
	if (blist == NULL)
	{
	    lprintf (outfile, "   None.\n");
	}
	else
	{
	    /* dump all bit vectors  */
	    for (b = blist; b != NULL; b = b->bnext)
	    {
		/* dump the bit vector name  */
		lprintf (outfile, "   Bit Vector %s defined by (right to left)\n",
			 b->bname);

		/* dump all the individual node names  */
		for (i = 0; i < b->bnbits; i++)
		{
		    lprintf (outfile, "      %s\n",
			     b->bbits[i]->nname);
		}
	    }
	}
    }
    else
    {
        /*  add/remove a bit vector from the display list  */

	for (i = 1; i < targc; i++)
	{
	    if (targv[i][0] == '-')
	    {
		/*
		 * if vector name begins with a "-", remove it from display
		 * list
		 */

		for (found_it = FALSE, b = blist, prev = NULL;
                     b != NULL; prev = b, b = b->bnext)
		{
		    if (!strcmp (b->bname, &targv[i][1]))
		    {
			/* we found the bit vector the user wants to remove  */
			/* unlink the vector  */
			if (prev == NULL)
			    blist = b->bnext;
			else
			    prev->bnext = b->bnext;

			/* free the memory  */
			free ((char *) b);

                        /*  mark that we found it  */
                        found_it = TRUE;
		    }
		}
		if (found_it == FALSE)
		{
		    verror ("cannot find vector %s on display list\n", &targv[i][1]);
		}
	    }
	    else
	    {
		/* define a bit vector, verify all nodes defining the vector exist  */

		if (targc <= (i + 1))
		{
		    verror ("not enough arguments to create bit vector\n");
		    return (CMD_SAME_LEVEL);
		}

		for (j = i + 1; j < targc; j++)
		{
		    if (match_net (targv[j], NULL, 0) == 0)
		    {
                        /*  no matches found  */

			verror ("cannot find nodes matching %s\n", targv[j]);
			verror ("creation of vector %s aborted\n", targv[i - 1]);
			node_err = TRUE;
		    }
		}

		if (node_err)
		    return (CMD_SAME_LEVEL);

		/* now we know that the nodes exist, create the bit vector  */

		/* allocate memory for the vector  */
		b = (bptr) malloc (sizeof (struct Bits) + strlen (targv[i]) + 1);

		/* link the vector in  */
		b->bnext = blist;
		blist = b;

		/* initialize the vector's fields  */
		b->bnbits = 0;
		b->bname = (char *) (b + 1);
		strcpy (b->bname, targv[i]);

                global_bit_vec_ptr = b;

                /*
                **  get pointers to all nodes which make up the bitvec
                */
		for (j = i + 1; j < targc; j++)
		{
                    match_net (targv[j], add_bit_vec, 0);
		}

                /*
                **  now sort the bitvec in ascending numberical order;
                **  shortest entries first
                */
                for (j = 0; j < (b->bnbits-1); j++)
                {
                    for (k = j+1; k < b->bnbits; k++)
                    {
		      if (swap_strings(b->bbits[j]->nname, b->bbits[k]->nname)) {
			/*  perform the swap  */
			tmp_nptr = b->bbits[k];
			b->bbits[k] = b->bbits[j];
			b->bbits[j] = tmp_nptr;
		      }
                    }
                }

		return (CMD_SAME_LEVEL);
	    }
	}
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_vec_clock
**
**   File:         cmd_vec.c
**
**   Called by:    invocation of 'clock' command
**
**   Description:  routine to display/setup clock vector of values
**                 for nodes
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_vec_clock ()
{
    int             i;

    /* check the arguments  */

    /* if no arguments, simply output the current clock vector  */
    if (targc == 1)
    {
	dump_vlist (VLIST_CLOCK);
    }
    else if ((targc % 2) != 1)
    {
	/* arguments should occur in pairs, this is an error  */
	verror ("error, arguments to set the clock vector should occur in pairs\n");
	verror ("%s\n", USAGE_CLOCK);
    }
    else
    {
	/* the user specified the correct number of arguments  */

	for (i = 1; i < targc; i += 2)
	{
	    add_to_vector_list (VLIST_CLOCK, targv[i], targv[i + 1]);
	}
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_vec_clear
**
**   File:         cmd_vec.c
**
**   Called by:    invocation of 'clrvec' command
**
**   Description:  routine to clear the specified vector list
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
cmd_vec_clear ()
{
    if (targc != 2 ||
	    (strcmp (targv[1], "clock") && strcmp (targv[1], "user")))
    {
	verror ("error parsing 'clrvec' command\n");
	verror ("%s\n", USAGE_CLRVEC);
    }
    else
    {
	/* we either have to clear the clock or user vector list  */
	if (!strcmp (targv[1], "clock"))
	{
	    /* clear the clock vector list  */
	    free_vlist (&vec_clock);
	}
	else if (!strcmp (targv[1], "user"))
	{
	    /* clear the user vector list  */
	    free_vlist (&vec_user);
	}
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_vec_run
**
**   File:         cmd_vec.c
**
**   Called by:    invocation of 'runvec' command
**
**   Description:  routine to clock through all vectors previously
**                 initialized
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_vec_run ()
{
    if (targc > 2)
    {
	verror ("illegal arguments to 'runvec'\n");
	verror ("%s\n", USAGE_RUNVEC);
    }
    else
    {
	if (targc == 1)
	{
	    targv[1] = NULL;
	}
	run_vector (targv[1]);
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_vec_user
**
**   File:         cmd_vec.c
**
**   Called by:    invocation of 'uvec' command
**
**   Description:  routine to display/setup vector of values
**                 for nodes
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_vec_user ()
{
    int             i;

    /* check the arguments  */

    /* if no arguments, simply output the current user vector  */
    if (targc == 1)
    {
	dump_vlist (VLIST_USER);
    }
    else if ((targc % 2) != 1)
    {
	/* arguments should occur in pairs, this is an error  */
	verror ("error, arguments to set the user vector should occur in pairs\n");
	verror ("%s\n", USAGE_UVEC);
    }
    else
    {
	/* the user specified the correct number of arguments  */

	for (i = 1; i < targc; i += 2)
	{
	    add_to_vector_list (VLIST_USER, targv[i], targv[i + 1]);
	}
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_sb_vec
**
**   File:         cmd_vec.c
**
**   Called by:    invocation of 'sbvec' command
**
**   Description:  routine to set a bitvec to a specified value
**
**   Arguments:    None.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_sb_vec ()
{
    int length = 0;
    char *c_ptr;
    register bptr   b;
    register int    i;

    /* if no arguments, simply output all bit vectors  */
    if (targc == 3)
    {
        /*  verify the user's values are acceptable  */
        for (c_ptr = targv[2]; *c_ptr != '\0'; c_ptr++)
        {
            /*  make sure the user specified a reasonable string  */
            if (*c_ptr != '1' && *c_ptr != '0' &&
                *c_ptr != 'x' && *c_ptr != 'z')
            {
                verror ("illegal character in vector value string\n");
                return (CMD_SAME_LEVEL);
            }

            /*  increment the length  */
            length++;
        }
        
        /*  search for the specified bitvec  */
        for (b = blist; b != NULL; b = b->bnext)
        {
            if (!strcmp (b->bname, targv[1]))
            {
                /* we found the bit vector the user wants to give a value to  */

                /*  make sure the number of nodes equals the string length  */
                if (b->bnbits != length)
                {
                    verror ("length mismatch trying to set bitvec value\n");
                    return (CMD_SAME_LEVEL);
                }

                /*  set the node values  */

		/* dump all the individual node names  */
		for (i = 0; i < length; i++)
		{
                    setin (b->bbits[i], targv[2][length - i - 1]);
		}

                /*  success return  */
                return (CMD_SAME_LEVEL);
	    }
        }

        verror ("cannot find vector %s on display list\n", targv[1]);
    }
    else
    {
        /*  error, inform the user how to invoke this command  */
        lprintf (outfile, "error usage: sbvec bitvec_name value\n");
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     swap_strings 
**
**   File:         cmd_vec.c
**
**   Called by:    cmd_sb_vec (cmd_vec.c), cmd_display (cmd_show.c)
**
**   Description:  routine to return TRUE if the first string is > second
**                 string (this implies the strings should be swapped)
**
**   Arguments:    str_1, str_2 - the strings to swap
**
**   Return Value: TRUE if str_1 > str_2, FALSE otherwise
*/
int
swap_strings(str_1, str_2)
char *str_1, *str_2;
{
    int             k, size_1, size_2, val;

  /*
   *  if the lengths aren't the same, only compare
   *  the number of characters of the shorter of the
   *  two, if they are equivalent, then the sorter
   *  of the two is lexically "less than" the
   *  longer of the two.
   */
  size_1 = strlen(str_1);
  size_2 = strlen(str_2);
  
  if (size_1 != size_2)
    {
      if (size_1 < size_2)
	{
	  /*  compare 'size_1-1' characters  */
	  val = strncmp (str_1, str_2, size_1-1);
	  if (val > 0)
	    {
	      /*  swap the j-th and k-th entries  */
	      return(TRUE);
	    }
	}
      else
	{
	  /*  compare 'size_2-1' characters  */
	  val = strncmp (str_1, str_2, size_2-1);
	  if (val >= 0)
	    {
	      /*  swap the j-th and k-th entries  */
	      return(TRUE);
	    }
	}
    }
  else
    {
      /*  they are the same length, do a simple comparison  */
      if (strcmp (str_1, str_2) > 0)
	{
	  /*  swap the j-th and k-th entries  */
	  return(TRUE);
	}
    }
    return(FALSE);
}
