
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        network.c
**
**   Functions:   idelete, iinset, info, setin
**
**   Description:
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include "net.h"





/*
**   Function:     idelete
**
**   File:         network.c
**
**   Called by:    xwatch (cmd_show.c), setin (network.c)
**
**   Description:  delete node 'n' from list 'list', and return
**                 the node's memory to the free list;  if the node
**                 doesn't exist on the list, just return (no error)
**
**   Arguments:    n - the node to delete
**                 list - the list to search for the node
**
**   Return Value: None.
*/
void
idelete (n, list)
    register nptr   n;
    register iptr  *list;
{
    register iptr   p = *list;
    register iptr   q;

    if (p == NULL)
    {
	/* no nodes in the list => nothing to delete  */
	return;
    }
    else if (p->inode == n)
    {
	/* the first node on the list  */

	*list = p->next;
	p->next = infree;
	infree = p;
	return;
    }
    else
    {
	/* search the list for the node  */

	for (q = p->next; q != NULL; p = q, q = p->next)
	{
	    if (q->inode == n)
	    {
		/* found the node  */

		p->next = q->next;
		q->next = infree;
		infree = q;
		return;
	    }
	}
    }
}

/*
**   Function:     iinsert
**
**   File:         network.c
**
**   Called by:    xwatch (cmd_show.c), setin (network.c)
**
**   Description:  insert node 'n' onto start of list 'list';
**                 if there is no more node on the free node list,
**                 allocate another bunch of them.
**
**   Arguments:    n - the node to insert
**                 list - the list to insert the node onto
**
**   Return Value: None.
*/
void
iinsert (n, list)
    nptr            n;
    iptr           *list;
{
    register iptr   p = infree;
    register int    i;

    if (p == NULL)
    {
	/* no room on the free node list, allocate more memory  */

	p = (iptr) malloc (10 * sizeof (struct Input));

	/* hook the list up  */
	for (i = 0; i++ < 9; p++)
	{
	    p->next = infree;
	    infree = p;
	}
    }
    else
    {
	/* we found a node, now unlink it from the list  */

	infree = p->next;
    }

    /* linkt the node onto the list  */
    p->next = *list;
    *list = p;
    p->inode = n;
}


/*
**   Function:     info
**
**   File:         network.c
**
**   Called by:    cmd_excl (cmd_misc.c), cmd_quest (cmd_show.c)
**
**   Description:  routine to output information concerning a node
**
**   Arguments:    n - the node to display information about
**                 which - 0 => print out nodes affected by 'n'
**                         1 => print out nodes affecting 'n'
**
**   Return Value: None.
*/
void
info (n, which)
    register nptr   n;
{
    register oblptr obl;
    register obptr  ob;
    register nlptr  nl;
    register nptr   nd;
    char           *name;

    /* can't display information if the node is NULL !!  */
    if (n == NULL)
	return;

    /* get the node's name  */
    name = pnode (n);

    /* skip the alias list  */
    while (n->nflags & ALIAS)
	n = n->nlink;

    /* print out the node's name and current value  */
    lprintf (outfile, "%s = %c\n", name, n->nvalue);

    /* print out if the node is an input  */
    if (n->nflags & FORCE_VALUE)
	lprintf (outfile, "[note:  node is an input]\n");

    if (which == 0)
    {
	/* print out who this node affects  */
	lprintf (outfile, "affects:\n");
	for (obl = n->in; obl; obl = obl->next)
	{
	    ob = obl->this;
	    lprintf (outfile, "\t%s\n", ob->oname);
	    for (nl = ob->out; nl; nl = nl->next)
	    {
		nd = nl->this;
		lprintf (outfile, "\t\t%s = %c\n", nd->nname, nd->nvalue);
	    }
	}
    }

    if (which == 1)
    {
	/* print out who affects this node  */

	lprintf (outfile, "affected by:\n");
	for (obl = n->out; obl; obl = obl->next)
	{
	    ob = obl->this;
	    lprintf (outfile, "\t%s\n", ob->oname);
	    for (nl = ob->in; nl; nl = nl->next)
	    {
		nd = nl->this;
		lprintf (outfile, "\t\t%s = %c\n", nd->nname, nd->nvalue);
	    }
	}
    }

    /* print out pending event information  */
    if (n->events)
    {
	register evptr  e;

	lprintf (outfile, "Pending events:\n");
	for (e = n->events; e; e = e->nlink)
	    lprintf (outfile, "   transition to %c at %2.1fns\n",
		     e->eval, 0.1 * (float) e->ntime);
    }
}



/*
**   Function:     setin
**
**   File:         network.c
**
**   Called by:     cmd_setvalue (cmd_set.c), cycle (vec.c)
**                  run_vector (vec.c)
**
**   Description:  routine to setup 'n' as an input node with the
**                 value 'which'
**
**   Arguments:    n - pointer to node to setup as an input
**                 which - value to give to input; either 'h', 'l', or 'x'
**
**   Return Value: None.
*/
void
setin (n, which)
    register nptr   n;
    char            which;
{
    /* skip the alias list  */
    while (n->nflags & ALIAS)
	n = n->nlink;

#if 0
this printout gets annoying!
    if (!(n->nflags & FORCE_VALUE))
	lprintf (stderr, "making node %s an input\n", pnode (n));
#endif

    /*
     * make sure the node only exists in one list at a time,  delete the node
     * from any  list it is already in.
     */
    idelete (n, &hinputs);
    idelete (n, &linputs);
    idelete (n, &xinputs);
    idelete (n, &zinputs);

        /* 
         *  set the value to '0' or '1' (hi or lo) 
         *  even before we post
         *  an event for it, this is in case someone
         *  else uses this node's value before the
         *  event we are posting happens, they use the
         *  new value, NOT the old value.  we need to
         *  post the event so other gates fed by this
         *  node are updated properly
         */

    switch (which)
    {
      case '1':
      case 'h':		/* set the input to a high value  */
	iinsert (n, &hinputs);
	n->nflags |= (FORCE_VALUE | FORCE_ONCE);

        /*  set the time that we last transitioned to now if the values are different  */
        if (n->nvalue != '1')
        {
            n->ctime = cur_delta;
            n->cause = NULL;

	    n->delta_stamp = delta_timestamp;

	    /*  clear the rising and falling bits  */ 
	    n->nflags &= ~(FLAG_GATE_RISING | FLAG_GATE_FALLING);

	    /*  we are transitioning from 0 to 1  */
	    n->nflags |= FLAG_GATE_RISING;
        }
        
        n->nvalue = '1'; 
	new_out (n, '1', NO_DELTA, INERTIAL_DELAY);
	break;

      case '0':
      case 'l':		/* set the input to a low value  */
	iinsert (n, &linputs);
	n->nflags |= (FORCE_VALUE | FORCE_ONCE);

        /*  set the time that we last transitioned to now if the values are different  */
        if (n->nvalue != '0')
        {
            n->ctime = cur_delta;
            n->cause = NULL;

	    n->delta_stamp = delta_timestamp;

	    /*  clear the rising and falling bits  */ 
	    n->nflags &= ~(FLAG_GATE_RISING | FLAG_GATE_FALLING);

	    /*  we are transitioning from 1 to 0  */
	    n->nflags |= FLAG_GATE_FALLING;
        }
        
        n->nvalue = '0'; 
	new_out (n, '0', NO_DELTA, INERTIAL_DELAY);
	break;


      case 'x':		/* set the input to a low value  */
	iinsert (n, &xinputs);
	n->nflags |= (FORCE_VALUE | FORCE_ONCE);

        /*  set the time that we last transitioned to now if the values are different  */
        if (n->nvalue != 'x')
        {
            n->ctime = cur_delta;
            n->cause = NULL;

	    n->delta_stamp = delta_timestamp;

	    /*  clear the rising and falling bits  */ 
	    n->nflags &= ~(FLAG_GATE_RISING | FLAG_GATE_FALLING);

	    /*  we are transitioning from 1 to 0  */
	    n->nflags |= FLAG_GATE_FALLING;
        }
        
        n->nvalue = 'x'; 
	new_out (n, 'x', NO_DELTA, INERTIAL_DELAY);
	break;


      case 'z':		/* set the input to a z value  */
	iinsert (n, &zinputs);
	n->nflags |= (FORCE_VALUE | FORCE_ONCE);

        /*  set the time that we last transitioned to now if the values are different  */
        if (n->nvalue != 'z')
        {
            n->ctime = cur_delta;
            n->cause = NULL;

	    n->delta_stamp = delta_timestamp;

	    /*  clear the rising and falling bits  */ 
	    n->nflags &= ~(FLAG_GATE_RISING | FLAG_GATE_FALLING);

	    /*  we are transitioning from 1 to 0  */
	    n->nflags |= FLAG_GATE_FALLING;
        }
        
        n->nvalue = 'z'; 
	new_out (n, 'z', NO_DELTA, INERTIAL_DELAY);
	break;


#if 0
      case 'x':		/* untie the input  */
	n->nflags &= ~(FORCE_VALUE | FORCE_ONCE);
	break;
#endif
    }
}
