
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        queue.c
**
**   Functions:   advance_simulated_time (static),
**                enqueue_event (static), free_event (static),
**                get_next_event (static), seq_init (static),
**                delete_events, enqueue_event, init_nodes, step
**
**
**   Description:
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**		  Anthony J. Bybell(AJB) - Penn State University
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**   08/22/88  SPL/ARM    Correct handling of inertial/transport delays
**   08/24/88    ARM      Add code to handle tying nodes high/low
**                        in enqueue event
**   07/02/90    ARM      Rewrite evaluation of events at current time
**                        to be breadth first instead of depth first
**   08/18/97    AJB	  Added references to xnor.
**   08/24/97    AJB	  Added AET support.
**   01/15/98    AJB      Added supply, tri, and wire primitive refs
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include "net.h"
#include "system_defs.h"

/*
**  structure to perform breadth first queue evaluation  
*/
#define MAX_EVAL_LIST_SIZE 50
struct eval_list {
  int 		    num_eval_entries;
  obptr             eval_array[MAX_EVAL_LIST_SIZE];
  struct eval_list *eval_next;
};
#define NULL_EVAL_LIST ((struct eval_list *) 0L)
static struct eval_list *eval_head, *eval_tail;
static void add_to_list();
static void post_next_events();
static int  get_next_event_time ();
static void eval_queued_gates();
static void process_events();

static void free_event ();

static nptr     cur_node;	/* node that belongs to current event */

/* array serves as head of doubly-linked lists  */
static struct evhdr
{
    evptr           flink, blink;	/* pointers in doubly-linked list */
}               ev_array[TSIZE];

static struct evhdr overflow;	/* head of sorted overflow event list */

static evptr    evfree = NULL;	/* list of free event structures */

static int      npending;	/* number of pending events on event wheel
				 * and overflow list  */
static int      init = 1;	/* network not initialized */




/*
**   Function:     advance_simulated_time (static)
**
**   File:         queue.c
**
**   Called by:    step (queue.c)
**
**   Description:  This routine advances the simulation time forward to
**                 the time specified by 'target_time'.  This assumes
**                 no events are scheduled to happen between 'cur_delta'
**                 and 'target_time'.  This is guaranteed if the
**                 'target_time' passed in is returned from the last
**                 call to get_next_event.  This routine updates
**                 'cur_delta', recalculates the event wheel index, and
**                 moves "near future" events from the overflow list onto
**                 the event wheel.
**
**   Arguments:    target_time - the time to advance to;  this time MUST
**                               be the time of the next event in the
**                               simulation queue / overflow list
**
**   Return Value:
*/
static void
advance_simulated_time (target_time)
    register long   target_time;
{
    register long   time_limit = target_time + TSIZE;
    register evptr  next, marker, target;

    /* set the new curent time to our target time  */
    cur_delta = target_time;

    /*
     * scan the overflow list looking for events to place on the event wheel;
     * since the overflow list is in ascending time order, once we find an
     * entry thay won't fit on the event wheel, we can stop; also once the
     * overflow list wraps around on itself, we can stop.
     */
    for (target = overflow.flink;
	    target->ntime < time_limit && target != (evptr) & overflow;
	    target = next)
    {
	/*
	 * '*target' should go onto the event wheel, compute the event wheel
	 * index --- NASTY!! TSIZE MUST BE A POWER OF 2 FOR THIS TO WORK !!!
	 */
	marker = (evptr) & ev_array[target->ntime & (TSIZE - 1)];

	/* unlink event from overflow list */
	next = target->flink;
	target->blink->flink = next;
	target->flink->blink = target->blink;

	/* then link it onto event wheel at appropriate place */
	target->flink = marker;
	target->blink = marker->blink;
	marker->blink->flink = target;
	marker->blink = target;
    }
}


/*
**   Function:     enqueue_event (static)
**
**   File:         queue.c
**
**   Called by:    new_out (queue.c), update_events (queue.c)
**
**   Description:  this routine is called because it is desired
**                 to drive node 'n' to value 'newval' in a
**                 delta time 'delta';  verify whether it is
**                 necessary to queue a new event and to delete
**                 any old events which exists (depending on the
**                 type of delay, inertial or transport) associated
**                 with this node
**
**   Arguments:    n - node this event is associated with
**                 newval - new value for node 'n'
**                 delta - time in the future (a true delta time) that this
**                         event should occur
**                 transport - type of delay the node has:
**                             0  => inertial delay
**                             !0 => transport delay
**
**   Return Value: None.
*/
static void
enqueue_event (n, newvalue, delta, delay_type)
    register nptr   n;
    char            newvalue;
    long            delta;
    int             delay_type;
{
    register evptr  marker, new, cur_ev;
    register long   etime;
    char            oldvalue;
    register int    nbunch;
    evptr           event;

    /* compute the time in the future that this event should occur  */
    etime = cur_delta + delta;

    oldvalue = n->nvalue;
    
    /*
     * do not let a node which is tied high/low have an event queued on it,
     * unless we are forcing the event
     */
    if ((n->nflags & FORCE_VALUE) && !(n->nflags & FORCE_ONCE))
	return;

    /*
     * if we are forcing the event high/low once (i.e. we have just tied it
     * high/low), change the value to let the code think we have just
     * transitioned now
     */
    if (!(n->nflags & FORCE_ONCE))
    {
	/* we don't want to force the event, do normal processing  */

	/*
	 * check to see if there is any reason to add the event; if we have
	 * no events for this node and our new value is the same as our old
	 * value, we are already in the state we were going to post an event
	 * for, and there are no events that can possibly effect this.
	 */
	if ((n->events == NULL) && (newvalue == oldvalue))
	    return;
#if 0
        /*  
         *  WARNING!!
         *  hack to fix the problem with uninitialized nodes
         *  being set to '0', don't delete pending events from this
         *
         *  THIS IS NOT NEEDED ONCE MULTI-VALUE LOGIC IS
         *  IMPLEMENTED PROPERLY
         */
        if (oldvalue == 'u')
        {
            /*
             * scan through all events associated with 
             * this node and delete those
             * events which are scheduled to occur at our posting
             */
            for (event = n->events; event != NULL; event = event->nlink)
            {
                if (event->ntime == etime)
                {
                    if (((n->nflags & TRACED) && (debug & DEBUG_EVENTS)) ||
                        (debug & DEBUG_SYSTEM))
                    lprintf (outfile,
			 "deleting transition to %c for %s at %ld\n",
			 event->eval, event->enode->nname, event->ntime);
                    free_event (event);
                }
            }
        }
        else
#endif
        {
            /*
             * delete old events that interfere with new events and report on
             * whether it is necessary to queue this one.
             */

            if (delete_events (n, etime, newvalue, delay_type))
            {
                return;
            }
        }
    }
    else
    {
	/* we want to force the event  */

	/* clear the flag  */
	n->nflags &= ~FORCE_ONCE;

#if 0
        /*  
         *  WARNING!!
         *  hack to fix the problem with uninitialized nodes
         *  being set to '0', don't delete pending events from this
         *  
         *  THIS IS NOT NEEDED ONCE MULTI-VALUE LOGIC IS
         *  IMPLEMENTED PROPERLY
         */
        if (oldvalue == 'u')
        {
            /*
             * scan through all events associated with 
             * this node and delete those
             * events which are scheduled to occur at our posting
             */
            for (event = n->events; event != NULL; event = event->nlink)
            {
                if (event->ntime == etime)
                {
                    if (((n->nflags & TRACED) && (debug & DEBUG_EVENTS)) ||
                        (debug & DEBUG_SYSTEM))
                    lprintf (outfile,
			 "deleting transition to %c for %s at %ld\n",
			 event->eval, event->enode->nname, event->ntime);
                    free_event (event);
                }
            }
        }
        else
#endif
        {
            /*
             * delete old events that interfere with new events
             */

            delete_events (n, etime, newvalue, delay_type);
        }
    }

    /* get event structure.  Allocate a bunch more if we've run out. */
    if ((new = evfree) == NULL)
    {
	new = (evptr) malloc (NBUNCH * sizeof (struct Event));
	for (nbunch = NBUNCH; --nbunch > 0; new++)
	{
	    new->flink = evfree;
	    evfree = new;
	}
    }
    else
	evfree = new->flink;

    /* remember facts about this event */

    /* set the absolute time in the future this event should 'fire'  */
    new->ntime = etime;

    /* set the new value to occur the event time  */
    new->eval = newvalue;

    /* set the new event to point to node n   */
    new->enode = n;

    /* this event was caused by the current node (by definition)  */
    new->cause = cur_node;

    /*
     * find out where new event should be inserted.  If scheduled time is too
     * far into the future, look through overflow list (which is kept sorted
     * by increasing event time).  Otherwise, just add new event to END of
     * event list for appropriate entry in event wheel.
     */
    if (delta >= TSIZE)
    {
	for (marker = overflow.flink;
		etime >= marker->ntime && marker != (evptr) & overflow;)
	    marker = marker->flink;
    }
    else
	marker = (evptr) & ev_array[(cur_delta + delta) & (TSIZE - 1)];

    /* insert event right before event pointed to by marker */
    new->flink = marker;
    new->blink = marker->blink;
    marker->blink->flink = new;
    marker->blink = new;

    /* increment the number of events pending  */
    npending++;

    if (((n->nflags & TRACED) && (debug & DEBUG_EVENTS)) ||
        (debug & DEBUG_SYSTEM))
    {
	lprintf (outfile, "Adding event:transition to %c for %s at %ld\n",
		 new->eval, new->enode->nname, new->ntime);
    }

    /*
     * link the event onto the head of the list of events which effect this
     * node (specified by new->enode)
     */
    new->nlink = n->events;

    /*
     * point this node's event pointer back to the event which we just
     * inserted
     */
    n->events = new;

    if (((n->nflags & TRACED) && (debug & DEBUG_EVENTS)) ||
        (debug & DEBUG_SYSTEM))
    {
	for (cur_ev = n->events; cur_ev != NULL; cur_ev = cur_ev->nlink)
	{
	    lprintf (outfile,
		     "comming event: node:%s, value:%c, time:%ld\n",
		     cur_ev->enode->nname, new->eval, new->ntime);
	}

    }
}



/*
**   Function:     free_event (static)
**
**   File:         queue.c
**
**   Called by:    delete_events (queue.c), step (queue.c)
**
**   Description:  unlink an event from the event wheel or
**                 overflow list and from the node's list of
**                 pending events and add it to the free event
*                  list
**
**   Arguments:    event - pointer to event to free
**
**   Return Value: None.
*/
 /*
  * remove event from all the structures it belongs to and return it to free
  * pool
  */
static void
free_event (event)
    register evptr  event;
{
    register evptr  e;

    /* unhook from doubly-linked event list */
    event->blink->flink = event->flink;
    event->flink->blink = event->blink;
    npending--;

    /* add to free storage pool */
    event->flink = evfree;
    evfree = event;

    /* remove event from node's list of pending events */
    if ((e = event->enode->events) == event)
	event->enode->events = event->nlink;
    else
    {
	while (e->nlink != event)
	    e = e->nlink;
	e->nlink = event->nlink;
    }

    if (debug & DEBUG_SYSTEM)
    {
	lprintf (outfile, "removing event:transition to %c for %s at %ld\n",
		 event->eval, event->enode->nname, event->ntime);
    }
}


/*
**   Function:     seq_init (static)
**
**   File:         queue.c
**
**   Called by:    init_nodes (queue.c)
**
**   Description:  perform an initial execution on all sequential objects
**                 to initialize them
**
**   Arguments:    None.
**
**   Return Value: None.
*/
static void
seq_init ()
{
#ifndef MAX_NAME_LEN
#define MAX_NAME_LEN 20
#endif

    oblptr          b;
    char            init_name[6*MAX_NAME_LEN];
    int             i;

    /*  
     *  loop through all process statements, and initialize each one 
     */
    for (b = seq_obs; b != NULL; b = b->next)
    {
        sprintf (init_name, "%s%s", INIT_FUN_PREFACE,
                 b->this->eval+strlen(INIT_FUN_PREFACE)-1);
        
        for (i = 0; proc_funs[i].fun != NULL; i++)
        {
            if (!strcmp (init_name, proc_funs[i].name))
            {
                /*  we found the initialization routine; call it  */
                (*(proc_funs[i].fun))();
                break;
            }
        }            

        if (proc_funs[i].fun == NULL)
        {
            fprintf (stderr, "error searching for initialization routine '%s'\n",
                     init_name);
            fprintf (stderr, "aborting...\n");
            exit (1);
        }
    }
}



/*
**   Function:     delete_events
**
**   File:         queue.c
**
**   Called by:    enqueue_event (queue.c), enqueue_seq (sequ.c)
**
**   Description:  this routine is called because it is desired
**                 to drive node 'n' to value 'newval' at
**                 absolute time 'etime'; delete old events
**                 already on the queue which are not needed
**                 (as defined by the type of delay, transport
**                 vs. inertial), and return whether we need to
**                 enqueue an event (i.e. we may not need to enqueue
**                 an event because the node may already be in that
**                 state and there are no events for that node, or
**                 another event will drive the node into the
**                 desired state prior to our node firing, etc..)
**
**   Arguments:    n - node this event is associated with
**                 etime - absolute time to drive 'n' to 'newval'
**                 newval - new value for node
**                 delay_type - type of delay the node has
**                             0  => inertial delay
**                             !0 => transport delay
**
**   Return Value: 0 => necessary to queue a new event
**                 1 => not necessary to queue new event
*/
delete_events (n, etime, newval, delay_type)
    register nptr   n;
    long            etime;
    char            newval;
    int             delay_type;
{
    register evptr  event, marked_event = NULL;
    register int    found_one = 0;
    register long   time_diff;

    /*
     * transport delay is performed by deleting all old transactions
     * projected to occur at or after the time at which the earliest new
     * transaction is projected to occur
     */
    /*
     * scan through all events associated with this node and delete those
     * events which are scheduled to occur after the event we are about to
     * post
     */
    for (event = n->events; event != NULL; event = event->nlink)
    {
	if (event->ntime >= etime)
	{
            if (((n->nflags & TRACED) && (debug & DEBUG_EVENTS)) ||
                (debug & DEBUG_SYSTEM))
		lprintf (outfile,
			 "deleting transition to %c for %s at %ld\n",
			 event->eval, event->enode->nname, event->ntime);
	    free_event (event);
	}
    }

    /*
     * Scan backwards from the time of the event, searching for the previous
     * event which effects this node.  If this event drives the node into the
     * desired state, it is not necessary to post a new event.
     */
    time_diff = BIG_NUMBER;

    for (event = n->events; event != NULL; event = event->nlink)
    {
	/*
	 * if this event will happen more immediately than any other event,
	 * remember this event
	 */
	if ((etime - event->ntime) < time_diff)
	{
	    marked_event = event;
	    time_diff = etime - event->ntime;
	}
    }

    /*
     * the event which immediately precedes the event desired at 'etime' is
     * now pointed to by 'marked_event, decide if we need to add another
     * event at 'etime'
     */
    if (marked_event != NULL && marked_event->eval == newval)
    {
	/* no, we don't need to add an event  */
	found_one = 1;
    }
    else
    {
	/* yes, add the event  */
	found_one = 0;
    }

    if (delay_type != TRANSPORT_DELAY)
    {
	/* INERTIAL DELAY PROCESSING  */

	/*
	 * Inertial delay is perfored by first performing transport delay
	 * processing, then delete all transactions (events) for this node
	 * which are to occur between now (cur_delta) and the time for our
	 * event (etime).  The only exception to this is if the transaction
	 * (event) which immediately precedes the new transaction is for this
	 * node and its value component is the same as that of the new
	 * transaction. If so, do not delete this event (transaction) and
	 * report that the new event is not necessary.
	 */

	for (event = n->events; event != NULL; event = event->nlink)
	{
	    /* delete this event if it is not our marked event  */
	    if (event != marked_event || found_one == 0)
	    {
		free_event (event);
	    }
	}

	/*
	 * check if we need to queue a new event; we do not need to queue a
	 * new event if our current value is the same as our desired (new)
	 * value and we have no other events between us
	 */

	if (found_one == 0 && newval == n->nvalue)
	{
	    found_one = 1;
	}
    }

    return (found_one);
}


/*
**   Function:     init_nodes
**
**   File:         queue.c
**
**   Called by:    main (vsim.c)
**
**   Description:  routine to initialize the doubly linked list
**                 event data structures and to initialize the
**                 sequential objects.
**
**   Arguments:    None.
**
**   Return Value: None.
*/
void
init_nodes ()
{
    register int    i;
    register struct evhdr *event;

    nevent = 0;

    /* initialize the double linked lists for all the event structures */
    for (i = 0, event = &ev_array[0]; i < TSIZE; i += 1, event += 1)
	event->flink = event->blink = (evptr) event;

    /* and for the overflow list  */
    overflow.flink = overflow.blink = (evptr) & overflow;

    /* no pending events  */
    npending = 0;

    /* no current node  */
    cur_node = NULL;

    /* run through all sequential statements to initialize them  */
    seq_init ();

    /* mark that initialization is complete  */
    init = 0;
}


/*
**   Function:     new_out
**
**   File:         queue.c
**
**   Called by:    enqueue_seq (sequ.c), prim_and (prims.c),
**                 prim_conditional (prims.c), prim_eq (prims.c),
**                 prim_ge (prims.c), prim_gt (prims.c),
**                 prim_le (prims.c), prim_lt (prims.c),
**                 prim_nand (prims.c), prim_ne (prims.c),
**                 prim_nor (prims.c), prim_not (prims.c),
**                 prim_or (prims.c), prim_selected (prims.c),
**                 prim_xor (prims.c), prim_xnor (prims.c).
**		   prim_supply1 (prims.c), prims_supply0 (prims.c),
**		   prim_tri1 (prims.c), prims_tri0 (prims.c),
**		   prim_wire (prims.c), prim_wor (prims.c),
**		   prim_wand (prims.c), prim_buf (prims.c),
**		   prim_bufif0 (prims.c), prim_bufif1 (prims.c),
**		   prim_notif0 (prims.c), prim_notif1 (prims.c),
**		   prim_nmos (prims.c), prim_pmos (prims.c).
**
**   Description:  this routine is called by the primitives
**                 to queue an event which should define
**                 the deisred transition for 'node'
**
**   Arguments:    node - the node to queue an output for
**                 value - the new value for 'node'
**                 delay - the delta in the future when 'node'
**                         should have 'value'
**                 delay_type - the type of delay associated with
**                             the transition
**
**   Return Value: None.
*/
void
new_out (node, value, delta_delay, delay_type)
    nptr            node;
    char            value;
    long int        delta_delay;
    int             delay_type;
{

    if (node->nflags & INITED)
	node->nflags &= ~INITED;
#if 0
    /*
     *  if we have a node transitioning from 'u' 
     *  with a delta delay of 0, post the event
     *  and force the value now.
     *  this is in case someone
     *  else uses this node's value before the
     *  event we are posting happens, they use the
     *  new value, NOT the old value.  we need to
     *  post the event so other gates fed by this
     *  node are updated properly
     */
    if (node->nvalue == 'u' && delta_delay == NO_DELTA)
    {
        /*  mark the node as being forced  */
	node->nflags |= FORCE_ONCE;

        /*
         *  we need to enqueue event prior to setting value
         *  in case we already have events pending
         */
        enqueue_event (node, value, delta_delay, delay_type);

        node->nvalue = value;
    }
    else
    {
        if (delta_delay == NO_DELTA && !(node->nflags & FORCE_VALUE))
        {
            /*  mark the node as being forced  */
            node->nflags |= FORCE_ONCE;
        }
        
        /*  queue up the event  */
        enqueue_event (node, value, delta_delay, delay_type);
    }
#endif

    /*  queue up the event  */
    enqueue_event (node, value, delta_delay, delay_type);

}

/*
**   Function:     step
**
**   File:         queue.c
**
**   Called by:    cmd_dostep (cmd_do.c), cycle (vec.c)
**                 run_vector (vec.c)
**
**   Description:  routine to run the simulator through time,
**                 firing all events scheduled to fire
**                 through 'stop_time' in a breadth first manner
**
**   Arguments:    stop_time - the time to run the simulator until;
**
**   Return Value: None.
*/
void step2(long stop_time)
{
    interrupt_flag = FALSE;

    /*
     * process events from the current event until specified stop time is
     * reached or until events run out
     */
    
    /*
     *	this is the two phase evaluation loop !!!
     *  Phase 1) set values of signals and queue gates to be modified
     *	Phase 2) update the specified gates
     *
     *  note that 'delta_timestamp' is used to signify which delta 
     *  we are in
     */

    while (get_next_event_time () <= stop_time && npending > 0 && interrupt_flag != TRUE)
    {
      /*  we have events which should happen now, update their values  */
      post_next_events();

      /*  now execute all gates which use the specified gates as inputs  */
      eval_queued_gates();

      /*
       *  the increment MUST be at the bottom of the loop since
       *  we set node->delta_stamp to delta_timestamp in "setin" (network.c)
       *  for both vectors and user hi/lo commands
       */
      delta_timestamp++;
    }

    /*  advance time to that which the user requested only if we didn't have an interrupt  */
    if (interrupt_flag == FALSE)
    {
        cur_delta = stop_time;
    }
    
    return;

}


int aetfirstframe=~0;	/* aet support is grafted into the step function */

void step(long stop_time)
{
if((stop_time>aeton)&&(aetfirstframe)&&(aethandle))
	{
	write_aet_frame(aeton);
	aetfirstframe=0;
	}

step2(stop_time);

if((stop_time>aeton)&&(aethandle)&&((stop_time<=aetoff)||(aetoff==-1)))
	{
	write_aet_frame(stop_time);
	}
}


static void
post_next_events()
{
    register int    i;
    register struct evhdr *event;

    /* return if there are no pending events  */
    if (npending == 0)
	return;

    /*
     * scan through the event wheel, if 'i < 0', we have checked every entry
     * on the event wheel
     */
    for (i = TSIZE, event = &ev_array[cur_delta & (TSIZE - 1)]; --i >= 0;)
    {
	/*
	 * if there is an entry for the current slot in the event wheel,
	 * return the first entry in the current slot on the event wheel
	 */
	if ((evptr) event != event->flink)
	  {
	    process_events (event);

	    /*  all done  */
	    return;
	  } 

	/*
	 * wrap the event wheel around properly when we reach the end of the
	 * array
	 */
	if ((event += 1) == &ev_array[TSIZE])
	  event = &ev_array[0];
      }

    /*
     * if we are here, there are no events on the event wheel, now try the
     * overflow list
     */

    if (overflow.flink != (evptr) & overflow)
    {
      /*  process all events which occur at time specified on the overflow list  */
      process_events (overflow.flink);
      return;
    }
    else
    {
	/*
	 * we are confused because the number of pending events was != 0, yet
	 * we couldn't find any events to process
	 */
	lprintf (stderr, "*** internal error: event list messed up!\n");
	return;
    }
}

static void
add_to_list(this_one)
obptr this_one;
{
  struct eval_list *head;
  int i;

  for (head = eval_head; head != NULL_EVAL_LIST; head = head->eval_next) {

    for (i = 0; i < head->num_eval_entries; i++) {
      if (head->eval_array[i] == this_one) {
	/*  'this_one' is already scheduled to run, just return  */
	return;
      }
    }
  }

  /*  if we get here, we didn't find 'this_one' in the list, place it in the list  */
  if (eval_tail == NULL_EVAL_LIST) {

    /*  initialize the new list  */

    eval_head = eval_tail = (struct eval_list *) malloc(sizeof(struct eval_list));
    eval_tail->eval_next = NULL_EVAL_LIST;
    eval_tail->num_eval_entries = 0;

  } else if (eval_tail->num_eval_entries >= MAX_EVAL_LIST_SIZE) {

    /*  allocate memory for another entry  */

    eval_tail->eval_next = (struct eval_list *) malloc(sizeof(struct eval_list));
    eval_tail = eval_tail->eval_next;
    eval_tail->eval_next = NULL_EVAL_LIST;
    eval_tail->num_eval_entries = 0;
  } 

  /*  ok, just copy the entry to the spot in the tail  */
  eval_tail->eval_array[eval_tail->num_eval_entries] = this_one;
  eval_tail->num_eval_entries++;
}

static void 
eval_queued_gates()
{
  struct eval_list *head, *tmp;
  int i;

  /*  evaluate all the queued gates  */

  for (head = eval_head; head != NULL_EVAL_LIST; head = tmp) {

    for (i = 0; i < head->num_eval_entries; i++) {
      /*  evaluate the gate */
      eval_obj (head->eval_array[i]);
    }

    tmp = head->eval_next;

    /*  don't free 'head' to save time for later  */
    if (head != eval_head)  free(head);
    else {
      eval_head->num_eval_entries = 0;
      eval_head->eval_next = NULL_EVAL_LIST;
      eval_tail = eval_head;
    }
  }
}

static void 
process_events(event)
struct evhdr *event;
{
    struct Event *tmp_e_ptr;
    struct Event *tmp2_e_ptr;
    register nptr   n;
    register oblptr l;

  /*  process all events at this time slot  */
  tmp_e_ptr = event->flink; 
  do {
    /*  update the value of the signal  */
    
    /* advance counter and simulated time to that of this event */
    nevent++;
    advance_simulated_time (tmp_e_ptr->ntime);
    
    /* assign new fields to this node based on the event  */
    n = cur_node = tmp_e_ptr->enode;
    
    /*  only update the time and cause if we are really changing value  */
    if (n->nvalue != tmp_e_ptr->eval)
      {
	n->ctime = tmp_e_ptr->ntime;
	n->cause = tmp_e_ptr->cause;
	n->delta_stamp = delta_timestamp;

	/*  clear the rising and falling bits  */ 
	n->nflags &= ~(FLAG_GATE_RISING | FLAG_GATE_FALLING);

	if (n->nvalue == '1') {
	  /*  we are transitioning from 1 to 0  */
	  n->nflags |= FLAG_GATE_FALLING;

	} else if (n->nvalue == '0') {
	  /*  we are transitioning from 0 to 1  */
	  n->nflags |= FLAG_GATE_RISING;

	}
      }
    
    /* output trace if necessary */
    if ((n->nflags & TRACED) || (debug & DEBUG_SYSTEM))
      {
	lprintf (outfile, "[event #%ld] node %s: %c -> %c @ %2.1fns\n",
		 nevent, n->nname, n->nvalue, tmp_e_ptr->eval,
		 0.1 * tmp_e_ptr->ntime);
      }

    /*
     *    HERE IS WHERE WE ASSIGN THE NEW VALUE TO THE NODE
     */
    n->nvalue = tmp_e_ptr->eval;
    
    /* free the event that just fired  */
    tmp2_e_ptr = tmp_e_ptr->flink;
    free_event(tmp_e_ptr);
    
    /*  queue all the gates which this signal feeds  */
    for (l = n->in; l != NULL; l = l->next)
      add_to_list(l->this);

    tmp_e_ptr = tmp2_e_ptr;

  } while (tmp_e_ptr != NULL && tmp_e_ptr != (evptr)event);
}


static int
get_next_event_time()
{
    register int    i;
    register struct evhdr *event;

    /* return if there are no pending events  */
    if (npending == 0)
	return(0);

    /*
     * scan through the event wheel, if 'i < 0', we have checked every entry
     * on the event wheel
     */
    for (i = TSIZE, event = &ev_array[cur_delta & (TSIZE - 1)]; --i >= 0;)
    {
	/*
	 * if there is an entry for the current slot in the event wheel,
	 * return the first entry in the current slot on the event wheel
	 */
	if ((evptr) event != event->flink)
	  {
	    /*  all done  */
	    return(event->flink->ntime);
	  } 

	/*
	 * wrap the event wheel around properly when we reach the end of the
	 * array
	 */
	if ((event += 1) == &ev_array[TSIZE])
	  event = &ev_array[0];
      }

    /*
     * if we are here, there are no events on the event wheel, now try the
     * overflow list
     */

    if (overflow.flink != (evptr) & overflow)
    {
      /*  process all events which occur at time specified on the overflow list  */
      return(overflow.flink->ntime);
    }
    else
    {
	/*
	 * we are confused because the number of pending events was != 0, yet
	 * we couldn't find any events to process
	 */
	lprintf (stderr, "*** internal error: event list messed up!\n");
	return;
    }
}
