
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        vsim.c
**
**   Functions:   parse_command (static),
**                input_command, main
**
**   Description:
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**		  Anthony J. Bybell(AJB) - Penn State University
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**   08/23/88    ARM      Add support for command and $X aliases.
**   08/24/88    ARM      Add initialization file support, program
**                        version number, and redefine command names.
**   08/24/97    AJB	  Added support for AET's.
**   08/30/97    AJB	  Removed compiler warnings.
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include <string.h>
#include <ctype.h>
#include <signal.h>

#include "system_defs.h"
#include "version.h"

#include "net.h"

struct Bits    *blist = NULL;
nptr            hash[HASHSIZE];

void int_routine(int);	/* proto gets rid of compiler warnings */


/* list of commands and their handlers */

/*
** NOTE:
**	A linear search is performed on this table for each command
**	specified; however, if this table is reordered, the 'help'
**	message will no longer be in alphabetical order.  A resonable
**	solution is to either use separate tables for 'help' and
**	command parsing (not easily maintainable) OR order the table
**	to optimize command search time (and let the 'help' routine
**	alphabetize the table.
*/

struct command  cmdtbl[] = {
    {"activity", "show histogram of circuit activity in specified time interval", cmd_doactivity},
    {"aet", "set on/off times for aet", cmd_aet},
    {"alias", "define/show command or '$X' aliases", cmd_alias},
    {"bitvec", "define a bit vector or display all bit vectors", cmd_set_bits},
    {"cl", "run simulator for one clock vector cycle", cmd_doclock},
    {"changes", "list nodes which changed value within specified time interval", cmd_dochanges},
    {"clock", "set/display clock vector", cmd_vec_clock},
    {"clrvec", "clear clock/user vector", cmd_vec_clear},
    {"cont"  , "continue executing events pending for now", cmd_cont },
    {"dl", "display specified node value of those from display list", cmd_pnlist},
    {"debug", "set debugging level to that specified", cmd_setdbg},
    {"exit", "exit simulator", cmd_doexit},
    {"hi", "strap specified nodes high (1) and set as inputs", cmd_set_hi},
    {"help", "print list of available commands", cmd_help},
    {"input", "display current inputs", cmd_set_xinput},
    {"lo", "strap specified nodes low (0) and set as inputs", cmd_set_lo},
    {"logfile", "close current log file and optionally open new log file", cmd_setlog},
    {"node", "print summary or detailed info on node specification", cmd_print_node},
    {"path", "show critical path of node specification", cmd_dopath},
    {"print", "output string to terminal/logfile", cmd_domsg},
    {"quit", "exit this level of command processing", cmd_quit},
    {"runvec", "run simulator for one user defined vector cycle", cmd_vec_run},
    {"sbvec" , "set a bitvec to a string of values", cmd_sb_vec},
    {"st", "step the network forward one time cycle, can optionally set stepsize", cmd_dostep},
    {"stepsize", "set stepsize to that specified", cmd_setstep},
#if INCLUDE_TIMER
    {"timer", "timer on/off/show - time operations", cmd_timer},
#endif
    {"trace", "add/remove specified node from trace", cmd_settrace},
    {"uvec", "set/display user defined vector", cmd_vec_user},
    {"watch", "add/delete specified nodes from display list", cmd_display},
    {"xinput", "remove specified nodes from input list", cmd_set_xinput},
    {"@", "source a command file", cmd_cmdfile},
    {"?", "print info on other nodes affected by this node", cmd_quest},
    {"!", "print info on other nodes that affect this node", cmd_excl},
    {"|", "comment", cmd_comment},
    {NULL, NULL, NULL}
};

 /*
  * global variable for alias list
  */
struct alias_str alias_list[MAX_ALIAS];


 /* the clock vector list  */
Vector         *vec_clock = NULL;

 /* the user-defined vector list  */
Vector         *vec_user = NULL;

int             lineno;		/* current line number in input file  */

FILE           *outfile;
char           *targv[MAXARGS];	/* pointer to tokens on current command line */
int             debug = 0;	/* <>0 if debugging 	 */
int             nfree = 0;	/* number of bytes of free storage remaining 	 */
int             nindex = 0;	/* number of nodes 	 */
int             pending;	/* <>0 means events are still pending 	 */
long            stepsize = 1000;/* duration, in Delta's, of simulation step	 */
int             targc;		/* number of args on current command line 	 */
iptr            hinputs;        /* list of inputs strapped high */
iptr            linputs;        /* list of inputs strapped low  */
iptr            xinputs;        /* list of inputs strapped x */
iptr            zinputs;        /* list of inputs strapped z  */

iptr            infree;
iptr            wlist = NULL;	/* list of nodes to be displayed 	 */
long            cur_delta = 0;	/* current simulated time 	 */

unsigned long	delta_timestamp = 1; 	/*  make it unsigned so when
					    it rolls over it is still ok  */

long            nevent;		/* number of current event 	 */
obptr           curobj = NULL;	/* pointer to current object	 */
int             interrupt_flag; /* flag used for ^C handling  */
#if INCLUDE_TIMER
int             timer_state = TIMER_OFF;  /*  the current timer state  */
long int        elapsed_time = 0L;

/*  two variables for statistics  */
long st_num_events;	/*  starting number of events  */
long end_num_events;	/*  ending number of events  */
#endif

 /*  global variables for sequential objects  */
oblptr seq_obs;
oblptr end_seq;

/*
**   Function:     parse_command (static)
**
**   File:         vsim.c
**
**   Called by:    input_command (vsim.c)
**
**   Description:  parse 'line' into tokens, store the tokens in
**                 targv and store the number of tokens in targc
**
**   Arguments:    line - the line to parse
**
**   Return Value: None.
*/
 /* parse input line into tokens, filling up targv and setting targc */
static void
parse_command (line)
    register char  *line;
{
    register char **carg = targv;
    register char   ch;

    targc = 0;
    while (ch = *line++)
    {
	if (ch <= ' ')
	    continue;

	if (targc >= MAXARGS - 1)
	{
	    verror ("too many command line arguments\n");
	    break;
	}
	else
	    targc++;

	*carg++ = line - 1;

	while ((ch = *line) && ch > ' ')
	    line++;

	if (ch)
	    *line++ = '\0';
    }
    *carg = 0;
}



/*
**   Function:     input_command
**
**   File:         vsim.c
**
**   Called by:    main (vsim.c), finput (cmd_misc.c)
**
**   Description:  routine to source the file descriptor
**                 passed in with simulator commands
**
**   Arguments:    in - the open file descriptor to source
**                      commands from
**
**   Return Value: None.
*/
void
input_command (in)
    FILE           *in;
{
    register struct command *cmd;
    char            line[LSIZE];
    char            line_copy[2 * LSIZE];
    int             i, j;
    register char  *c_ptr, *d_ptr;
    int             alias_error;

    char           *look_ahd_ptr = NULL;
    int             spot;
    int             mult_alias_found;
    int             one_more_comment;

#if INCLUDE_TIMER
#include <sys/types.h>
#include <sys/timeb.h>
    struct timeb start_time;
    struct timeb end_time;
    int timer_command;
#endif

    while (1)
    {
	/* flush pending output, then output prompt */

	if (outfile == stdout)
	    fflush (outfile);

	if (in == stdin && look_ahd_ptr == NULL)
	    fprintf (stderr, "vsim> ");

        one_more_comment = FALSE;
        
	/* read command in from the lookahead buffer or input stream  */
        if (look_ahd_ptr != NULL)
        {
            /*  get the next line from the lookahead buffer  */
            c_ptr = strchr (look_ahd_ptr, MULT_LINE_ALIAS);
            
            /*  advance the lookahead pointer  */
            if (c_ptr == NULL)
            {
                strcpy (line, look_ahd_ptr);
                look_ahd_ptr = NULL;
                one_more_comment = TRUE;
            }
            else
            {
                spot = c_ptr - look_ahd_ptr;
                strncpy (line, look_ahd_ptr, spot);
                line[spot] = '\0';
                look_ahd_ptr = c_ptr+1;
            }
        }
	else if (fgets (line, LSIZE, in) != line)
        {
	    break;
        }
        else
        {
            /* remember where we are at  */
            lineno++;
        }

        /*  if the first thing on the line is a '|', simply dump the line  */
        for (c_ptr = line; *c_ptr == ' ' || *c_ptr == '\t'; c_ptr++);
        if (*c_ptr == '|')
        {
            if (logfile != NULL)
	        fprintf (logfile, "%s", line);
            continue;
        }

	/* map the line to all lowercase  */
	for (c_ptr = line; *c_ptr != '\0'; c_ptr++)
	{
	    if (isupper (*c_ptr))
	    {
		*c_ptr = tolower (*c_ptr);
	    }
	}

	/* perform '$X' alias substitution  */
	for (c_ptr = line, d_ptr = line_copy; *c_ptr != '\0';)
	{
	    /* if we have a normal character (any but a '$'), simply copy it  */
	    if (*c_ptr != '$')
	    {
		*d_ptr++ = *c_ptr++;
		continue;
	    }

	    /* we have found an alias, try to expand it  */

	    c_ptr++;		/* *c_ptr now points to the macro char  */

	    /* search for the macro character  */
	    for (alias_error = TRUE, i = 0;
	      (i < MAX_ALIAS) && (alias_list[i].alias_name[0] != '\0'); i++)
	    {
		if ((*c_ptr == alias_list[i].alias_name[0]) &&
			(strlen (alias_list[i].alias_name) == 1))
		{
		    /*
		     * bingo, we found the alias character, copy the
		     * substituted chatacters
		     */
		    *d_ptr = '\0';
		    strcat (line_copy, alias_list[i].actual_cmd);

		    /* point *d_ptr to just past the last character  */
		    d_ptr = &line_copy[strlen (line_copy)];

		    alias_error = FALSE;
		    break;
		}
	    }

	    if (alias_error)
	    {
		/* we couldn't convert the character  */
		verror ("error expanding macros $%c\n", *c_ptr);
	    }

	    /* point c_ptr past the macro character  */
	    c_ptr++;
	}

	/* nul-terminate the copy  */
	*d_ptr = '\0';

	/* copy the line  */
	strncpy (line, line_copy, LSIZE - 1);

	/* finished performing '$X' alias substitution  */

        /*  log the output, (only if NOT multi-line alias) */
	if (logfile != NULL && look_ahd_ptr == NULL && !one_more_comment)
	    fprintf (logfile, "%s", line);

	/* parse the line into tokens  */
	parse_command (line);

	/* ignore blank lines  */
	if (targv[0] == NULL)
	    continue;

	/* substitute command alias in if appropriate  */
        mult_alias_found = FALSE;
	for (i = 0; (i < MAX_ALIAS) && (alias_list[i].alias_name[0] != '\0'); i++)
	{
	    if (!strcmp (targv[0], alias_list[i].alias_name))
	    {
		/*
		 * bingo, we found an alias
                 */

                /*  check if we have a multi-line alias  */
                if (strchr (alias_list[i].actual_cmd, MULT_LINE_ALIAS) != NULL)
                {
                    /*  this is a multi-line alias  */

                    /*  get a pointer to the alias line  */
                    look_ahd_ptr = alias_list[i].actual_cmd;
                    
                    /*  set a flag to go to the top of the loop  */
                    mult_alias_found = TRUE;

                    if (in == stdin)
                    {
                        fprintf (stderr, "      %s\n", look_ahd_ptr);
                    }
                }                    
                else
                {
                    /* copy the line and insert the alias  */
                    
                    strcpy (line_copy, alias_list[i].actual_cmd);
                    
                    for (j = 1; j < targc; j++)
                    {
                        strcat (line_copy, " ");
                        strcat (line_copy, targv[j]);
                    }
                    
                    /* copy the line  */
                    strncpy (line, line_copy, LSIZE - 1);
                    
                    /* echo the new aliased command line  */
                    if (in == stdin)
                    {
                        fprintf (stderr, "      %s\n", line);
                    }
                
                    /* parse the command again  */
                    parse_command (line);

                    break;
                }
	    }
	}

        if (mult_alias_found == TRUE)
        {
            continue;
        }

	/* search command table, dispatch to handler, if any */
	for (cmd = &cmdtbl[0]; cmd->name != NULL; cmd++)
	    if (strcmp (targv[0], cmd->name) == 0)
	    {
#if INCLUDE_TIMER              
	        timer_command = (timer_state == TIMER_ON && strcmp(targv[0], "timer"));

                if (timer_command)
                {
                    /*  get the starting time   */
                    ftime(&start_time);
                }
#endif
		if ((*cmd->handler) ())
                {
#if INCLUDE_TIMER                
                    /*  get the ending time   */
                    if (timer_command)
                    {
                        ftime(&end_time);
                        elapsed_time += ((end_time.time * 1000L + end_time.millitm) -
                                         (start_time.time * 1000L + start_time.millitm));
                    }
#endif
                    
		    return;
		  }

#if INCLUDE_TIMER
                /*  get the ending time   */
                if (timer_command)
                {
                    ftime(&end_time);
                    elapsed_time += ((end_time.time * 1000L + end_time.millitm) -
                                     (start_time.time * 1000L + start_time.millitm));
                }
#endif
		break;
	    }

	/* no builtin command found	 */
	if (cmd->name == NULL)
	{
	    verror ("unrecognized command: %s\n", targv[0]);

	    /* if at the CRT, give the user some advice  */
	    if (in == stdin)
	    {
		fprintf (stderr, "type 'help' for command summary\n");
	    }

	}

    }
}



/*
**   Function:     main
**
**   File:         vsim.c
**
**   Called by:    the operating system
**
**   Description:  the main entry point for the
**                 University of Pittsburgh VHDL Simulator
**
**   Arguments:    argc, argv - the old standards
**                 argv should contain one argument, the filename
**                 which contains the database format XXXXX.ivf
**
**   Return Value: None.
*/
main (argc, argv)
    int             argc;
    char           *argv[];

{
    char            local_fname[50];
    char            db_fname[50];
    int             c;
    extern int      optind;
    extern char    *optarg;
    int             err_flag = 0;
    static char     init_fname[50] = "NULL FILE NAME" ;
    
    outfile = stderr;

    fprintf (stderr, "    vsim VHDL Simulator, Version %d.%03d\n",
	     MAJOR_VERSION_NUM, MINOR_VERSION_NUM);
    fprintf (stderr, "Copyright (C) 1988, University of Pittsburgh\n\n");


    /* flag if this release is an alpha or beta release  */

#if (MINOR_VERSION_NUM >= 200)
    /* we have a beta release  */
    fprintf (stderr, "**  BETA Release  --  Limited Distribution  **\n\n");
#endif

#if ((MINOR_VERSION_NUM >= 100) && (MINOR_VERSION_NUM < 200))
    /* we have an alpha release  */
    fprintf (stderr, "**  ALPHA Release  --  Internal Pitt & Penn State Use Only  **\n\n");
#endif

    while ((c = getopt_2(argc, argv, "i:")) != EOF)
    {
        switch(c)
        {
            case 'i':   /*  we have an init file  */
            strcpy(init_fname, optarg);
            break;
            
            default:
            err_flag++;
        }
    }            

    if (err_flag || optind > argc)
    {
	usage ();
    }

    if (argc == optind+1)
    {
        /* parse the file name  */
        strcpy (local_fname, &argv[optind][strlen (argv[optind]) - 4]);

        /* make sure we have the proper extension  */
        if (strcmp (local_fname, ".ivf"))
        {
            usage();
        }
	else
	{
	aetname=(char *)malloc(strlen(argv[optind])+1);
	strcpy(aetname,argv[optind]);
	strcpy(aetname+strlen(argv[optind])-4,".aet");
	}
	
        /* all is ok, strip off the front for the initialization file  */
        strcpy (local_fname, argv[optind]);

        /*  nul-terminate the input file  */
        *strrchr (local_fname, '.') = '\0';
        
        /* append the '_init'  */
        strcat (local_fname, "_init");

        /*  copy the database file name  */
        strcpy (db_fname, argv[optind]);
    }
    else
    {
        /*  the user only provided the command 'vsim_xxx', strip off 'xxx'  */
        if ((argv[0][4] != '_') || (strlen(argv[0]) <= 5))
        {
            usage();
        }

        /*  setup the initialization file name  */
        /*  copy the base for the database and aet file name  */
        strcpy (local_fname, &argv[0][5]);

	/* build the aet filename */
	aetname=(char *)malloc(strlen(local_fname)+5);
	strcpy(aetname,local_fname);
	strcat(aetname,".aet");

        /* append the '_init'  */
        strcat (local_fname, "_init");

        /*  setup the database file name  */
        /*  copy the base for the database file name  */
        strcpy (db_fname, &argv[0][5]);

        /* append the '.ivf'  */
        strcat (db_fname, ".ivf");
    }
    
    /* call presim to set up network for simulation	 */
    presim (db_fname);

    /* settle network	 */
    init_nodes ();

    /*  tie the one and zero nodes high and low  */
    setin (find (SIG_ONE), 'h');
    setin (find (SIG_ZERO), 'l');
    setin (find (SIG_X), 'x');
    setin (find (SIG_Z), 'z');

    /*
     * attempt to read the initialization files in the following order:
     * ~/.vsim_init,  ./.vsim_init,  XXXXX_init where XXXXX is the same as
     * the database XXXXX.ivf loaded
     */

    finput ("~/.vsim_init");
    finput ("./.vsim_init");
    finput (local_fname);
    finput (init_fname);

    /*  setup the ^C handler  */
    signal (SIGINT, int_routine);

    /* read commands from user	 */
    debug = 0;
    filename = "tty";

    lineno = 0;
    input_command (stdin);

    exit (0);
}



usage ()
{
    	fprintf (stderr, " usage: vsim_xxx [-i init_file_name] [database.ivf]\n");
	fprintf (stderr, " database.ivf is only needed if different from 'xxx.ivf'\n");
	exit (1);
}

void int_routine (int dummy)
{
    fprintf (stderr,"\n-- Interrupt --\n");
    interrupt_flag = TRUE;
}
